{-# LANGUAGE ParallelListComp, RecordWildCards, NamedFieldPuns, TemplateHaskell, DeriveDataTypeable, DeriveGeneric, ScopedTypeVariables, TypeFamilies, RankNTypes, MultiParamTypeClasses, FlexibleInstances, FlexibleContexts #-}



{- |

Module: BaselMachine.MutableArena

Description : Data structures representing a mutable arena in which colonies of cells interact

Copyright   : © 2019 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)


-}

module BaselMachine.MutableArena where

import GHC.Generics

import Control.DeepSeq
import Control.Exception
import Control.Monad
import Control.Monad.ST

import Data.Aeson
import Data.Aeson.TH
import Data.Binary
import Data.Data
import Data.DeriveTH
import Data.Foldable
import Data.Functor.Identity
import qualified Data.Graph.Inductive.Graph as G
import Data.Graph.Inductive.Graph (Node, LNode, Edge, LEdge)
import Data.Graph.Inductive.PatriciaTree
import Data.IntMap (IntMap)
import qualified Data.IntMap as IntMap
import Data.IntSet (IntSet)
import qualified Data.IntSet as IntSet
import Data.Maybe
import Data.Sequence (
  Seq, (<|), (|>), ViewL(..), ViewR(..), viewl, viewr)
import qualified Data.Sequence as Seq
import Data.STRef
import Data.Typeable
import qualified Data.Vector.Generic as V
import qualified Data.Vector.Generic.Mutable as MV
import qualified Data.Vector.Unboxed as UV

import BaselMachine.CircularBuffer
import BaselMachine.Compiler
import BaselMachine.MVM
import BaselMachine.Signal
import BaselMachine.AesonUtil
import WGM.ShortShow
import WGM.Snapshot

-- === General implementation notes ===
--
-- I've tried to make state transitions run with as little
-- allocation as possible.  That means there's a lot of places
-- where I put changed objects into an STRef or something similar
-- rather than allocate a new record.
--
-- The reason for including the Arena in Basel Machine is that
-- for speed and memory efficiency, I need Arena code to have
-- access to the mutable virtual machines in each cell, but I
-- don't want anything outside of the Basel Machine module to
-- have that access.


data Source =
  Source
  {
    sourceCellIx :: !Int,
    sourcePattern :: !Int,
    sourceThreshold :: !Int,
    sourceActivationDelta :: !Int
  }
  deriving (Read, Show, Eq, Ord, Data, Generic, Typeable)

$(derive makeNFData ''Source)
$(derive makeBinary ''Source)
$(deriveJSON aesonOptions ''Source)

instance Monad m => Snapshot m Source



data Target =
  Target
  {
    targetCellIx :: !Int,
    targetPattern :: !Int,
    targetActivationDelta :: !Int
  }
  deriving (Read, Show, Eq, Ord, Data, Generic, Typeable)

instance Monad m => Snapshot m Target

$(derive makeNFData ''Target)
$(derive makeBinary ''Target)
$(deriveJSON aesonOptions ''Target)


data ChannelSpec =
  ChannelSpec
  {
    channelSource :: !Source,
    channelTarget :: !Target,
    channelDelay :: !Int
  }
  deriving (Read, Show, Eq, Ord, Data, Generic, Typeable)

instance Monad m => Snapshot m ChannelSpec

$(derive makeNFData ''ChannelSpec)
$(derive makeBinary ''ChannelSpec)
$(deriveJSON aesonOptions ''ChannelSpec)


data MChannelState s =
  ChannelState
  {
    channelSpec :: ChannelSpec,
    channelSpikeTrainState :: (CircularBuffer s)
  }

type ChannelState = TChannelState (Seq Bool)

$(derive makeNFData ''TChannelState)
$(derive makeBinary ''TChannelState)
$(deriveJSON aesonOptions ''TChannelState)



instance Snapshot (ST s) (MChannelState s) where
  type Memento (MChannelState s) = ChannelState

  snapshot ChannelState{..} = do
    spikeTrainMemento <- snapshot channelSpikeTrainState
    return ChannelState {
      channelSpikeTrainState = spikeTrainMemento,
      ..}



data TCellState (m :: * -> *) vm =
  CellState
  {
    cellVM :: vm,
    cellTotalReactionCount :: MBox m Int,
    cellTotalReactionDeltas :: MBox m Int
  }

$(derive makeNFData ''TCellState)
$(derive makeBinary ''TCellState)
$(deriveJSON aesonOptions ''TCellState)

type MCellState s =
  TCellState (ST s) (MVM s)

type CellState =
  TCellState Identity (UV.Vector Int)

instance Snapshot (ST s) (MCellState s) where

  type Memento (MCellState s) = CellState

  snapshot CellState{..} = do
    sVM <- snapshot cellVM
    trc <- snapshot cellTotalReactionCount
    trd <- snapshot cellTotalReactionDeltas
    return CellState {
      cellVM = sVM,
      cellTotalReactionCount = trc,
      cellTotalReactionDeltas = trd }


-- | The signal should be something immutable that can serve as its own 'Memento'
data BroadcastSpec signal =
  BroadcastSpec
  {
    broadcastSource :: !Source,
    broadcastMaximumAge :: !Int,
    broadcastTarget :: !Target,
    broadcastSignal :: !signal
  }
  deriving (Read, Show, Eq, Ord, Data, Generic, Typeable)

instance (Monad m, Snapshot m signal) =>
   Snapshot m (BroadcastSpec signal)


data TBroadcastState m location signal =
  BroadcastState
  {
    broadcastSpec :: BroadcastSpec signal,
    broadcastFromColonyIx :: Maybe Int,
    broadcastOrigin :: location,
    broadcastAge :: MBox m Int
  }


type MBroadcastState location signal s =
  TBroadcastState (ST s) location signal

type BroadcastState location signal =
  TBroadcastState Identity location signal

instance Snapshot (ST s) (MBroadcastState location signal s) where

  type Memento (MBroadcastState location signal s) =
    BroadcastState location signal

  snapshot BroadcastState{..} = do
    age <- snapshot broadcastAge
    return BroadcastState{
      broadcastAge = age,
      ..}

instance Snapshot m (BroadcastState location signal)



instance Reconstruct (ST s) (MBroadcastState location signal s) where
  reconstruct BroadcastState{..} = do
    ageRef <- reconstruct broadcastAge
    return BroadcastState{
      broadcastAge = ageRef,
      ..}



data ColonySpec signal =
  ColonySpec
  {
    colonyCellCount :: Int,
    colonyCellIxToName :: IntMap String,
    colonySynapseSpecs :: [ChannelSpec],
    colonyBroadcastSpecs :: [BroadcastSpec signal]
  }
  deriving (Read, Show, Eq, Ord, Data, Generic, Typeable)


data TColonyState m location signal cellState synapseState =
  ColonyState
  {
    colonySpec :: ColonySpec signal,
    colonyName :: String,
    colonyLocation :: MBox m location,
    colonyCellGraph :: MBox m (Gr cellState synapseState)
  }

type MColonyState location signal s =
  TColonyState (ST s) location signal (MCellState s) (MChannelState s)

type ColonyState location signal =
  TColonyState Identity location signal CellState ChannelState


instance
  Snapshot (ST s) (MColonyState location signal s) where

  type Memento (MColonyState location signal s) =
    ColonyState location signal

  snapshot ColonyState{..} = do
    location <- readSTRef colonyLocation
    cellGraphM <- snapshot colonyCellGraph
    return ColonyState {
      colonyLocation = location,
      colonyCellGraph = cellGraphM,
      ..}





data ColonyInitialState location signal =
  ColonyInitialState {
  initName :: String,
  initColonySpec :: ColonySpec signal,
  initProgram :: (ProgramInformation, ReactionProgram),
  initLocation :: location
  }
  deriving (Show, Generic, Typeable)




data TArenaState m location signal cellState synapseState =
  ArenaState
  {
    arenaColonyIxToState ::
      MBox m (IntMap (TColonyState m location signal cellState synapseState)),
    arenaBroadcastStates ::
      MBox m [TBroadcastState m location signal]
  }



type MArenaState location signal s =
  TArenaState (ST s) location signal (MCellState s) (MChannelState s)

type ArenaState location signal =
  TArenaState Identity location signal CellState ChannelState


instance
  Snapshot (ST s) (MArenaState location signal s) where

  type Memento (MArenaState location signal s) =
    ArenaState location signal

  snapshot ArenaState{..} = do
    colonyIxToState <- snapshot arenaColonyIxToState
    broadcastStates <- snapshot arenaBroadcastStates
    return ArenaState {
      arenaColonyIxToState = colonyIxToState,
      arenaBroadcastStates = broadcastStates,
      ..}



-- === Utilities ===

mkCheckedLookup ::
  (key -> map -> Bool)
  -> (key -> map -> Maybe value)
  -> key -> map -> value

mkCheckedLookup ck lk k m =
  assert (ck k m) $ fromJust (lk k m)


imGet = mkCheckedLookup IntMap.member IntMap.lookup


nodeLabel ::
  (G.Graph gr)
  =>
  Node -> gr nLabel eLabel -> nLabel
nodeLabel = mkCheckedLookup G.gelem lk
  where
    lk g n = G.lab n g


-- === Phases related to broadcasts ===

-- | Add external broadcasts to the arena
mArenaIncorporateInputs ::
  [BroadcastState location signal]
  -> MArenaState location signal s
  -> ST s ()

mArenaIncorporateInputs [] _ =
  return ()

mArenaIncorporateInputs inputMs ArenaState{..} = do
  externalBroadcasts <- reconstruct inputMs
  modifySTRef' arenaBroadcastStates (externalBroadcasts ++)

-- | Send broadcasts to agents
mArenaPhaseBroadcastSend ::
  forall location signal s.
  (BroadcastSignal location signal)
  =>
  MArenaState location signal s -> ST s ()

mArenaPhaseBroadcastSend ArenaState{..} = do
  broadcastStates <- readSTRef arenaBroadcastStates
  colonyIxToState <- readSTRef arenaColonyIxToState
  sequence_ $ do -- build a list of ST s actions
    BroadcastState{..} <- broadcastStates -- pull from list
    let BroadcastSpec{..} = broadcastSpec
        Target{..} = broadcastTarget
    (colIx, ColonyState{..}) <- IntMap.toList colonyIxToState
    return $ do -- build ST s action
      loc <- readSTRef colonyLocation
      cellGraph <- readSTRef colonyCellGraph
      age <- readSTRef broadcastAge
      let CellState{..} = nodeLabel targetCellIx cellGraph
          str = (strengthAt
                 broadcastSignal
                 broadcastOrigin
                 age
                 loc)
          delta = floor (str * fromIntegral targetActivationDelta)
      return (mvmApplyDelta cellVM targetPattern delta)


-- | Add 1 to the age of all broadcasts and get rid of any that
-- have passed their maximum age.
mArenaPhaseBroadcastIncrementAge ::
  MArenaState location signal s -> ST s ()

mArenaPhaseBroadcastIncrementAge ArenaState{..} = do
  broadcastStates <- readSTRef arenaBroadcastStates
  survivingBroadcasts <- filterM notTooOld broadcastStates
  forM_ survivingBroadcasts $ \ BroadcastState{..} -> do
    modifySTRef' broadcastAge (1 +)
  writeSTRef arenaBroadcastStates survivingBroadcasts
  where
    -- Return whether a broadcast can be aged another step before
    -- expiring
    notTooOld BroadcastState{..} = do
        age <- readSTRef broadcastAge
        return (age < broadcastMaximumAge broadcastSpec)



-- | Check all agents to see whether they are emitting a
-- broadcast.
mArenaPhaseBroadcastEmit ::
  forall location signal s.
  (BroadcastSignal location signal)
  =>
  MArenaState location signal s -> ST s ()

mArenaPhaseBroadcastEmit ArenaState{..} = do
  colonyIxToState <- readSTRef arenaColonyIxToState
  forM_ (IntMap.toList colonyIxToState) $ \ (colonyIx, colonyState) -> do
    let ColonyState{..} = colonyState
        ColonySpec{..} = colonySpec
        broadcastFromColonyIx = Just colonyIx
    cellGraph <- readSTRef colonyCellGraph
    newEmissions <- forM colonyBroadcastSpecs $ \ broadcastSpec -> do
      let BroadcastSpec{..} = broadcastSpec
          Source{..} = broadcastSource
          CellState{..} = nodeLabel sourceCellIx cellGraph
      a <- mvmReadActivation cellVM sourcePattern
      if a >= sourceThreshold
        then do
        mvmApplyDelta cellVM sourcePattern sourceActivationDelta
        broadcastOrigin <- readSTRef colonyLocation
        broadcastAge <- newSTRef 0
        return (Just BroadcastState{..})
        else
        return Nothing
    oldEmissions <- readSTRef arenaBroadcastStates
    writeSTRef arenaBroadcastStates ((catMaybes newEmissions) ++ oldEmissions)



-- === Synapse phase ===

-- | Check all cells to see whether they send a spike down a
-- synapse.
mColonyRunSynapseEmitters ::
  MColonyState location signal s
  -> ST s ()

mColonyRunSynapseEmitters ColonyState{..} = do
  cellGraph <- readSTRef colonyCellGraph
  -- run each pre-synaptic emitter
  forM_ (G.labEdges cellGraph) $ \ (senderIx, receiverIx, channelState) -> do
    -- assert senderIx == sourceCellIx
    let ChannelState{..} = channelState
        ChannelSpec{..} = channelSpec
        Source{..} = channelSource
        CellState{..} = nodeLabel sourceCellIx cellGraph
    a <- mvmReadActivation cellVM sourcePattern
    let spike = a >= sourceThreshold
    when spike $ do
      mvmApplyDelta cellVM sourcePattern sourceActivationDelta
    pushBackBool channelSpikeTrainState spike



-- | Checke all synapsese to see whether a spike arrived at the
-- end and must be received by a cell.
mColonyRunSynapseReceptors ::
  MColonyState location signal s
  -> ST s ()

mColonyRunSynapseReceptors ColonyState{..} = do
  cellGraph <- readSTRef colonyCellGraph
  -- run each post-synaptic receptor
  forM_ (G.labEdges cellGraph) $ \ (senderIx, receiverIx, channelState) -> do
    -- assert receiverIx == targetCellIx
    let ChannelState{..} = channelState
        ChannelSpec{..} = channelSpec
        Target{..} = channelTarget
        CellState{..} = nodeLabel targetCellIx cellGraph
    justArrived <- popFrontBool channelSpikeTrainState
    when justArrived $ do
      mvmApplyDelta cellVM targetPattern targetActivationDelta



-- | Run both the emission and reception parts of synapses.
mArenaPhaseSynapses ::
  MArenaState location signal s -> ST s ()

mArenaPhaseSynapses ArenaState{..} = do
  colonyIxToState <- readSTRef arenaColonyIxToState
  forM_ (IntMap.toList colonyIxToState) $ \ (colonyIx, colonyState) -> do
    mColonyRunSynapseEmitters colonyState
    mColonyRunSynapseReceptors colonyState



-- === Computation phase ===

-- | Run all reactions in a 'Colony'.  Accumulate the total count
-- of reactions that were run and the total absolute value of all
-- deltas applied.
mColonyRunReactions ::
  MColonyState location signal s
  -> ST s ()

mColonyRunReactions ColonyState{..} = do
  cellGraph <- readSTRef colonyCellGraph
  forM_ (G.labNodes cellGraph) $ \ (node, cellState) -> do
    let CellState{..} = cellState
    (reactionCount, reactionDeltas) <- mvmRunReactions cellVM
    modifySTRef' cellTotalReactionCount (reactionCount +)
    modifySTRef' cellTotalReactionDeltas (reactionDeltas +)


-- | Run all reactions in every 'Colony'.
mArenaPhaseReactions ::
  MArenaState location signal s -> ST s ()

mArenaPhaseReactions ArenaState{..} = do
  colonyIxToState <- readSTRef arenaColonyIxToState
  forM_ (IntMap.toList colonyIxToState) $ \ (colonyIx, colonyState) -> do
    mColonyRunReactions colonyState



-- === Single big step and monitor functions ===


mArenaStep ::
  forall location signal s.
  (BroadcastSignal location signal)
  =>
  ArenaStepMonitor location signal s
  -> [BroadcastState location signal]
  -> MArenaState location signal s
  -> ST s ()

mArenaStep ArenaStepMonitor{..} inputs a@ArenaState{..} = do

  monitorInit a

  mArenaIncorporateInputs inputs a
  mArenaPhaseBroadcastSend a
  monitorAfterBroadcastSend a

  mArenaPhaseBroadcastIncrementAge a
  monitorAfterBroadcastIncrementAge a

  mArenaPhaseSynapses a
  monitorAfterSynapses a

  mArenaPhaseReactions a
  monitorAfterReactions a

  mArenaPhaseBroadcastEmit a
  monitorAfterBroadcastEmit a


type ArenaPhaseMonitor location signal s =
  MArenaState location signal s -> ST s ()


data ArenaStepMonitor location signal s =
  ArenaStepMonitor {
  monitorInit ::
      ArenaPhaseMonitor location signal s,
  monitorAfterBroadcastSend ::
      ArenaPhaseMonitor location signal s,
  monitorAfterBroadcastIncrementAge ::
      ArenaPhaseMonitor location signal s,
  monitorAfterSynapses ::
      ArenaPhaseMonitor location signal s,
  monitorAfterReactions ::
      ArenaPhaseMonitor location signal s,
  monitorAfterBroadcastEmit ::
      ArenaPhaseMonitor location signal s
  }


nothingArenaPhaseMonitor :: ArenaPhaseMonitor location signal s
nothingArenaPhaseMonitor a = return ()

nothingArenaStepMonitor :: ArenaStepMonitor location signal s
nothingArenaStepMonitor =
  ArenaStepMonitor {
  monitorInit = nothingArenaPhaseMonitor,
  monitorAfterBroadcastSend = nothingArenaPhaseMonitor,
  monitorAfterBroadcastIncrementAge = nothingArenaPhaseMonitor,
  monitorAfterSynapses = nothingArenaPhaseMonitor,
  monitorAfterReactions = nothingArenaPhaseMonitor,
  monitorAfterBroadcastEmit = nothingArenaPhaseMonitor
  }


-- === Various Monitor Functions ===

logSnapshot ::
  (Snapshot (ST s) t)
  =>
  STRef s (Seq (Memento t))
  -> t
  -> ST s ()

logSnapshot archive x = logSnapshotSx archive id x


logSnapshotSx archive view x = do
  memento <- snapshot (view x)
  modifySTRef' archive (|> memento)



recordActiveBroadcasts ::
  STRef s (Seq [BroadcastState location signal])
  -> ArenaPhaseMonitor location signal s

recordActiveBroadcasts archive ArenaState{..} = do
  logSnapshot archive arenaBroadcastStates


-- === Running an arena ===

mArenaRun ::
  (BroadcastSignal location signal)
  =>
  ArenaStepMonitor location signal s
  -> (MArenaState location signal s -> ST s Bool)
  -> MArenaState location signal s
  -> Int -- ^ maximum number of steps to run
  -> [[BroadcastState location signal]] -- ^ inputs for each time step
  -> ST s ()

mArenaRun _ _ _ 0 _ = return ()

mArenaRun stepMonitor stopCheckAxn arenaState numSteps [] =
  mArenaRun stepMonitor stopCheckAxn arenaState numSteps [[]]

mArenaRun stepMonitor stopCheckAxn arenaState numSteps (input:moreInputs) = do
  mArenaStep stepMonitor input arenaState
  stopSignalled <- stopCheckAxn arenaState
  if stopSignalled
    then return ()
    else mArenaRun stepMonitor stopCheckAxn arenaState (numSteps - 1) moreInputs


runArena ::
  forall location signal.
  (Eq signal,
   Show signal,
   Typeable signal,
   BroadcastSignal location signal)
  =>
  [ColonyInitialState location signal]
  -> Maybe signal -- ^ stop when this signal is broadcast
  -> Int -- ^ maximum number of steps to run
  -> [[BroadcastState location signal]] -- ^ inputs to the arena
  -> Seq [BroadcastState location signal] -- ^ snapshots of all arena broadcasts

runArena colonyInits mStopSignal numSteps arenaInputs = runST $ do
  mArena <- makeMutableArena colonyInits
  broadcastArchive <- newSTRef Seq.empty
  let stepMonitor = nothingArenaStepMonitor {
        monitorAfterBroadcastEmit =
            logSnapshotSx broadcastArchive arenaBroadcastStates
        }
  mArenaRun stepMonitor stopCheckAxn mArena numSteps arenaInputs
  -- The rest of this should be simply
  -- > snapshot broadcastArchive
  -- but for some reason GHC can't do the type inference properly
  archive <- readSTRef broadcastArchive
  let archiveAsList = toList archive
  mementoList <- snapshot archiveAsList
  return (Seq.fromList mementoList)
  --snapshot broadcastArchive

  where
    stopCheckAxn :: MArenaState location signal s -> ST s Bool
    stopCheckAxn ArenaState{..} =
      case mStopSignal of
        Nothing -> return False -- never stop
        Just stopSignal -> do -- stop if particular signal emitted
          broadcastStates <- readSTRef arenaBroadcastStates
          let checks = [
                stopSignal == broadcastSignal broadcastSpec
                | BroadcastState{..} <- broadcastStates]
          return (or checks)


-- === Total reactions and deltas expended per colony ===

mColonyTotalReactionsAndDeltas ::
  MColonyState location signal s
  -> ST s (Int, Int)

mColonyTotalReactionsAndDeltas ColonyState{..} = do
  cellGraph <- readSTRef colonyCellGraph
  let cellIxStates = G.labNodes cellGraph
  reactionCounts <- forM cellIxStates $ \ (j, CellState{..}) -> readSTRef cellTotalReactionCount
  reactionDeltas <- forM cellIxStates $ \ (j, CellState{..}) -> readSTRef cellTotalReactionDeltas
  return (sum reactionCounts, sum reactionDeltas)




-- === Activation range analysis ===


data TCellActivationRangeTable v =
  CellActivationRangeTable {
  activationMin :: v Int,
  activationMax :: v Int
  }

type MCellActivationRangeTable s =
  TCellActivationRangeTable (UV.MVector s)

type CellActivationRangeTable =
  TCellActivationRangeTable UV.Vector

instance
  Snapshot (ST s) (MCellActivationRangeTable s) where
  type Memento (MCellActivationRangeTable s) = CellActivationRangeTable

  snapshot CellActivationRangeTable{..} = do
    frozenMin <- V.freeze activationMin
    frozenMax <- V.freeze activationMax
    return CellActivationRangeTable {
      activationMin = frozenMin,
      activationMax = frozenMax }



data TColonyActivationRangeTable v =
  ColonyActivationRangeTable (IntMap (TCellActivationRangeTable v))

type MColonyActivationRangeTable s =
  TColonyActivationRangeTable (UV.MVector s)

type ColonyActivationRangeTable =
  TColonyActivationRangeTable UV.Vector

instance Snapshot (ST s) (MColonyActivationRangeTable s)



data TArenaActivationRangeTable v =
  ArenaActivationRangeTable (IntMap (TColonyActivationRangeTable v))

type MArenaActivationRangeTable s =
  TArenaActivationRangeTable (UV.MVector s)

type ArenaActivationRangeTable =
  TArenaActivationRangeTable UV.Vector

instance Snapshot (ST s) (MArenaActivationRangeTable s)


updateCellActivationRangeTable ::
  MCellActivationRangeTable s
  -> MCellState s
  -> ST s ()

updateCellActivationRangeTable CellActivationRangeTable{..} CellState{..} = do
  forM_ [0.. mvmNumPatterns cellVM - 1] $ \j -> do
    a <- mvmReadActivation cellVM j
    MV.modify activationMin (min a) j
    MV.modify activationMax (max a) j



updateColonyActivationRangeTables ::
  MColonyActivationRangeTable s
  -> MColonyState location signal s
  -> ST s ()

updateColonyActivationRangeTables
  (ColonyActivationRangeTable colonyART)
  ColonyState{..} = do
  cellGraph <- readSTRef colonyCellGraph
  forM_ (G.labNodes cellGraph) $ \ (j, cellState) -> do
    let cellART = imGet j colonyART
    updateCellActivationRangeTable cellART cellState



updateArenaActivationRangeTables ::
  MArenaActivationRangeTable s
  -> MArenaState location signal s
  -> ST s ()

updateArenaActivationRangeTables
  (ArenaActivationRangeTable arenaART)
  ArenaState{..} = do
  colonyIxToState <- readSTRef arenaColonyIxToState
  forM_ (IntMap.toList colonyIxToState) $ \(j, colonyState) -> do
    let colonyART = imGet j arenaART
    updateColonyActivationRangeTables colonyART colonyState



-- === Post-run analysis ===

data ColonyTraceResult =
  ColonyTraceResult {
  trPerCellOutputPatterns :: IntMap IntSet,
  -- ^ indexed by cell index, a set of all patterns ever read
  -- from that cell for any form colony output

  trPerCellPresynapticPatterns :: IntMap IntSet,
  -- ^ indexed by cell index, a set of all patterns ever read
  -- from that cell by a synapse

  trPerCellActivationRanges :: IntMap (UV.Vector Int, UV.Vector Int),
  -- ^ indexed by cell index, each pair consists of arrays of the
  -- (minimum, maximum) activation levels of each pattern in that
  -- cell over the entire trace

  trOverallActivationRanges :: (UV.Vector Int, UV.Vector Int)
  -- ^ arrays of the (minimum, maximum) activation levels of each
  -- pattern in all cells over the entire trace
  } deriving (Typeable, Generic)


-- | Return an 'IntMap' from cell index to an 'IntSet' of all
-- patterns read from that cell to generate output.

perCellOutputPatterns :: ColonySpec signal -> IntMap IntSet

perCellOutputPatterns ColonySpec{..} =
  (IntMap.fromListWith IntSet.union
    [ (sourceCellIx, IntSet.singleton sourcePattern)
      | BroadcastSpec{..} <- colonyBroadcastSpecs,
        let Source{..} = broadcastSource])



-- | Return an 'IntMap' from cell index to an 'IntSet' of all
-- patterns read from that cell by a synapse.

perCellPresynapticPatterns :: ColonySpec signal -> IntMap IntSet

perCellPresynapticPatterns ColonySpec{..} =
  (IntMap.fromListWith IntSet.union
   [ (sourceCellIx, IntSet.singleton sourcePattern)
   | ChannelSpec{..} <- colonySynapseSpecs,
     let Source{..} = channelSource])



-- === Construction ===

data BadColonySpecException signal =
  BadColonySpec {
    badCellNames :: [(Int, String)],
    badBroadcastSpecs :: [BroadcastSpec signal],
    badSynapseSpecs :: [ChannelSpec]
    } deriving (Show, Eq, Typeable, Data, Generic)

instance (Typeable signal, Show signal, Eq signal) =>
  Exception (BadColonySpecException signal)


checkColonySpec ColonySpec{..} =
  if null badCellNames
     && null badSynapseSpecs
     && null badBroadcastSpecs
  then
    return ()
  else
    throw BadColonySpec{..}
  where
    cellIxBad j =
      j < 0 || j >= colonyCellCount

    cellNameBad (j, _) =
      cellIxBad j

    badCellNames =
      filter cellNameBad (IntMap.toList colonyCellIxToName)

    synapseBad ChannelSpec{..} =
      let
        Source{..} = channelSource
        Target{..} = channelTarget
      in
        cellIxBad sourceCellIx || cellIxBad targetCellIx

    badSynapseSpecs =
      filter synapseBad colonySynapseSpecs

    broadcastBad BroadcastSpec{..} =
      let
        Source{..} = broadcastSource
        Target{..} = broadcastTarget
      in
        cellIxBad sourceCellIx || cellIxBad targetCellIx

    badBroadcastSpecs =
      filter broadcastBad colonyBroadcastSpecs



-- | Return the highest pattern mentioned by any instruction from
-- the program that was compiled, or any of the synapses, inputs,
-- or emitters.

maxPattern :: ColonySpec signal -> ProgramInformation -> Int

maxPattern ColonySpec{..} ProgramInformation{..} =
  overallMax
  where
    maxRefReactions = patternsReferenced

    maxRefSyn = maximum (0:[
      max targetPattern sourcePattern
      | ChannelSpec{..} <- colonySynapseSpecs,
        let Source{..} = channelSource,
        let Target{..} = channelTarget])

    maxRefIn = maximum (0:[
      max targetPattern sourcePattern
      | BroadcastSpec{..} <- colonyBroadcastSpecs,
        let Source{..} = broadcastSource,
        let Target{..} = broadcastTarget
        ])

    overallMax = maximum [
      IntSet.findMax maxRefReactions,
      maxRefSyn,
      maxRefIn
      ]



-- | Make a mutable colony in an initial state

makeMutableColony ::
  (Eq signal,
   Show signal,
   Typeable signal)
  =>
  String  -- ^ name
  -> location -- ^ location
  -> ColonySpec signal -- ^ specification
  -> (ProgramInformation, ReactionProgram) -- ^ genetic program for the cells
  -> ST s (MColonyState location signal s)

makeMutableColony colonyName location colonySpec@ColonySpec{..} (progInfo, reactions) = do
  -- Throw an exception if something is off
  checkColonySpec colonySpec
  -- construct vitrual machines
  let vmStateSize = 1 + maxPattern colonySpec progInfo
  annotatedNodes <- forM [0 .. colonyCellCount - 1] $ \ j -> do
    cellVM <- makeZeroMVM vmStateSize reactions
    cellTotalReactionCount <- newSTRef 0
    cellTotalReactionDeltas <- newSTRef 0
    return (j, CellState{..})
  annotatedEdges <- forM colonySynapseSpecs $ \ channelSpec -> do
    let ChannelSpec{..} = channelSpec
        Source{..} = channelSource
        Target{..} = channelTarget
    channelSpikeTrainState <- synapseWire channelDelay
    return (sourceCellIx, targetCellIx, ChannelState{..})
  let cellGraph = G.mkGraph annotatedNodes annotatedEdges
  colonyCellGraph <- newSTRef cellGraph
  colonyLocation <- newSTRef location
  return ColonyState{..}



-- | Make a wire with @n@ zeros in it, and enough room for at
-- least @n+1@ items.

synapseWire :: Int -> ST s (CircularBuffer s)

synapseWire numSteps =
  let
    -- This is kind of overkill, but there's no point in skimping
    -- here.  There has to be room for at least two more bytes,
    -- so we can push a new byte in one end and pop one out the
    -- other, and at all times have room in the buffer so that
    -- the two pointers don't hit each other: numSteps + 1 here
    -- would theoretically be enough, except that it allows the
    -- two pointers to hit each other after the push and before
    -- the pop, and the resulting memento would indicate that the
    -- buffer is empty.
    q = quot (numSteps + 2) 16
    bSize = 16 * (q + 1)
  in do
     b <- makeCircularBuffer bSize
     replicateM_ numSteps (pushFront b 0)
     return b



-- | Make a mutable arena

makeMutableArena ::
  (Eq signal,
   Show signal,
   Typeable signal)
  =>
  [ColonyInitialState location signal]
  -> ST s (MArenaState location signal s)

makeMutableArena colonyInits = do
  mColonyInfos <- forM colonyInits $ \ ColonyInitialState{..} -> do
    makeMutableColony initName initLocation initColonySpec initProgram
  let colonyIxToState = IntMap.fromList [
        (j, mColony)
        | j <- [0..]
        | mColony <- mColonyInfos]
  arenaColonyIxToState <- newSTRef colonyIxToState
  arenaBroadcastStates <- newSTRef []
  return ArenaState{..}
