{-# LANGUAGE
UndecidableInstances,
FlexibleInstances,
MultiParamTypeClasses,
TypeFamilies,
RecordWildCards,
NamedFieldPuns,
ScopedTypeVariables,
DeriveDataTypeable,
DeriveGeneric,
TemplateHaskell,
FlexibleContexts,
OverloadedStrings #-}


module Digiostelium.BasicReport where

import GHC.Generics

-- import Control.DeepSeq
import Control.Exception
-- import Control.Monad.Trans

import Data.Aeson
import Data.Aeson.TH
-- import Data.Binary
-- import qualified Data.ByteString.UTF8 as U
import Data.Data
-- import Data.IntMap (IntMap)
-- import qualified Data.IntMap as IntMap
-- import Data.IntSet (IntSet)
-- import qualified Data.IntSet as IntSet
-- import Data.Map (Map)
-- import qualified Data.Map as Map
-- import Data.Typeable
-- import qualified Data.Yaml as Yaml
import Data.String

import Text.Pandoc.Builder
import Text.Printf

import Data.BitString

-- import BaselMachine.Arena
-- import BaselMachine.Compiler
-- import BaselMachine.EfficaciousAnalysis
import BaselMachine.Instruction
-- import BaselMachine.Signal

import Digiostelium.AesonUtil
import Digiostelium.Agent as Agent
import Digiostelium.Genome as G
import Digiostelium.GraphML
import Digiostelium.PandocUtil
-- import Digiostelium.PopulationDatabase
import Digiostelium.Problem
import Digiostelium.Query
import Digiostelium.Report
import Digiostelium.ReportException

import WGM.ShortShow

-- | Basic information that can be extracted from a single 'Agent', independent of the problem or genome archivetcture.
-- These are parellel to 'AgentResult'.
data BasicReportSpec =
  JustAgent
  | Minimal
  | Summary
  | Export
  | Program
  | GenomeDetails
  deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)

$(deriveJSON myJSONOpts ''BasicReportSpec)


-- | Information that can be retrieved from an 'Agent'.  Normally
-- I use 'identifier' for an agent's id number so as not to clash
-- with the identity function in the prelude.  I'm using 'id'
-- here rather than 'identifier' because a lot of this stuff will
-- get exported to JSON and those extra letters could add up to a
-- lot of unnecessary data.
data BasicReport genome rating =
  JustAgentReport (Agent genome rating)
  | MinimalAgentReport {
    id :: Integer,
    rating :: rating
    }
  | SummaryReport {
    id :: Integer,
    rating :: rating,
    lifespan :: Lifespan,
    parentIds :: [Integer],
    genomeLength :: Integer
    }
    -- ^ quick summary: id, rating, lifespan, parents, genome length
  | ExportReport {
    id :: Integer,
    genome :: Value
    }
    -- ^ genome information only; for moving genomes from one
    -- simulation to another, even one that solves a different
    -- problem, has different settings, etc.
  | ProgramReport {
    id :: Integer,
    rating :: rating,
    lifespan :: Lifespan,
    parentIds :: [Integer],
    genomeLength :: Integer,
    instructions :: [Instruction]
    }
    -- ^ summary of the agent plus a listing of its instructions,
    -- but no bit-level genome data; requires no significant
    -- computation
  | GenomeDetailsReport {
    id :: Integer,
    rating :: rating,
    lifespan :: Lifespan,
    parentIds :: [Integer],
    genomeLength :: Integer,
    geneDescriptions ::  [String],
    instructionDetails :: [(Instruction, BitString, BitString, EncodingDetails)]
    -- ^ architecture-independent instruction information, @(the
    -- instruction, codon bits, nucleotide bits, encoding details)@
    }
    -- ^ summary of the agent plus a complete listing of the
    -- genome and additional information; requires no
    -- significant computation
    deriving (Read, Show, Eq, Data, Typeable, Generic)

$(deriveJSON myJSONOpts ''BasicReport)



instance
  (Show (Signal problem),
   ShortShow (Interval genomeSpec),
   GenomeSpec genomeSpec) =>
  QueryIndividual genomeSpec problem BasicReportSpec where

  type QueryResult genomeSpec problem BasicReportSpec = BasicReport (Genome genomeSpec) (Rating problem)


  queryIndividual _genomeSpec _problem JustAgent a =
    JustAgentReport a

  queryIndividual _genomeSpec _problem Minimal a =
    minimalAgentReport a

  queryIndividual genomeSpec _problem Summary Agent{egg=Egg{..}, ..} =
    SummaryReport {
      id = identifier,
      genomeLength = G.genomeLength genomeSpec genome,
      ..}

  queryIndividual genomeSpec _problem Export Agent{egg=Egg{..}, ..} =
    ExportReport {
      id = identifier,
      genome = exportGenomeData genomeSpec genome
    }

  queryIndividual genomeSpec _problem Program Agent{egg=Egg{..}, ..} =
    ProgramReport {
      id = identifier,
      genomeLength = G.genomeLength genomeSpec genome,
      instructions = interpretGenome genomeSpec genome,
      ..}

  queryIndividual genomeSpec _problem GenomeDetails a =
    genomeDetailsReport genomeSpec a


-- | Produce a 'MinimalAgentReport'.  This is popped out of
-- 'queryIndividual' so that it can be called cleanly without
-- passing in unneeded objects.

minimalAgentReport ::
  Agent genome rating
  -> BasicReport genome rating

minimalAgentReport Agent{..} =
  MinimalAgentReport {
    id = identifier,
    ..
    }

-- | Produce a 'GenomeDetailsReport'.  This is popped out of
-- 'queryIndividual' so that it can be called cleanly without
-- passing in an unneeded problem object.

genomeDetailsReport ::
  (GenomeSpec genomeSpec,
   ShortShow (Interval genomeSpec))
  => genomeSpec
  -> Agent (Genome genomeSpec) rating
  -> BasicReport genome rating

genomeDetailsReport genomeSpec Agent{egg=Egg{..}, ..} =
    GenomeDetailsReport {
      id = identifier,
      genomeLength = G.genomeLength genomeSpec genome,
      geneDescriptions = map shortShow geneDetails,
      instructionDetails =
      [ (instruction, codonBits, nucleotideBits, instructionEncodingDetails)
      | G.GeneDetails{..} <- geneDetails],
    ..}
    where
      geneDetails = G.geneDetails genomeSpec genome


instance
  (ToJSON genome,
   Show rating,
   ToJSON rating) =>
  ToPandoc opt (BasicReport genome rating)



instance GetTitle opt (BasicReport genome rating) where

  getTitle _ (JustAgentReport _) =
    text "Agent code"

  getTitle _ MinimalAgentReport{..} =
    text ""

  getTitle _ SummaryReport{..} =
    text "Summary"

  getTitle _ ExportReport{..} =
    mempty

  getTitle _ ProgramReport{..} =
    text "Program Summary"

  getTitle _ GenomeDetailsReport{..} =
    text "Genome Details"



instance
  (ToJSON genome,
   Show rating,
    ToJSON rating) =>
  Snip opt (BasicReport genome rating) where

  toBlocks _ SummaryReport{..} =
    (header 1 $ text $ fromString $ printf "Agent %d" id)
    <>
    (para
     ("Rating:" <> space <> texts rating <> linebreak <>
      "Lifespan:" <> space <> texts lifespan <> linebreak <>
      "Parents:" <> space <> texts parentIds <> linebreak <>
      "Genome length:" <> space <> texts genomeLength))

  toBlocks _ MinimalAgentReport{..} =
    (header 1 $ text $ fromString $ printf "Agent %d" id)
    <>
    (para
     ("Rating:" <> space <> texts rating <> linebreak))

  toBlocks _ ProgramReport{..} =
    (header 1 $ text $ fromString $ printf "Agent %d" id)
    <>
    (para
     ("Rating:" <> space <> texts rating <> linebreak <>
      "Lifespan:" <> space <> texts lifespan <> linebreak <>
      "Parents:" <> space <> texts parentIds <> linebreak <>
      "Genome length:" <> space <> texts genomeLength))
    <>
    (header 2 "Instructions:")
    <>
    (instructionBlock (zip [0..] instructions))


  -- for other branches of BasicReport, convert to YAML
  toBlocks _ r =
    toYAMLCodeBlock r


instance ToGraphML opt (BasicReport genome rating) where

  toGraphML _ _r =
    throw (UnsupportedFormat "Expression as GraphML is not possible for a BasicReport")

-- | Just about everything should work with a 'BasicReportSpec'
instance
  (GenomeSpec genomeSpec,
   ToJSON (Genome genomeSpec),
   ShortShow (Interval genomeSpec),
   Show (Signal problem),
   Show (Rating problem),
   ToJSON (Rating problem)
   ) =>
  Reportable genomeSpec problem BasicReportSpec
