{-# LANGUAGE RecordWildCards, NamedFieldPuns, NoMonomorphismRestriction, FlexibleContexts #-}

{-|

Module: Digiostelium.ColonyToStyledNetwork

Description: Tools for drawing a Colony as a GraphML document

Copyright: © 2019 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

Implementation notes 2018-07-04: I went to a lot of trouble to
make it correctly mark the edges with the multiplicities of
reactions.  Copies of 'Reaction's are mostly kept around because
they differ in their index member.  I modified
'BaselMachine.EfficaciousAnalysis' to produce a map from
'Reaction's to their multiplicities, and to make it work, the
'Reaction's used as keys in that map are modified to have an
index of 0.  The main function in this module,
'efficaciousAnalysisToStyledNetwork', modifies the per-cell
'Reaction' sets by likewise setting their index fields to 0.  The
secondary function 'colonyToStyledNetwork' expects the
'Reaction's passed to it to also have zero 'index'.

-}

module Digiostelium.ColonyToStyledNetwork
  (
    colonyToStyledNetwork,
    efficaciousAnalysisToStyledNetwork
  ) where


import Data.IntMap.Strict (IntMap)
import qualified Data.IntMap.Strict as IntMap
import Data.Map.Strict (Map)
import qualified Data.Map.Strict as Map
import Data.Set (Set)
import qualified Data.Set as Set

import Text.Printf

import Digiostelium.StyledNetwork

import BaselMachine.Arena
import BaselMachine.EfficaciousAnalysis
import BaselMachine.Instruction

-- | (Private) Make an edge (e) key from a from-key and a to-key
eKey ::
  String
  -> String
  -> String

eKey fromKey toKey =
  printf "E[%s,%s]" fromKey toKey

-- | (Private) Make a cell+pattern (cp) key from the cell index and pattern
cpKey ::
  Int
  -> Int
  -> String

cpKey cellIx p =
  printf "CP[%d,%d]" cellIx p


-- | (Private) Make a network (n) key from a network label
nKey ::
  String
  -> String

nKey networkLabel =
  printf "Net[%s]" networkLabel


-- | (Private) Make a reaction (rxn) key from a reaction
rxnKey ::
  Reaction
  -> String

rxnKey Reaction{..} =
 printf "Rxn[%d,%d,%d,%d]" sPat thr tPat delta



dStyle :: Int -> EdgeStyle
dStyle delta =
  if delta > 0
  then ActivationEdge
  else InhibitionEdge


-- (Private) Make an edge label from a threshold, a delta, and a delay
dLabel :: Int -> Int -> Int -> String
dLabel threshold delta delay =
  if delay == 0
  then
    printf "≥%d⇒%+d" threshold delta
  else
    printf "≥%d⇒%+d [%d]" threshold delta delay


-- | (Private) Objects generated as we traverse a 'ColonySpec'

data Object =
  PNode Int Int StyledNode -- ^ cell ix, pattern, node
  | SigNode StyledNode -- ^ these are always intercellular
  | Edge NetIx StyledEdge


-- | (Private) Which cell an 'Edge' belongs to, or intercellular.

data NetIx =
  CellNet Int
  | Intercellular

traverseColonySpec :: SignalToNode signal => ColonySpec signal -> [Object]
traverseColonySpec ColonySpec{..} =
  let
    fromSynapses = map traverseSynapseSpec colonySynapseSpecs
    fromSensors = map traverseSensorSpec colonySensorSpecs
    fromBroadcasts = map traverseBroadcastSpec colonyBroadcastSpecs
  in concat [concat fromSynapses, concat fromSensors, concat fromBroadcasts]

traverseSynapseSpec :: ChannelSpec -> [Object]
traverseSynapseSpec ChannelSpec{
  channelSource = Source{..},
  channelTarget = Target{..},
  ..} =
  let
    sourceKey = cpKey sourceCellIx sourcePattern
    targetKey = cpKey targetCellIx targetPattern
    edgeKey = eKey sourceKey targetKey
    sourceNode =
      StyledNode {
      nodeKey = sourceKey,
      nodeLabel = printf "%d" sourcePattern,
      nodeStyle = SynapseSourceNode
      }
    targetNode =
      StyledNode {
      nodeKey = targetKey,
      nodeLabel = printf "%d" targetPattern,
      nodeStyle = SynapseTargetNode
      }
    synapseEdge =
      StyledEdge {
      fromKey = sourceKey,
      toKey = targetKey,
      edgeKey,
      edgeLabel = dLabel sourceThreshold targetActivationDelta channelDelay,
      edgeStyle = dStyle targetActivationDelta
      }
    in [
    -- source node
    PNode sourceCellIx sourcePattern sourceNode,
    -- target node
    PNode targetCellIx targetPattern targetNode,
    -- edge
    Edge Intercellular synapseEdge]


traverseSensorSpec :: SignalToNode signal => SensorSpec signal -> [Object]
traverseSensorSpec SensorSpec{
  sensorTarget = Target{..},
  ..} =
  let
    sigNode = signalToNode sensorSignal
    targetNodeKey = cpKey targetCellIx targetPattern
    sensorEdgeKey = eKey (nodeKey sigNode) targetNodeKey

    targetNode =
      StyledNode {
      nodeLabel = printf "%d" targetPattern,
      nodeKey = targetNodeKey,
      nodeStyle = SensorNode
      }
    sensorEdge =
      StyledEdge {
      edgeKey = sensorEdgeKey,
      fromKey = nodeKey sigNode,
      toKey = nodeKey targetNode,
      edgeLabel = printf "%+d" targetActivationDelta,
      edgeStyle = dStyle targetActivationDelta
      }
    in [
    -- input source node
    SigNode sigNode,
    -- input target node
    PNode targetCellIx targetPattern targetNode,
    -- edge
    Edge Intercellular sensorEdge]


traverseBroadcastSpec :: SignalToNode signal => BroadcastSpec signal -> [Object]
traverseBroadcastSpec BroadcastSpec{
  broadcastSource = Source{..},
  ..} =
  let
    sourceNodeKey = cpKey sourceCellIx sourcePattern
    sigNode = signalToNode broadcastSignal
    edgeKey = eKey sourceNodeKey (nodeKey sigNode)

    sourceNode =
      StyledNode {
      nodeKey = sourceNodeKey,
      nodeLabel = printf "%d" sourcePattern,
      nodeStyle = OutputSourceNode
      }
    broadcastEdge =
      StyledEdge {
      edgeKey,
      edgeLabel = printf "≥%d" sourceThreshold,
      fromKey = nodeKey sourceNode,
      toKey = nodeKey sigNode,
      edgeStyle = ActivationEdge
      }
    in [
    -- broadcast source node
    PNode sourceCellIx sourcePattern sourceNode,
    -- target node
    SigNode sigNode,
    -- edge
    Edge Intercellular broadcastEdge]


traverseReaction :: Map Reaction Int -> Int -> Reaction -> [Object]
traverseReaction multiplicityMap cellIx r@Reaction{..} = do
  let
    targetNodeKey = cpKey cellIx tPat
    switchNodeKey = cpKey cellIx sPat
    edgeKey = rxnKey r

    switchNode =
      StyledNode {
      nodeKey = switchNodeKey,
      nodeLabel = show sPat,
      nodeStyle = OrdinaryNode
      }
    targetNode =
      StyledNode {
      nodeKey = targetNodeKey,
      nodeLabel = show tPat,
      nodeStyle = OrdinaryNode
      }

    n :: Int
    n = Map.findWithDefault 0 r multiplicityMap

    multiplicityMark :: String
    multiplicityMark =
      if n == 1
      then ""
      else
      printf " ×%d" n

    reactionEdge =
      StyledEdge {
      edgeKey,
      fromKey = nodeKey switchNode,
      toKey = nodeKey targetNode,
      edgeLabel = printf "≥%d⇒%+d%s" thr delta multiplicityMark,
      edgeStyle = dStyle delta
      }
    in [
    -- switch node
    PNode cellIx sPat switchNode,
    -- target node
    PNode cellIx tPat targetNode,
    -- edge
    Edge (CellNet cellIx) reactionEdge]



traverseReactions :: Map Reaction Int -> Int -> Set Reaction -> [Object]
traverseReactions multiplicityMap cellIx reactions =
  concatMap (traverseReaction multiplicityMap cellIx) (Set.toList reactions)


-- | (Private) Go through @m@, find the item associated with key,
-- which is a list, and prepend the given value to it.
insertPrepend :: Int -> value -> IntMap [value] -> IntMap [value]
insertPrepend key value m =
  IntMap.insertWith j key [value] m
  where
    j newValue oldValue =
      newValue ++ oldValue

-- | (Private) Go through a list of objects and sort them into
-- 'Intercellular' and those that belong in a particular cellular
-- sub-graph.  Order of 'Object's is preserved.

collectByCellIx :: [Object] -> ([Object], IntMap [Object])
collectByCellIx = collectByCellIxLoop [] IntMap.empty

collectByCellIxLoop :: [Object] -> IntMap [Object] -> [Object] -> ([Object], IntMap [Object])
collectByCellIxLoop intercellular cellular [] =
  (reverse intercellular, fmap reverse cellular)
collectByCellIxLoop intercellular cellular (object:objects) =
  case object of
    PNode cellIx _ _ ->
      (collectByCellIxLoop
       intercellular
       (insertPrepend cellIx object cellular)
       objects)

    SigNode _ ->
      (collectByCellIxLoop
       (object : intercellular)
       cellular
       objects)

    Edge (CellNet cellIx) _ ->
      (collectByCellIxLoop
       intercellular
       (insertPrepend cellIx object cellular)
       objects)

    Edge Intercellular _ ->
      (collectByCellIxLoop
       (object : intercellular)
       cellular
       objects)


-- | (Private) Go through the objects generated by traversing a
-- 'Colony' and collect the pattern nodes ('PNode's), and if
-- there are duplicates for a given cell index and pattern, keep
-- only the first one.  (That way if a pattern in a cell has some
-- special meaning, a specially styled node can be created for it
-- first, and it will override any duplicate nodes created when
-- scanning instructions.)

collectPNodes :: [Object] -> Map (Int, Int) StyledNode
collectPNodes objects =
  (foldl Map.union Map.empty
    [ Map.singleton (cellIx, pat) node
    | PNode cellIx pat node <- objects])



-- | (Private) Assuming all the given objects belong to the same
-- network, build the 'StyledNetwork' object
-- buildSubnet :: Int -> String -> [Object] -> StyledNetwork
buildSubnet :: Int -> String -> [Object] -> StyledNetwork
buildSubnet cellIx networkLabel objects =
  let
    networkKey = nKey networkLabel
    pNodeMap = collectPNodes objects

    subnetworks = []
    nodes = Map.elems pNodeMap
    edges = [e | Edge (CellNet j) e <- objects, cellIx == j]
  in
    StyledNetwork{..}


colonyToStyledNetwork ::
  (SignalToNode signal)
  =>
  Map Reaction Int
  -- ^ map from a 'Reaction' (with 'index' set to 0) to how many
  -- copies of that reaction are present in the genome
  -> String
  -- ^ overall label for the network
  -> ColonySpec signal
  -> IntMap (Set Reaction)
  -- ^ map from cell index to its efficacious 'Reaction's;
  -- assumes their 'index' fields have been set to 0.
  -> StyledNetwork

colonyToStyledNetwork multiplicityMap networkLabel colonySpec@ColonySpec{..} cellToReactions =
  let
    objsFromSpec = traverseColonySpec colonySpec
    objssFromReactions =
      [traverseReactions multiplicityMap cellIx reactions
      | (cellIx, reactions) <- IntMap.toList cellToReactions]

    traversalObjects = objsFromSpec ++ concat objssFromReactions
    (intercellularObjects, cellularObjects) = collectByCellIx traversalObjects

    cellNetworks =
      [buildSubnet cellIx cellName (IntMap.findWithDefault [] cellIx cellularObjects)
      | (cellIx, cellName) <- IntMap.toList colonyCellIxToName]

    inSrcNodes = [n | SigNode n <- intercellularObjects, InputSourceNode == nodeStyle n]
    outTgtNodes = [n | SigNode n <- intercellularObjects,
                   elem (nodeStyle n) [OutputTargetNode, StopNode]]
    otherIntercellularNodes =
      [n | SigNode n <- intercellularObjects,
          not (elem (nodeStyle n) [InputSourceNode, OutputTargetNode, StopNode])]
    intercellularEdges =
        [e | Edge Intercellular e <- intercellularObjects]

    inSubnetKey = nKey "Input"
    outSubnetKey = nKey "Output"
    fullNetworkKey = nKey networkLabel

    inSubnet = StyledNetwork {
        networkLabel = "Input",
        networkKey = inSubnetKey ,
        nodes = inSrcNodes,
        edges = [],
        subnetworks = []
        }

    outSubnet = StyledNetwork {
        networkLabel = "Output",
        networkKey = outSubnetKey,
        nodes = outTgtNodes,
        edges = [],
        subnetworks = []
        }
  in StyledNetwork{
    networkKey = fullNetworkKey,
    networkLabel,
    nodes = otherIntercellularNodes,
    edges = intercellularEdges,
    subnetworks = inSubnet : outSubnet : cellNetworks
    }



efficaciousAnalysisToStyledNetwork ::
  (SignalToNode signal) =>
  EfficaciousAnalysis
  -> String
  -> ColonySpec signal
  -> StyledNetwork

efficaciousAnalysisToStyledNetwork effAn networkLabel colonySpec =
  colonyToStyledNetwork multiplicityMap networkLabel colonySpec cellToReactions
  where
    multiplicityMap = Map.fromList (allEfficaciousReactionCounts effAn)
    cellToReactions =
      IntMap.map (Set.map discardIndex . cellEfficaciousReactionSet) (perCellEfficacious effAn)


discardIndex :: Reaction -> Reaction
discardIndex r = r{index=0}
