{-# LANGUAGE
FlexibleInstances,
MultiParamTypeClasses,
TemplateHaskell,
DeriveDataTypeable,
DeriveGeneric,
RecordWildCards,
NamedFieldPuns
#-}

{-|

Module: Digiostelium.CopyProblemBase

Description : Definitions common to 'Digistelium.CopyProblem' and 'Digistelium.PairCopyProblem'

Copyright   : © 2019 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

-}

module Digiostelium.CopyProblemBase
where

import GHC.Generics

import Control.DeepSeq

import Data.Aeson.TH
import Data.Binary
import Data.Data
import qualified Data.List as List
import Data.Sequence (Seq, ViewR(..), viewr)
import qualified Data.Sequence as Seq
import Data.String

import Text.Pandoc.Builder
import Text.Printf

import BaselMachine.Arena

import Digiostelium.AesonUtil
import Digiostelium.PandocUtil
import Digiostelium.StyledNetwork

import BaselMachine.Signal

import WGM.Util


data CopyProblemSignal =
  SenderRole
  | ReceiverRole
  | Stop
  | Beacon
  | InputBit !Int
  | OutputBit !Int
  | TransmitBit !Int -- only used by PairCopyProblem
  deriving (Eq, Ord, Read, Show, Data, Generic, Typeable)

instance NFData CopyProblemSignal
instance Binary CopyProblemSignal
$(deriveJSON myJSONOpts ''CopyProblemSignal)

instance BroadcastSignal location CopyProblemSignal where
  strengthAt _ _ _ _ = 1.0

data CopyProblemInput = CopyProblemInput {
  inputBits :: [Int] -- ^ list of inputs bits set
  }
  deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)

instance NFData CopyProblemInput
$(deriveJSON defaultOptions ''CopyProblemInput)

data CopyProblemOutput = CopyProblemOutput {
  numTimeSteps :: Int,
  -- ^ how long until the stop signal was sent
  nonDataOutputs :: [CopyProblemSignal],
  -- ^ output signals active at the stop time, other than the data bits
  dataBits :: [Int]
  -- ^ data bit signals active at the stop time
  }
  deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)

instance NFData CopyProblemOutput
$(deriveJSON myJSONOpts ''CopyProblemOutput)

isInputBit :: CopyProblemSignal -> Bool
isInputBit (InputBit _) = True
isInputBit _ = False


-- | (Private) Produce all sublists of 0 .. n-1
allInputBitLists :: Int -> [[Int]]
allInputBitLists n = reverse (map reverse (allSubLists [n-1, n-2 .. 0]))


-- | (Private) Count the number of elements of 0 .. n-1 common to
-- both lists.
countCorrectBits :: Int -> [Int] -> [Int] -> Int
countCorrectBits n inputs outputs =
  sum [fromEnum ((j `elem` inputs) == (j `elem` outputs))
      | j <- [0 .. n-1] ]



-- | Given a list of indexes, return a string where the
-- characters at those indices are 1 and the rest are 0.
patternIxsTo01s :: Int -> [Int] -> String
patternIxsTo01s nMax ks =
  [ if j `elem` ks then '1' else '0'
  | j <- [0 .. (nMax - 1)]]



instance SignalToNode CopyProblemSignal where
  signalToNode SenderRole =
    StyledNode {
     nodeKey = "SenderRole",
     nodeLabel = "S",
     nodeStyle = InputSourceNode }

  signalToNode ReceiverRole =
    StyledNode {
      nodeKey = "ReceiverRole",
      nodeLabel = "R",
      nodeStyle = InputSourceNode }

  signalToNode Stop =
    StyledNode {
      nodeKey = "Stop",
      nodeLabel = "Stop",
      nodeStyle = StopNode }

  signalToNode Beacon =
    StyledNode {
      nodeKey = "Beacon",
      nodeLabel = "B",
      nodeStyle = OutputTargetNode }

  signalToNode (InputBit n) =
    StyledNode {
      nodeKey = printf "In[%d]" n,
      nodeLabel = printf "%d" n,
      nodeStyle = InputSourceNode }

  signalToNode (OutputBit n) =
    StyledNode {
      nodeKey = printf "Out[%d]" n,
      nodeLabel = printf "%d" n,
      nodeStyle = OutputTargetNode }

  signalToNode (TransmitBit n) =
    StyledNode {
      nodeKey = printf "Trans[%d]" n,
      nodeLabel = printf "%d" n,
      nodeStyle = TransmitNode }


-- | Convert input to string of 0s and 1s.
-- The 'Int' parameter is interpreted as the word size.

instance Sniplet Int CopyProblemInput where
  toInlines wordSize CopyProblemInput{..} =
    code $ fromString $ patternIxsTo01s wordSize inputBits


-- | Convert output to 0s and 1s, and how many steps.
-- The 'Int' parameter is interpreted as the word size.

instance Sniplet Int CopyProblemOutput where
  toInlines wordSize CopyProblemOutput{..} =
    code
    $ fromString
    $ (printf "%s (%d steps) %s"
       (patternIxsTo01s wordSize dataBits)
       numTimeSteps
       (show nonDataOutputs))



genericArenaOutputToCopyProblemOutput
  :: Seq
    [BroadcastState t CopyProblemSignal]
  -> CopyProblemOutput

genericArenaOutputToCopyProblemOutput colonyOutputSeq =
  CopyProblemOutput{..}
  where
    numTimeSteps = Seq.length colonyOutputSeq
    _ :> lastBroadcast = viewr colonyOutputSeq
    nonDataOutputs =
      [ broadcastSignal
      | BroadcastState {..} <- lastBroadcast,
        not (isInputBit broadcastSignal)]
    dataBits =
      List.sort [ j
           | BroadcastState {broadcastSignal = OutputBit j, ..} <- lastBroadcast]


allInputsFromWordSize ::
  Int -- ^ word size
  -> [CopyProblemInput] -- ^ all inputs of the given word size

allInputsFromWordSize wordSize =
  [ CopyProblemInput{..}
  | inputBits <- allInputBitLists wordSize]
