{-# LANGUAGE BangPatterns, RecordWildCards, NamedFieldPuns, ScopedTypeVariables,
  NoMonomorphismRestriction, StandaloneDeriving, DeriveDataTypeable, TemplateHaskell, DeriveGeneric #-}

{-# OPTIONS_GHC -fno-ignore-asserts #-}

module Digiostelium.Database (
  -- * Opening and closing
  DatabaseSpec(..),
  mkStdDBSpec,
  Database(..),
  openDatabase,
  closeDatabase,
  DBMonad,
  withDatabase,

  -- * Direct access
  dbGetSpec,
  dbTouch,
  dbPut,
  dbGet,
  dbGetMany,
  dbHasKey,
  KeyNotFoundException,
  dbUnsafeGet,
  dbDelete,

  -- * 'DBMonad' 'Cursor' traversal
  Cursor(),
  withCursor,
  withSubcursor,
  getAtCursor,
  hopCursorToFirst,
  hopCursorToFirst_,
  hopCursorToLast,
  hopCursorToLast_,
  hopCursorNext,
  hopCursorNext_,
  hopCursorPrevious,
  hopCursorPrevious_,
  hopCursorTo,
  putAtCursor,
  putWithCursor,
  deleteAtCursor,
  cursorForEach,
  cursorForEach_,
  cursorSearch,
  cursorWhile,

  -- * Transactions
  withTransaction
  ) where

--import Debug.Trace

import GHC.Generics

import Control.DeepSeq
import Control.Monad
import Control.Monad.Trans.State
import Control.Monad.IO.Class
import Control.Exception
import qualified Control.Exception as E

import System.FilePath
import System.Directory
import System.Process
import System.Exit

import Data.Aeson.TH
import Database.Berkeley.Db
import qualified Data.ByteString.Lazy as LB -- used by Binary
import qualified Data.ByteString as SB -- used by Db *sigh*
import qualified Data.Binary as Bin
import Data.Data
import Data.IORef
import Data.Maybe
import qualified Data.Map as M

import Digiostelium.AesonUtil
import Digiostelium.Logging


{- Only needed for making debugging messages
debugMessageTrucateTo :: Int
debugMessageTrucateTo = 16
-}

data DatabaseSpec =
  DatabaseSpec {
    environmentDirectory :: FilePath
    }
  deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)

instance NFData DatabaseSpec
$(deriveJSON myJSONOpts ''DatabaseSpec)

-- | Make a standard spec.  The environment directory is really
-- the only field.

mkStdDBSpec :: FilePath -> DatabaseSpec
mkStdDBSpec environmentDirectory =
  DatabaseSpec { .. }

data Database tableName =
  Database {
    dbSpec :: DatabaseSpec,
    environment :: DbEnv,
    currentTransaction :: Maybe DbTxn,
    tableFromNameRef :: IORef (M.Map tableName Db)
    }

{- Bug fix: I was keeping tableFromName as an ordinary member and
updating it through the usual StateT mechanism.  The problem is
that when you do (bracket open close axn) any changes to the
state object made by axn are lost when close is run, because
close is passed the original state object.  Instead,
tableFromName has to be kept as a mutable pointer, so databases
opened during an action are noted in the original object and
closed correctly. -}

-- | (High level) Open a population database, or create an empty
-- one if none exists.

openDatabase :: DatabaseSpec -> IO (Database tableName)
openDatabase dbSpec@(DatabaseSpec {..}) = do
  dirExists <- doesDirectoryExist environmentDirectory
  when (not dirExists) (createDirectory environmentDirectory)
  -- Create a DoNotIndex file in the directory so that I can make sure tracker isn't
  -- going through gigs of stuff:
  -- ad = add directory
  let ad filePath = (environmentDirectory </> filePath)
  writeFile (ad "DoNotIndex") ""
  environment <- openEnvironment environmentDirectory
  let currentTransaction = Nothing
  tableFromNameRef <- newIORef M.empty
  return Database {..}

closeDatabase :: Show tableName => Database tableName -> IO ()
closeDatabase (Database {..}) = do
  tableFromName <- readIORef tableFromNameRef
  msgDebug ("closeDatabase: starting")
  msgDebug ("closeDatabase: open tables: " ++ show (M.keys tableFromName))
  forM_ (M.toList tableFromName) $ \(tableName, table) -> do
    msgDebug ("closeDatabase: closing table " ++ show tableName)
    closeTable table
  excMsg "closeDatabase" $ dbEnv_close [] environment
  msgDebug ("closeDatabase: done")

{-- Database interface helpers --}

-- | Open a 'DbEnv' with a certain amount of housekeeping.  For
-- some reason, a database created on one machine sometimes
-- triggers @DB_VERSION_MISMATCH@ error when opened on another
-- (possibly a 32 bit / 64 bit clash).  Also, when you compress
-- and restore a database file, you get weird errors from cursors
-- on it.  These errors seem to be harmless and easy to fix by
-- removing the __db.001, ... files before doing anything else.

-- I'm still working on whether these additional flags actually
-- accomplish anything, and are there any bad side effects, such as
-- a slow-down or huge log files.

openEnvironment :: FilePath -> IO DbEnv
openEnvironment envDir = excMsg "openEnvironment" $ do
  env <- dbEnv_create []
  excMsg "openEnvironment.2" $ (
    dbEnv_open [DB_CREATE,
                DB_INIT_MPOOL,
                DB_INIT_LOCK, DB_INIT_TXN, DB_INIT_LOG, DB_RECOVER,
                DB_THREAD] 0 env envDir)
  return env

-- | Helper functions for cleanly closing a table in the database.

closeTable :: Db -> IO ()
closeTable db = do
  excMsg "closeTable" $ db_close [] db

-- | Helper function for cleanly opening a table in the database,
-- possibly decompressing the file.

openTable :: DbEnv -> FilePath -> FilePath -> IO Db
openTable environment environmentDirectory fileName = do
  -- If the requested file doesn't exist but a .bz2 does,
  -- uncompress and try again
  fileEx <- doesFileExist (environmentDirectory </> fileName)
  if fileEx
    then openOrCreateTable environment fileName
    else do
    let fileNameZ = environmentDirectory </> fileName <.> "bz2"
    fileEx <- doesFileExist fileNameZ
    if fileEx
      then do
      msgInfo ("Decompressing " ++ fileNameZ)
      (eCode, bzOut, bzErr) <-
        readProcessWithExitCode "bzip2" (["-d"] ++ [fileNameZ]) ""
      case eCode of
        ExitSuccess -> do
          msgInfo ("Done")
          openTable environment environmentDirectory fileName
        ExitFailure n -> do
          msgError ("Decompression failed with error code " ++ show n)
          msgError ("Decompression reported (stdout) " ++ bzOut)
          msgError ("Decompression reported (stderr) " ++ bzErr)
          exitWith (ExitFailure 1)
      else openOrCreateTable environment fileName


-- | Lower level than 'openTable'.  Helper function for opening a
-- database file or creating a fresh one if none is present. No
-- automatic decompression.

openOrCreateTable :: DbEnv -> FilePath -> IO Db
openOrCreateTable environment fileName = excMsg "openOrCreateTable" $ do
  msgDebug ("openOrCreateTable: Opening table file " ++ fileName)
  -- Construct a db handle
  db <- (excMsg "openOrCreateTable.1"
        $ db_create [] environment)
  -- Set the page size to 16K (will be replaced with whatever
  -- page size was used to create the file originally, but this
  -- value will be used for new files)
  -- Must be called before db_open.
  (excMsg "openOrCreateTable.2"
        $ db_set_pagesize db 16384)
  -- Try to open the file and attach it to this handle.
  -- Create the file if it doesn't exist
  (excMsg "openOrCreateTable.3"
   $ db_open [DB_CREATE] DB_BTREE 0 db Nothing fileName Nothing)

  return db

type DBMonad tableName = StateT (Database tableName) IO

withDatabase :: (Show tableName) =>
  DatabaseSpec -> (DBMonad tableName c) -> IO c

withDatabase dbSpec axn =
  (bracket -- close the database even if an exception occurs during axn
   (openDatabase dbSpec)
   closeDatabase
   (\dbState -> evalStateT axn dbState))

-- | Return the 'DatabaseSpec' used to open this database.
dbGetSpec :: DBMonad tableName DatabaseSpec
dbGetSpec =
  gets dbSpec

-- | Open a table on demand.

getTable :: (Show tableName, Ord tableName) => tableName -> DBMonad tableName Db
getTable tableName = do
  -- msgDebug ("getTable: request for table " ++ show tableName)
  Database {..} <- get
  let DatabaseSpec {..} = dbSpec
  tableFromName <- liftIO $ readIORef tableFromNameRef
  case (M.lookup tableName tableFromName) of
    Just db -> do
      -- msgDebug ("getTable: table already open")
      return db
    Nothing -> do
      -- msgDebug ("getTable: must open " ++ show tableName)
      db <- liftIO $ openTable environment environmentDirectory (show tableName)
      liftIO $ writeIORef tableFromNameRef (M.insert tableName db tableFromName)
      return db

-- | Make sure a table exists and is available in the monad.

dbTouch :: (Ord tableName, Show tableName) =>
           tableName -> DBMonad tableName ()
dbTouch tableName = do
  _ <- getTable tableName
  return ()

-- | Check for the existence of a table.

doesTableExist :: (Show tableName, Ord tableName) => tableName -> DBMonad tableName Bool
doesTableExist tableName = do
  Database {..} <- get
  let DatabaseSpec {..} = dbSpec
  liftIO $ doesFileExist (environmentDirectory </> show tableName)

adapt :: (Show tableName, Ord tableName) =>
         tableName
         -> (Db -> Maybe DbTxn -> IO b)
         -> DBMonad tableName b

adapt tableName axn = do
  Database {..} <- get
  db <- getTable tableName
  liftIO $ do
    axn db currentTransaction

dbPut :: (Bin.Binary value, Bin.Binary key, Show tableName, Ord tableName) =>
         key
         -> value
         -> tableName
         -> DBMonad tableName ()
dbPut key value tableName = adapt tableName (dbPutTxn key value)

-- | Fetch a value from the table if possible.  Implementation
-- note: Do not create an empty table if the requested table does
-- not already exist.

dbGet :: (Bin.Binary value, Bin.Binary key, Show tableName, Ord tableName) =>
         key
         -> tableName
         -> DBMonad tableName (Maybe value)

dbGet key tableName = do
  f <- doesTableExist tableName
  if f
    then adapt tableName (dbGetTxn key)
    else return Nothing


-- | Check whether the table includes a key.  Implementation
-- note: Do not create an empty table if the requested table does
-- not already exist.

dbHasKey :: (Bin.Binary key, Show tableName, Ord tableName) =>
         key
         -> tableName
         -> DBMonad tableName Bool

dbHasKey key tableName = do
  f <- doesTableExist tableName
  if f
    then let
    keyEnc = encodeToDB key
    in do
      db <- getTable tableName
      mValueEnc <- liftIO $ excMsg "dbHasKey" $ db_get [] db Nothing keyEnc
      return (isJust mValueEnc)
    else
    return False


-- | Load as many of the requested values as possible, returning
-- an a-list.  Implementation note: Check for the existence of
-- the table once, then use the underlying 'dbGetTxn' function
-- repeatedly without checking again.

dbGetMany :: (Bin.Binary value, Bin.Binary key, Show tableName, Ord tableName) =>
             [key]
             -> tableName
             -> DBMonad tableName [(key, value)]

dbGetMany keys tableName = do
  f <- doesTableExist tableName
  if f
    then do
    mItems <- forM keys $ \ key -> do
      mValue <- adapt tableName (dbGetTxn key)
      case mValue of
        Nothing -> return Nothing
        Just value -> return (Just (key, value))
    return (catMaybes mItems)
    else
    return []


data KeyNotFoundException = KeyNotFoundException String
                          deriving (Eq, Ord, Read, Show, Typeable)

instance E.Exception KeyNotFoundException


-- | Get the record and throw an exception if it's not there.
dbUnsafeGet :: (Bin.Binary value, Bin.Binary key, Show tableName, Ord tableName) =>
               key
               -> tableName
               -> DBMonad tableName value
dbUnsafeGet key tableName = do
  mResult <- dbGet key tableName
  case mResult of
    Just r -> return r
    Nothing ->
      liftIO $ throwIO (KeyNotFoundException (show $ encodeToDB key))

dbDelete :: (Bin.Binary key, Show tableName, Ord tableName) =>
            key
            -> tableName
            -> DBMonad tableName ()

dbDelete key tableName = adapt tableName (dbDeleteTxn key)

withTransaction :: DBMonad tableName r -> DBMonad tableName r
withTransaction axn = do
  s0@(Database {..}) <- get
  -- To avoid resource leaks with open-on-demand database files,
  -- we have to keep most of the new state object
  (r,s1) <- liftIO
           $ dbEnv_withTxn [] [] environment currentTransaction
           $ \ txn -> do
             runStateT axn (s0 {currentTransaction = Just txn})
  -- But we absolutely have to get rid of the just-run transaction
  put (s1 {currentTransaction = currentTransaction})
  return r

-- These db???Txn functions are a layer below the easy access
-- functions and take an explicit transaction parameter.  My design
-- was to do everything through a withTransaction function, but
-- since transactions turned out to be uselessly slow and run out of
-- resources when the simulation was running, I gave up on them.

dbPutTxn :: (Bin.Binary key, Bin.Binary value) =>
  key -> value -> Db -> Maybe DbTxn -> IO ()

dbPutTxn key value db txn =
  let
    keyEnc = (encodeToDB key)
    valueEnc = (encodeToDB value)
  in do
    -- msgDebug ("dbPutTxn: keyEnc=" ++ take debugMessageTrucateTo (showAsHex keyEnc))
    -- msgDebug ("dbPutTxn: valueEnc=" ++ take debugMessageTrucateTo (showAsHex valueEnc))
    excMsg "dbPutTxn" $ db_put [] db txn keyEnc valueEnc

dbGetTxn :: (Bin.Binary key, Bin.Binary value) =>
  key -> Db -> Maybe DbTxn -> IO (Maybe value)

dbGetTxn key db txn =
  let
    keyEnc = (encodeToDB key)
  in do
    -- msgDebug ("dbGetTxn: keyEnc=" ++ take debugMessageTrucateTo (showAsHex keyEnc))
    mValueEnc <- excMsg "dbGetTxn" $ db_get [] db txn keyEnc
    case mValueEnc of
      Nothing ->
        return Nothing
      Just valueEnc -> do
        -- msgDebug ("dbGetTxn: valueEnc=" ++ take debugMessageTrucateTo (showAsHex valueEnc))
        return (Just (decodeFromDB valueEnc))

dbDeleteTxn :: (Bin.Binary key) =>
  key -> Db -> Maybe DbTxn -> IO ()

dbDeleteTxn key db txn =
  let
    keyEnc = (encodeToDB key)
  in do
    -- msgDebug ("dbDeleteTxn: keyEnc=" ++ take debugMessageTrucateTo (showAsHex keyEnc))
    excMsg "dbDeleteTxn" $ db_del [] db txn keyEnc

encodeToDB :: Bin.Binary t => t -> SB.ByteString
encodeToDB x =
  let lazyBS = Bin.encode x
      ws = LB.unpack lazyBS
      stBS = SB.pack ws
  in stBS

decodeFromDB :: Bin.Binary t => SB.ByteString -> t
decodeFromDB stBS =
  let ws = SB.unpack stBS
      lazyBS = LB.pack ws
      x = Bin.decode lazyBS
  in x

{-
-- A debugging function
showAsHex :: SB.ByteString -> [Char]
showAsHex byteStr =
  let
    bytes :: [Word8]
    bytes = SB.unpack byteStr
    hexReps = map (\b -> showHex b "") bytes
  in
   concat hexReps
-}

--- Cursors ---

data Cursor = Cursor { internalCursor :: DbCursor }

-- | Create a cursor on a database and apply an action to it,
-- safely closing the cursor and attaching it to any active
-- transaction.

withCursor :: (Show tableName, Ord tableName) =>
              tableName
              -> (Cursor -> DBMonad tableName b)
              -> DBMonad tableName b

withCursor tableName axn = do
  originalState@(Database {..}) <- get
  db <- getTable tableName
  (v, newState) <- liftIO $ excMsg "withCursor" $ do
    db_withCursor [] db  currentTransaction $ \cursor ->
      runStateT  (axn (Cursor cursor)) originalState
  put newState
  return v

withSubcursor :: Cursor -> (Cursor -> DBMonad tableName a) -> DBMonad tableName a
withSubcursor (Cursor {..}) axn = do
  originalState@(Database {..}) <- get
  (v, newState) <- liftIO $
    dbCursor_withCursor [] internalCursor $ \subCursor ->
    runStateT (axn (Cursor subCursor)) originalState
  put newState
  return v

deriving instance Show DbFlag

{- A complication here: I like to think in terms of python
indices (which are sort of between items in a list) and
C++/STL iterators, but Berkeley DB doesn't quite do this.

If a cursor has not been initialized,
calling dbCursor_get with DB_NEXT moves it to the first record
and returns that; calling with DB_PREV moves it to the last
record and returns that.  So an uninitialized cursor effectively
points to the zero-index prior to the first record, like container.begin().

If a cursor is on the last record and you call dbCursor_get with
DB_NEXT, it returns Nothing but the cursor isn't moved to an
index just past the last element like container.end().  When you
call dbCursor_get with DB_CURRENT on such a cursor, it returns
the last element.  This means you have to be careful about
handling the end when searching through an entire database.

-}

getCursorAdapter :: (Bin.Binary key, Bin.Binary value) =>
                    [DbFlag] -> Cursor -> DBMonad tableName (Maybe (key, value))
getCursorAdapter flags (Cursor {..}) = do
  mRecord <- (liftIO
              $ excMsg ("getCursorAdapter " ++ show flags)
              $ dbCursor_get flags internalCursor)
  case mRecord of
    Nothing ->
      return Nothing
    Just (keyBytes, valueBytes) ->
      return $ Just (decodeFromDB keyBytes, decodeFromDB valueBytes)

hopCursorAdapter_ :: [DbFlag] -> Cursor -> DBMonad tableName Bool
hopCursorAdapter_ flags (Cursor {..}) = do
  mRecord <- (liftIO
              $ excMsg ("hopCursorAdapter_" ++ show flags)
              $ dbCursor_get flags internalCursor)
  return (isJust mRecord)


-- | Return the key-value pair at the cursor as @Just (key,
-- value)@ , or 'Nothing' if the cursor is not at a record (which
-- I don't think is actually possible because of the slightly
-- messed up way cursors work here...)

getAtCursor :: (Bin.Binary key, Bin.Binary value) =>
               Cursor -> DBMonad tableName (Maybe (key, value))
getAtCursor = getCursorAdapter [DB_CURRENT]

-- | Move the cursor to the first record and return it.
hopCursorToFirst :: (Bin.Binary key, Bin.Binary value) =>
                    Cursor -> DBMonad tableName (Maybe (key, value))
hopCursorToFirst = getCursorAdapter [DB_FIRST]

-- | Move the cursor to the first record.  Return 'True' unless
-- the database is empty.

hopCursorToFirst_ :: Cursor -> DBMonad tableName Bool
hopCursorToFirst_ = hopCursorAdapter_ [DB_FIRST]

-- | Move the cursor to the last record and return it.

hopCursorToLast :: (Bin.Binary key, Bin.Binary value) =>
                   Cursor -> DBMonad tableName (Maybe (key, value))
hopCursorToLast = getCursorAdapter [DB_LAST]

-- | Move the cursor to the last record.  Return 'True' unless
-- the database is empty.

hopCursorToLast_ :: Cursor -> DBMonad tableName Bool
hopCursorToLast_ = hopCursorAdapter_ [DB_LAST]

-- | Move the cursor to the next record and return it.
hopCursorNext :: (Bin.Binary key, Bin.Binary value) =>
                 Cursor -> DBMonad tableName (Maybe (key, value))
hopCursorNext = getCursorAdapter [DB_NEXT]

-- | Move the cursor to the next record if possible.  Return
-- 'True' if the motion was successful.

hopCursorNext_ :: Cursor -> DBMonad tableName Bool
hopCursorNext_ = hopCursorAdapter_ [DB_NEXT]


-- | Move the cursor to the previous record and return it.
hopCursorPrevious :: (Bin.Binary key, Bin.Binary value) =>
                     Cursor -> DBMonad tableName (Maybe (key, value))
hopCursorPrevious = getCursorAdapter [DB_PREV]

-- | Move the cursor to the previous record. Return 'True' if the
-- motion was successful.

hopCursorPrevious_ :: Cursor -> DBMonad tableName Bool
hopCursorPrevious_ = hopCursorAdapter_ [DB_PREV]

hopCursorTo :: (Bin.Binary key, Bin.Binary value) =>
                 key -> Cursor -> DBMonad tableName (Maybe (key, value))
hopCursorTo key (Cursor {..})= do
  mValue <- liftIO $ excMsg "hopToCursor" $ dbCursor_set [] internalCursor (encodeToDB key)
  case mValue of
    Nothing ->
      return Nothing
    Just valueBytes ->
      return $ Just (key, decodeFromDB valueBytes)

-- | Replace the data at the record pointed to by the cursor.

putAtCursor :: (Bin.Binary value) =>
                value -> Cursor -> DBMonad tableName ()

putAtCursor value (Cursor {..}) =
   liftIO $ excMsg "putAtCursor" $ dbCursor_put [DB_CURRENT] internalCursor (SB.empty) (encodeToDB value)

putWithCursor :: (Bin.Binary key, Bin.Binary value) =>
                 key -> value -> Cursor -> DBMonad tableName ()

putWithCursor key value (Cursor {..}) =
  liftIO $ dbCursor_put [] internalCursor (encodeToDB key) (encodeToDB value)

deleteAtCursor :: Cursor -> DBMonad tableName ()

deleteAtCursor (Cursor {..}) =
  liftIO $ dbCursor_del [] internalCursor

cursorForEach :: (Bin.Binary key, Bin.Binary value) =>
                 Cursor -> (key -> value -> DBMonad tableName a) -> DBMonad tableName [a]
cursorForEach c axn = do
  mRecord <- hopCursorNext c
  case mRecord of
    Nothing ->
      return []
    Just (key, value) -> do
      result <- axn key value
      moreResults <- cursorForEach c axn
      return (result : moreResults)

-- | Advance the cursor and apply the action to the record there.
-- To apply this to every record, you have to give it an un-moved
-- cursor.  -}

cursorForEach_ :: (Bin.Binary key, Bin.Binary value) =>
                 Cursor
                 -> (key -> value -> DBMonad tableName a)
                 -> DBMonad tableName ()
cursorForEach_ c axn = do
  mRecord <- hopCursorNext c
  case mRecord of
    Nothing ->
      return ()
    Just (key, value) -> do
      _ <- axn key value
      cursorForEach_ c axn

-- | Given an action that can reject a key-value pair (returning
-- 'Nothing') or accept and process it (returning 'Just result'),
-- hop the cursor until the first accepted pair is found, and return
-- 'Just result', or 'Nothing' if no acceptable pair is found.  The
-- cursor starts at its current record, and is left pointing to
-- the acceptable pair, or at the end of its contents if non can be
-- found.

cursorSearch :: (Bin.Binary key, Bin.Binary value) =>
                 Cursor ->
                 (key -> value -> DBMonad tableName (Maybe b))
                 -> DBMonad tableName (Maybe b)
cursorSearch c axn = do
  mRecord <- getAtCursor c
  case mRecord of
    Nothing -> -- at the end
      return Nothing
    Just (key, value) -> do -- got a record, check it
      mResult <- axn key value
      case mResult of
        Nothing -> do -- not the one we want, keep going
          more <- hopCursorNext_ c
          if more
            then cursorSearch c axn
            else return Nothing
        Just result -> -- found it
          return (Just result)

-- | Given a cursor and a predicate action, repeatedly keep the
-- current item if the predicate returns 'True', and hop the cursor.
-- Return a list of pairs that yielded 'True' in the order they were
-- visited.  The cursor starts at its current record, and is left
-- pointing to the first pair resulting in 'False', or at the end of
-- its contents.

cursorWhile :: (Bin.Binary key, Bin.Binary value) =>
               Cursor
               -> (key -> value -> DBMonad tableName Bool)
               -> DBMonad tableName [(key, value)]

{- See comment at the top of this section for why this is so
absurdly complicated. -}

cursorWhile c testAxn = helper []
  where
    helper soFar = do
      mRecord <- getAtCursor c
      case mRecord of
        Nothing -> do
          return (reverse soFar)
        Just p@(key, value) -> do
          check <- testAxn key value
          if check
            then do
            more <- hopCursorNext_ c
            if more
              then helper (p:soFar)
              else return (reverse (p:soFar))
            else return (reverse soFar)
