{-# LANGUAGE DeriveDataTypeable, ScopedTypeVariables, RecordWildCards, DeriveGeneric, ConstraintKinds, FlexibleContexts, OverloadedStrings #-}



{-|
Module      : Digiostelium.EngineUtil

Description : Various loading, checking, and handling utilities for Engine and related executables.

Copyright   : © 2019 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

-}

module Digiostelium.EngineUtil where

import GHC.Generics

import Control.Concurrent
import Control.Exception
import Control.Monad
import Control.Monad.State

import Data.Aeson hiding (Array)
import qualified Data.Aeson as J
import qualified Data.Aeson.Types as JT
import qualified Data.ByteString as ByteString
import qualified Data.ByteString.Lazy.UTF8 as LU
import Data.Data
import Data.Map (Map)
import qualified Data.Map as M
import Data.Time.Clock
import Data.Time.Format
import Data.Time.LocalTime
import qualified Data.Yaml as Yaml
import Data.Yaml.Pretty

-- import qualified Shelly

import System.Directory
import System.Exit
import System.FilePath
import System.IO
import qualified System.Log.Handler as LH
import qualified System.Log.Handler.Simple as LHS
import System.Log.Formatter
import System.Log.Logger
import System.Posix.Signals

import Text.Printf

import Digiostelium.AesonUtil
import Digiostelium.Database
import Digiostelium.Genome
import Digiostelium.Logging
import Digiostelium.PandocUtil
import Digiostelium.PopulationDatabase
import Digiostelium.Problem
import Digiostelium.ProblemInventory

import System.Posix.FileMutex

-- | Make a log handler that sends messages to a 'Handle'
logToHandle :: Handle -> Priority -> IO (LHS.GenericHandler Handle)
logToHandle dest logLevel = do
  h0 <- LHS.streamHandler dest logLevel
  let h1 = LH.setFormatter h0
           (simpleLogFormatter "-\n priority: $prio\n timeStamp: $time\n logger: $loggername\n message: $msg")
      h = LH.setLevel h1 logLevel
  return h

-- | Make a log handler that sends messages to a file.
logToFile :: FilePath -> Priority -> IO (LHS.GenericHandler Handle)
logToFile path logLevel = do
  dest <- openFile path AppendMode
  hSetBuffering dest LineBuffering
  h <- logToHandle dest logLevel
  return h{LHS.closeFunc = hClose}

-- | Add a handler to the global logger that sends messages to a
-- file.
sendLogToFile :: FilePath -> Priority -> IO ()
sendLogToFile path logLevel = do
  h <- logToFile path logLevel
  updateGlobalLogger rootLoggerName $ addHandler h

handleConfigurationEx :: ConfigurationException -> IO any
handleConfigurationEx ex = do
  msgError "Error while interpreting configuration:"
  case ex of
    UnknownObject{..} ->
      msgError $ printf "Unknown %s requested: %s" objectType objectName
    DoNotOverwriteFile{..} ->
      msgError $ printf "File %s exists but shouldn't be overwritten" path
    FileDoesNotExist{..} ->
      msgError $ printf "File %s does not exist" path
    DirectoryDoesNotExist{..} ->
      msgError $ printf "Directory %s does not exist" path
    FileIsEmpty{..} ->
      msgError $ printf "File %s exists but is empty" path
    BadRating{..} ->
      msgError
       (printf "Unable to interpret JSON object as rating\n%s\n%s"
        (LU.toString $ J.encode badRating)
        parseErrorMessage)
    ConflictingSettings{..} ->
      msgError $ printf "Conflicting settings: %s" message
    MalformedConfiguration{..} ->
      msgError $ printf "Malformed configuration %s: %s" path parseErrorMessage
    ImpossibleReport{..} ->
      msgError $ printf "Impossible report request: %s" message
    EmptyConfiguration{..} ->
      msgError $ printf "Empty configuration %s" path

  exitWith (ExitFailure 10)


-- | Exception signalling a problem reading the configuration
data ConfigurationException =
  -- | An unknown object was requested
  UnknownObject {
    objectType :: String, -- ^ type (problem, genome architecture, etc)
    objectName :: String -- ^ name of requested object
    }
  |
  -- | A file exists that would be overwritten
  DoNotOverwriteFile {
    path :: FilePath -- ^ path to a file that exists but shouldn't
    }
  |
  -- | A file that needs to be read does not exit
  FileDoesNotExist {
    path :: FilePath -- ^ path to a file that should exist but doesn't
    }
  |
  -- | A directory that needs to be read does not exit
  DirectoryDoesNotExist {
    path :: FilePath -- ^ path to a directory that should exist but doesn't
    }
  |
  -- | A file that needs to be read is empty
  FileIsEmpty {
    path :: FilePath -- ^ path to a file that exists but is empty
    }
  |
  -- | Could not interpret a rating
  BadRating {
    badRating :: Value, -- ^ JSON value that could not be parsed
    parseErrorMessage :: String
    }
  |
  -- | Conflicting settings
  ConflictingSettings {
    message :: String
    }
  |
  -- | Requested report is not possible
  ImpossibleReport {
    message :: String
    }
  |
  -- | Malformed configuration file
  MalformedConfiguration {
    path :: FilePath, -- ^ path to read that resulted in malformed configuration
    parseErrorMessage :: String -- ^ message from 'fromJSON'
    }
  |
  -- | No configuation found
  EmptyConfiguration {
    path :: FilePath
    }
  deriving (Show, Eq, Data, Generic, Typeable)

instance Exception ConfigurationException


checkedLookup :: Map String t -> String -> String -> IO t
checkedLookup inventory objectType name = do
  unless (M.member name inventory) $
    throwIO (UnknownObject objectType name)
  return (inventory M.! name)


getProblemTemplate :: String -> IO (Box UsefulProblem)
getProblemTemplate = checkedLookup problemTemplates "problem"

getGenomeSpecTemplate :: String -> IO (Box UsefulGenomeSpec)
getGenomeSpecTemplate = checkedLookup genomeSpecTemplates "genome architecture"

assertDoNotOverwrite :: MonadIO m => FilePath -> m ()
assertDoNotOverwrite path = liftIO $ do
  x <- doesFileExist path
  when x $ throwIO DoNotOverwriteFile{..}


assertFileExists :: MonadIO m => FilePath -> m ()
assertFileExists path = liftIO $ do
  x <- doesFileExist path
  unless x $ throwIO FileDoesNotExist{..}

assertDirectoryExists :: MonadIO m => FilePath -> m ()
assertDirectoryExists path = liftIO $ do
  x <- doesDirectoryExist path
  unless x $ throwIO DirectoryDoesNotExist{..}

assertFileNotEmpty :: MonadIO m => FilePath -> m ()
assertFileNotEmpty path = liftIO $ do
  assertFileExists path
  n <- withFile path ReadMode $ \h ->
    hFileSize h
  when (n == 0) $ throwIO FileIsEmpty{..}


-- | Alias for 'checkedLoadLayeredYAML'

checkedLoadYAML ::
  (MonadIO m,
  FromJSON t) =>
  FilePath -> m t

checkedLoadYAML = checkedLoadLayeredYAML

-- | Given a path @d1/d2/.../dn/name.yaml@, attempt to load
-- @d1/d2/.../dn/name.yaml@,
-- @d1/d2/.../d(n-1)/name.yaml@, ...,
-- @d1/name.yaml@,
-- @name.yaml@.
-- Merge the resulting 'Value' objects, giving priority to the more deeply nested files.

checkedLoadLayeredYAML ::
  (MonadIO m,
  FromJSON t) =>
  FilePath -> m t

checkedLoadLayeredYAML filePath0 =
  liftIO $ do
  msgDebug (printf "checkedLoadLayeredYAML: Loading from path '%s'" filePath0)
  filePath1 <- makeAbsolute filePath0
  msgDebug (printf "checkedLoadLayeredYAML: Absolute path '%s'" filePath1)
  filePath <- canonicalizePath filePath1
  msgDebug (printf "checkedLoadLayeredYAML: Canonicalized to '%s'" filePath)
  mv <- checkedLoadLayeredYAMLHelper filePath Nothing
  case mv of
    Nothing ->
      throwIO (EmptyConfiguration filePath)
    Just v -> do
      msgDebug (printf "checkedLoadLayeredYAML: Loaded this object from '%s'" filePath)
      msgDebug (show (encode v))
      case fromJSON v of
        JT.Error msg ->
          throwIO (MalformedConfiguration filePath msg)
        JT.Success x ->
          return x

-- | (Private)
checkedLoadLayeredYAMLHelper ::
  FilePath -> Maybe Value -> IO (Maybe Value)

checkedLoadLayeredYAMLHelper filePath mUpperLayer = do
  fileExists <- doesFileExist filePath
  mObject <- if fileExists then
    withFile filePath ReadMode $ \h -> do
               n <- hFileSize h
               if n == 0 then
                 return Nothing
                 else do
                 v <- readYAMLFile filePath
                 msgDebug (printf "checkedLoadLayeredYAMLHelper: Loaded this object from '%s'" filePath)
                 msgDebug (show (encode v))
                 return (Just v)
    else
    return Nothing
  let mMerged = mergeJSONTreeM mUpperLayer mObject
      pathElements = init (splitPath filePath)
  if pathElements == ["/"] || null pathElements then
    return mMerged
    else do
    let upElements = init pathElements -- all but the last
        upDir = joinPath upElements
        upPath = upDir </> takeFileName filePath
    msgDebug (printf "checkedLoadLayeredYAMLHelper: looking at upPath '%s' built from upElements %s"
             upPath (show upElements))
    checkedLoadLayeredYAMLHelper upPath mMerged



checkedLoadSingleYAML ::
  (MonadIO m,
   FromJSON t) =>
  FilePath -> m t

checkedLoadSingleYAML filePath = liftIO $ do
  assertFileNotEmpty filePath
  readYAMLFile filePath


checkedLoadObject ::
  forall c t m. (FromJSON t, c t, MonadIO m)
  =>
  FilePath
  -> Impl c t
  -> m (Box c)
checkedLoadObject filePath Impl = liftIO $ do
  (p :: t) <- checkedLoadLayeredYAML filePath
  return (Box p)


readYAMLFile :: FromJSON t => FilePath -> IO t
readYAMLFile filePath = do
  mp <- Yaml.decodeFileEither filePath
  case mp of
    Left pError ->
      throwIO pError
    Right p ->
      return p

loadProblem :: FilePath -> String -> IO (Box UsefulProblem)
loadProblem filePath name =
  withNamedObject
 problemInventory
 (checkedLoadObject filePath)
 (throwIO (UnknownObject "problem" name))
 name

loadGenomeSpec :: FilePath -> String -> IO (Box UsefulGenomeSpec)
loadGenomeSpec filePath name =
  withNamedObject
 genomeArchitectureInventory
 (checkedLoadObject filePath)
 (throwIO (UnknownObject "genome architecture" name))
 name


-- | Exceptions thrown when a run hits some constraint, such as
-- running out of time.
data RunConstraintException =
  ExternalStopSignal
  | MaxSecondsElapsed
  deriving (Read, Show, Eq, Data, Typeable, Generic)

instance Exception RunConstraintException


-- | Exceptions thrown when a run hits some important condition, such as
-- time for a snapshot, but doesn't need to stop.
data RunNotificationException =
  SnapshotSecondsElapsed
  deriving (Read, Show, Eq, Data, Typeable, Generic)

instance Exception RunNotificationException


-- | Install signal handlers for SIGINT, SIGHUP, SIGTERM, and
-- SIGUSR1.  When one of these signals is received, set a flag
-- (to stop gracefully) and reset the signal handler as described
-- below.  Make and return an IO action called a /sig checker/
-- that throws 'ExternalStopSignal' if the flag has been set.
-- The idea is to keep the sig checker, and to run it every so
-- often inside a time-consuming loop in IO, surrounded by an
-- exception handler, so the loop can exit gracefully if a signal
-- is received.
--
-- Here's how the signal handler is set after a signal is
-- received: Pass in in @Just (threadId, execption)@ to specify
-- that a thread should be notified and the given execption
-- thrown to it if another signal is received; and the signal
-- handler is returned to its default value.  Pass in @Nothing@
-- to set the handler to its default value, which is probably to
-- exit forcefully.
makeSigStopper ::
  (MonadIO m, Exception e) =>
  Maybe (ThreadId, e) -- ^ thread to send an exception to, and execption to send
  -> IO (m ())
makeSigStopper mToNotify = do
  stopper <- newMVar False
  let sigs = [sigINT, sigHUP, sigUSR1, sigTERM]
      stop sig = do
        msgInfo (printf "Caught signal %s; calling for stop" (show sig))
        _v <- swapMVar stopper True
        _ <- case mToNotify of
          Nothing -> do
            msgDebug (printf "Setting signal handler to default for %s" (show sig))
            installHandler sig Default Nothing
          Just (threadId, e) -> do
            let seriouslyStop = do
                  msgWarning (printf "Caught %s again; throwing exception" (show sig))
                  throwTo threadId e
                  _ <- installHandler sig Default Nothing
                  return ()
            msgDebug (printf "Setting serious handler for %s" (show sig))
            installHandler sig (Catch seriouslyStop) Nothing
        return ()

  forM_ sigs $ \sig ->
    installHandler sig (Catch (stop sig)) Nothing

  return $ liftIO $ do
    b <- readMVar stopper
    when b (throwIO ExternalStopSignal)


-- | Given a 'Maybe' and an action, if the 'Maybe' object is
-- 'Just x', perform the action, otherwise do a do-nothing
-- action.  This is used to build checker actions out of
-- configuration that may or may not specify a contstraint to
-- check.

maybeHelperIO ::
  (MonadIO m) =>
  Maybe t -- ^ maybe an object
  -> (t -> IO (m ())) -- ^ action to perform
  -> IO (m ())
maybeHelperIO m axn =
  case m of
    Nothing -> return (return ())
    Just x ->
      axn x

-- | Make an action called a /clock checker/ that throws
-- the given execption if the current time is more than
-- @maxClockSeconds@ from when the checker was made.

makeClockChecker ::
  (MonadIO m, Integral a, Show e, Exception e) =>
  e -- ^ exception to throw
  -> a -- ^ @maxClockSeconds@
  -> IO (m ())

makeClockChecker exc maxClockSeconds =
  assert (maxClockSeconds > 0) $ do
  startTime <- getCurrentTime
  let stopTime = addUTCTime (fromIntegral maxClockSeconds) startTime
  stopTimeInLocalZone <- utcToLocalZonedTime stopTime
  msgInfo ("At time " ++ formatTime defaultTimeLocale "%c" stopTimeInLocalZone
           ++ " raise " ++ show exc)
  return $ liftIO $ do
    now <- getCurrentTime
    when (now >= stopTime) (throwIO exc)


-- | Encode an object as YAML to the given file path, using
-- 'assertDoNotOverwrite' to avoid overwriting an existing file.
yamlEncodeFileSafe :: ToJSON a => FilePath -> a -> IO ()
yamlEncodeFileSafe path obj = do
  assertDoNotOverwrite path
  let bytes = encodePretty myYAMLPrettyConf obj
  ByteString.writeFile path bytes


-- | This does exactly the same thing as 'withPopCache' but by
-- accepting a 'UsefulGenomeSpec' and a 'UsefulProblem' as dummy
-- arguments, the type checking works a lot better, no silly
-- ambiguities.
withPopCacheX ::
  (UsefulGenomeSpec gs,
   UsefulProblem prob) =>
  gs -> prob -> PopMonad (Genome gs) (Rating prob) c -> DBMonad TableName c

withPopCacheX _genomeSpec _problem = withPopCache


-- | Global lock path
globalLockFileName :: FilePath
globalLockFileName = "GlobalLock.txt"

-- | Exception thrown when the global lock can't be acquired.
data LockException = LockException String
                   deriving (Show, Eq, Data, Generic, Typeable)

instance Exception LockException

-- | Acquire the global lock and perform the given action, or throw a 'LockException'
withGlobalLock :: FilePath -> IO a -> IO a
withGlobalLock sampleDir axn = do
  let lockPath = sampleDir </> globalLockFileName
  mRes <- withLock lockPath axn
  case mRes of
    Just result ->
      return result
    Nothing -> do
      mLockHolder <- readLockHolder lockPath
      case mLockHolder of
        Just (hostname, pid) ->
          throwIO (LockException $
                  printf "Could not acquire global mutex lock %s: held by process %s on %s"
                  lockPath (show pid) hostname)
        Nothing ->
          throwIO (LockException $
                  printf "Could not acquire global mutex lock %s: no information available on lock holder"
                  lockPath)

-- | How long to go between snapshots and how many snapshots to keep.
-- If there are @snapshotMaxCount@ or more, the oldest (highest numbered) is deleted first.
data SnapshotSpec = SnapshotSpec {
  snapshotSeconds :: Int, -- ^ how many seconds between snapshots
  snapshotMaxCount :: Int -- ^ @snapshotMaxCount@
  }
  deriving (Read, Show, Eq, Ord, Data, Generic, Typeable)

-- | Make an action that checks for whether it's time to make a snapshot
-- and raises 'SnapshotSecondsElapsed' when it is.
makeSnapshotChecker ::
  MonadIO m =>
  SnapshotSpec
  -> IO (m ())
makeSnapshotChecker SnapshotSpec{..} = do
  makeClockChecker SnapshotSecondsElapsed snapshotSeconds

-- | Make a snapshot of all the important files in this sample.
-- They are placed in a subdirectory.
makeSnapshot :: DatabaseSpec -> SnapshotSpec -> IO ()
makeSnapshot DatabaseSpec{..} SnapshotSpec{..} = do
  msgInfo "Time to make a snapshot"
  cDir <- getCurrentDirectory
  msgDebug (printf "makeSnapshot: CWD = %s" cDir)
  let topDir = environmentDirectory
  let snapshotDir = topDir </> "Snapshot"
  -- Move existing snapshots up a notch
  let lastSnapshotPath = snapshotDir </> show (snapshotMaxCount - 1)
  lastSnapshotExists <- doesDirectoryExist lastSnapshotPath
  when lastSnapshotExists $ do
    msgDebug (printf "makeSnapshot: Removing %s" lastSnapshotPath)
    removeDirectoryRecursive lastSnapshotPath
  forM_ [snapshotMaxCount - 2, snapshotMaxCount - 3 .. 0] $ \j -> do
    let snapshotFromPath = snapshotDir </> show j
    let snapshotToPath = snapshotDir </> show (j + 1)
    snapshotExists <- doesDirectoryExist snapshotFromPath
    when snapshotExists $ do
      msgDebug (printf "makeSnapshot: Renaming %s to %s" snapshotFromPath snapshotToPath)
      renameDirectory snapshotFromPath snapshotToPath
  -- Copy key files to Snapshot/0
  let newSnapshotPath = snapshotDir </> "0"
  msgDebug (printf "makeSnapshot: Making snapshot in %s" newSnapshotPath)
  createDirectoryIfMissing True newSnapshotPath
  forM_ ["InitialState", "CurrentState", "Landmark", "Archive"] $ \keyFile -> do
    let fromPath = topDir </> keyFile
    let toPath = newSnapshotPath </> keyFile
    msgDebug (printf "Copying %s to %s" fromPath toPath)
    copyFileWithMetadata fromPath toPath
  msgInfo "Snapshot complete"
