{-# LANGUAGE
RecordWildCards,
NamedFieldPuns,
FlexibleContexts,
DeriveDataTypeable,
DeriveGeneric,
TemplateHaskell,
MultiParamTypeClasses
#-}

{-|

Module: Digiostelium.Query

Description : Class for generating GraphML

Copyright   : © 2019 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC)

<http://graphml.graphdrawing.org/ GraphML> is a file format read
by the <http://www.yworks.com YED> graph editor.  As well as
basic GraphML, the functions in this module incorporate styling
and other information particular to YED.
-}


module Digiostelium.GraphML where

import Text.XML.Light

infix 3 <:>, +>

-- | For objects that can be expressed as GraphML.
-- The @opt@ object is a way to specify options.
class ToGraphML opt t where

  toGraphML :: opt -> t -> Element



-- | Build an attribute setting

(+>) :: String -> String -> Attr
key +> value = Attr (unqual key) value


-- | Build a qualified name

(<:>) :: String -> String -> QName
pre <:> name =
  blank_name {qPrefix = Just pre, qName = name}


-- | Build a node from an id, a label to display in the node,
-- a shape string, and a color

prettyNode :: Node t => String -> t -> String -> String -> Element
prettyNode nodeId label shapeType color =
  (unode "node"
   (["id" +> nodeId],
    nodeGraphicsData $ shapeNode label shapeType color))


-- | Build an element containing node data as expected by YED.

nodeGraphicsData :: Node ([Attr], t) => t -> Element
nodeGraphicsData dataContent =
  unode "data" (["key" +> "NG"], dataContent)


-- | Build a shape node from a label, a shape string, and a fill
-- color string

shapeNode :: Node t => t -> String -> String -> Element
shapeNode label shapeType color =
  (node ("y" <:> "ShapeNode")
   [node ("y" <:> "NodeLabel") label,
    node ("y" <:> "Fill")
    ["color" +> color,
     "transparent" +> "false"],
    node ("y" <:> "BorderStyle")
    ["color" +> "#000000"],
    node ("y" <:> "Shape")
    ["type" +> shapeType]])


-- | Build an edge given the edge ID, source ID, the targetID, the edge
-- label, and whether the link should be styled to indicate
-- positive or negative action on the target

prettyEdge ::
  (Node t,
   Node ([Attr], t))
  =>
  String -> String -> String -> Maybe t -> Bool -> Element
prettyEdge edgeId sourceId targetId mLabel positive =
  (unode "edge"
   (["id" +> edgeId,
     "source" +> sourceId,
     "target" +> targetId],
    edgeGraphicsData $ polyLineEdge mLabel positive))

-- | Build an element containing edge data as expected by YED.
edgeGraphicsData :: Node ([Attr], t) => t -> Element
edgeGraphicsData dataContent =
  unode "data" (["key" +> "EG"], dataContent)


-- | Build a polyline edge from a label and whether it should be
-- styled to indicate positive or negative action on its
-- target.

polyLineEdge ::
  (Node t,
   Node ([Attr], t))
  =>
  Maybe t -> Bool -> Element
polyLineEdge mLabel positive =
  (node ("y" <:> "PolyLineEdge")
   ([node ("y" <:> "LineStyle")
     (if positive
      then ["color" +> "#000000", "type" +> "line", "width" +> "1.0"]
      else ["color" +> "#000000", "type" +> "dashed", "width" +> "1.0"]),
     node ("y" <:> "Arrows")
     (if positive
      then ["source" +> "none", "target" +> "standard"]
      else ["source" +> "none", "target" +> "white_delta"])] -- "t_shape"
    ++
    case mLabel of
      Nothing -> []
      Just label ->
        [node ("y" <:> "EdgeLabel")
         (["backgroundColor" +> "#ffffff"],
          label)]
   ))


-- | Make a directed graph element from an ID and content
digraphElement :: Node ([Attr], t) => String -> t -> Element
digraphElement graphId nodesAndEdgesContent =
  unode "graph"
  (["id" +> graphId,
    "edgedefault" +> "directed"],
   nodesAndEdgesContent)


-- | Wrap the given data in a GraphML document.  Yed requires the
-- @graphml@ wrapping element, complete with the schema
-- information and @y:@ namespace.
-- Furthermore, it adds global @key@ elements under @graphml@ to specify that
--
-- * @<data key="NG" ...@ within @node@ means the element includes yed-specific node styling
--
-- * @<data key="EG" ...@ within @edge@ means the element includes yed-specific edge styling

scaffold :: Element -> Element
scaffold documentContent =
  (unode "graphml"
   (["xmlns" +> "http://graphml.graphdrawing.org/xmlns",
     "xmlns:xsi" +> "http://www.w3.org/2001/XMLSchema-instance",
     "xsi:schemaLocation" +> "http://graphml.graphdrawing.org/xmlns http://www.yworks.com/xml/schema/graphml/1.1/ygraphml.xsd",
     "xmlns:y" +> "http://www.yworks.com/xml/graphml",
     "xmlns:yed" +> "http://www.yworks.com/xml/yed/3"],
    -- These elements declare YED-specific attributes
    [unode "key"
     ["for" +> "node",
      "id" +> "NG",
      "yfiles.type" +> "nodegraphics"],
     unode "key"
     ["for" +> "edge",
      "id" +> "EG",
      "yfiles.type" +> "edgegraphics"],
     documentContent]))
