{-# LANGUAGE TypeFamilies, FlexibleContexts, RecordWildCards, NamedFieldPuns #-}

{-|

Module: Digiostelium.InputOutputProblem

Description : Abstractions for representing a problem where everything is based on input to and output from an 'Arena'.

Copyright   : © 2019 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

-}

module Digiostelium.InputOutputProblem where

import Data.Map (Map)
import qualified Data.Map as Map
import Data.Maybe (maybeToList)
import Data.Sequence (Seq)

import BaselMachine.Arena
-- import BaselMachine.Signal

import Digiostelium.Problem



-- | A 'InputOutputProblem' is a problem that supplies a stream of
-- input to an 'Arena', and the ratings of agents depend only on the
-- broadcast outputs and resources used during the computation.

class InputOutputProblem problem where

  type Input problem

  type Output problem

  problemInputToArenaInput ::
    problem
    -> Input problem
    -> (Int, [[BroadcastState (Location problem) (Signal problem)]])
    -- ^ maximum time steps, and the list of inputs

  arenaOutputToProblemOutput ::
    problem
    -> Seq [BroadcastState (Location problem) (Signal problem)]
    -> Output problem


  -- | List all inputs that are used for rating purposes.
  -- Default is to call 'inputsForTesting' and fetch the value
  -- associated to the key "Rating".
  inputsForRating ::
    problem -> [Input problem]

  inputsForRating problem =
    getTestSet problem "Rating"

  -- | 'Map' from names to sets up inputs used for testing purposes.
  -- Default a 'Map' with the association ("Rating", rating inputs).
  inputsForTesting ::
    problem -> Map String [Input problem]

  inputsForTesting problem =
    Map.singleton "Rating" (inputsForRating problem)


-- | Fetch a test set for a problem.  Return '[]' if the test set
-- key is not found.

getTestSet ::
  (InputOutputProblem problem) =>
  problem
  -> String -- ^ test set key
  -> [Input problem]

getTestSet problem testSetKey =
  concat $ maybeToList $ Map.lookup testSetKey (inputsForTesting problem)
