{-# LANGUAGE RecordWildCards, NamedFieldPuns, DeriveDataTypeable, TemplateHaskell, Arrows, DeriveGeneric, StandaloneDeriving #-}

{-|
Module      : Digiostelium.Job

Description : Parsing the command line

Copyright   : © 2019 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

-}

module Digiostelium.Job where

import GHC.Generics

import Control.Applicative
import Control.Exception
import Control.Monad

import Data.Aeson
import Data.Aeson.TH
import Data.Array
-- import qualified Data.Attoparsec.ByteString as AP
-- import qualified Data.ByteString.UTF8 as BSU
import Data.Data
import Data.Default
import Data.Maybe
import Data.Semigroup ( (<>) )
--import Data.Typeable
import Data.Version
--import Data.Yaml
import Options.Applicative
import Options.Applicative.Arrows
import System.Log.Logger
import Text.Regex.Base
import Text.Regex.PCRE.String (Regex)

-- Generated by cabal
import qualified Paths_Digiostelium as P

import Digiostelium.AesonUtil
import Digiostelium.ArchivePredicate
import Digiostelium.BasicReport
import Digiostelium.PairCopyProblemReport
import Digiostelium.PredicateTools
import Digiostelium.ProblemInventory
import Digiostelium.Query
import Digiostelium.Report
import Digiostelium.ReportException
import Digiostelium.SingletonProblem

-- These got left out of old versions of System.Log.Logger
deriving instance Typeable Priority
-- deriving instance Data Priority

data CommonOpts =
  CommonOpts {
    logLevel :: Priority
    }
  deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)

data Target =
  Target {
    maxGenerations :: Maybe Integer,
    landmarkBorn :: Maybe Value,
    landmarkDied :: Maybe Value
    }
  deriving (Show, Data, Typeable, Generic)

$(deriveJSON myJSONOpts ''Target)


instance Default Target where
  def = Target{..}
    where
      maxGenerations = Nothing
      landmarkBorn = Nothing
      landmarkDied = Nothing


data Config =
  Config {
    problemName :: String,
    genomeArchitecture :: String,
    archive :: CompoundMPredicate ArchivePredicate,
    target :: Target
    } deriving (Show, Data, Typeable, Generic)

$(deriveJSON myJSONOpts ''Config)


data Args = Args CommonOpts Command
  deriving (Show, Data, Typeable, Generic)

data Command =
  List
  | Template {
    sampleDir :: FilePath, -- ^ diretory where sample will be kept
    problemTemplateName :: String, -- ^ name of the problem to solve
    genomeArchitectureTemplateName :: String -- ^ name of the genome architecture
    }
  | Init {
    sampleDir :: FilePath, -- ^ diretory where sample will be kept
    importFrom :: FilePath,
    identical :: Bool -- ^ whether all initial genomes should be the same
    }
  | Reset {
    sampleDir :: FilePath -- ^ diretory where sample will be kept
    }
  | Run {
    sampleDir :: FilePath, -- ^ diretory where sample will be kept
    maxSeconds :: Maybe Integer -- ^ run for no more than this many seconds
    }
  | Continue {
    sampleDir :: FilePath, -- ^ diretory where sample will be kept
    moreGenerations :: Integer -- ^ run this many more generations
    }
  | Query {
    sampleDir :: FilePath, -- ^ diretory where sample are kept
    select :: SelectAgents, -- ^ which agents
    reportSpec :: PossibleReportSpec, -- ^ what information
    outputFormat :: OutputFormat, -- ^ which format
    outputDestination :: OutputDestination -- ^ where to write it
    }
  deriving (Show, Data, Typeable, Generic)

$(deriveJSON myJSONOpts ''Command)


jobOpts :: Parser Args
jobOpts = runA $ proc () -> do
  cmds <- ((asA . hsubparser)
           (command "list"
            (info listOpts
             (progDesc "List available problems, genome architectures, etc."))
            <>
            command "template"
            (info templateOpts
             (progDesc "Create configuration files from templates"))
            <>
            command "init"
            (info initOpts
             (progDesc "Initialize a sample run"))
            <>
            command "reset"
            (info resetOpts
             (progDesc "Reset (reinitialize) a sample run"))
            <>
            command "run"
            (info runOpts
             (progDesc "Run a sample to completion target in Config.yaml"))
            <>
            command "continue"
            (info continueOpts
             (progDesc "Continue a sample for more generations"))
            <>
            command "query"
            (info queryOpts
             (progDesc "Query a sample"))
           )) -< ()
  opts <- asA commonOpts -< ()
  A helper -< Args opts cmds


commonOpts :: Parser CommonOpts
commonOpts =
  (CommonOpts <$>
   (option auto
    (long "log-level"
     <> metavar "LEVEL"
     <> help "Append log messages of this level (ERROR, WARNING, NOTICE, INFO, DEBUG) to an appropriate file. Default is NOTICE."
     <> value NOTICE))
  )

listOpts :: Parser Command
listOpts = runA $ proc () -> do
  returnA -< List

templateOpts :: Parser Command
templateOpts = runA $ proc () -> do
  sampleDir <-
    (asA $ argument str
     (metavar "SAMPLE_DIR"
      <> help "Data directory holding sample")
    ) -< ()
  genomeArchitectureTemplateName <-
    (asA $ argument str
     (metavar "ARCHITECTURE"
      <> help "Genome architecture")) -< ()
  problemTemplateName <-
    (asA $ argument str
     (metavar "PROBLEM"
      <> help "Problem name")) -< ()
  returnA -< Template{..}

initOpts :: Parser Command
initOpts = runA $ proc () -> do
  sampleDir <-
    (asA $ argument str
     (metavar "SAMPLE_DIR"
      <> help "Data directory holding sample")
    ) -< ()
  importFrom <-
    (asA $ option str
     (long "import"
      <> metavar "GENOME_FILE"
      <> value ""
      <> help "Import genomes exported from another population")
     ) -< ()
  identical <-
    (asA $ switch
     (long "identical"
      <> help "All genomes are created identical")) -< ()

  returnA -< Init{..}


resetOpts :: Parser Command
resetOpts = runA $ proc () -> do
  sampleDir <-
    (asA $ argument str
     (metavar "SAMPLE_DIR"
      <> help "Data directory holding sample")
    ) -< ()
  returnA -< Reset{..}

runOpts :: Parser Command
runOpts = runA $ proc () -> do
  sampleDir <-
    (asA $ argument str
     (metavar "SAMPLE_DIR"
      <> help "Data directory holding sample")
    ) -< ()
  maxSeconds <-
    (asA $ optional $ option auto
      (long "max-seconds"
       <> metavar "SECONDS"
       <> help "Run for no more than this many seconds")
    ) -< ()
  returnA -< Run{..}


continueOpts :: Parser Command
continueOpts = runA $ proc () -> do
  sampleDir <-
    (asA $ argument str
     (metavar "SAMPLE_DIR"
      <> help "Data directory holding sample")
    ) -< ()
  moreGenerations <-
    (asA $ argument auto
      (metavar "GENERATIONS"
       <> help "Run this many more generations")
    ) -< ()
  returnA -< Continue{..}

queryOpts :: Parser Command
queryOpts = runA $ proc () -> do
  sampleDir <-
    (asA $ argument str
     (metavar "SAMPLE_DIR"
      <> help "Data directory holding sample")
    ) -< ()
  -- select agents
  mlandmarks <-
    (asA $ flag Nothing (Just Landmarks)
     (long "landmarks"
      <> help "Select landmark agents (default)")) -< ()
  mbyID <-
    (asA $ optional $ strOption
     (long "by-id"
      <> metavar "IDs"
      <> help "Select agents with these IDs (can be comma or space separated)")) -< ()
  mtop <-
    (asA $ flag Nothing (Just TopRated)
     (long "top-rated"
      <> help "Select top rated living agents")) -< ()
  mratedAtLeastStr <-
    (asA $ optional $ strOption
     (long "rated-at-least"
      <> help "Select living agents with at least this rating (in JSON format)")) -< ()
  mliving <-
    (asA $ flag Nothing (Just Living)
     (long "living"
      <> help "Select all living agents")) -< ()
  mnone <-
    (asA $ flag Nothing (Just None)
     (long "none"
      <> help "Select no agents")) -< ()

  -- report sepcification
  msummary <-
    (asA $ flag Nothing (Just $ Basic Summary)
     (long "summary"
      <> help "Request a summary (default)")) -< ()
  magent <-
    (asA $ flag Nothing (Just $ Basic JustAgent)
     (long "agent"
      <> help "Request the whole agent structure")) -< ()
  mexport <-
    (asA $ flag Nothing (Just $ Basic Export)
     (long "export"
      <> help "Request an export record")) -< ()
  mprogram <-
    (asA $ flag Nothing (Just $ Basic Program)
     (long "program"
      <> help "Request information about the agent's program")) -< ()
  mgenome <-
    (asA $ flag Nothing (Just $ Basic GenomeDetails)
     (long "genome"
      <> help "Request information about the agent's genome")) -< ()

  mratingInputs <-
    (asA $ flag Nothing (Just $ Singleton RatingInputs)
     (long "rating-inputs"
      <> help "Request summary of inputs used for rating")) -< ()
  mtestSetStr <-
    (asA $ optional $ strOption
     (long "test-set"
      <> metavar "KEY"
      <> help "Request summary of inputs from the test set named by the given key")) -< ()
  mefficacious <-
    (asA $ flag Nothing (Just $ Singleton Efficacious)
     (long "efficacious"
      <> help "Request efficacious analysis of the agent's genome")) -< ()
  mfull <-
    (asA $ flag Nothing (Just $ Singleton FullDetails)
     (long "full"
      <> help "Request full analysis of the agent's genome")) -< ()
  mreactionNetwork <-
    (asA $ flag Nothing (Just $ Singleton ReactionNetwork)
     (long "reaction-network"
      <> help "Request GraphML showing efficacious reactions; use with --path-template")) -< ()

  mpairRatingInputs <-
    (asA $ flag Nothing (Just $ PairCopy PairRatingInputs)
     (long "pair-rating-inputs"
      <> help "Request a summary of inputs used for rating in which all pairs of selected agents interact")) -< ()
  mpairDetailsStr <-
    (asA $ optional $ strOption
     (long "pair-details"
      <> metavar "ID"
      <> help "Request details for agent with that ID when paired against other selected agents")) -< ()
  mpairReactionNetworkStr <-
    (asA $ optional $ strOption
     (long "pair-reaction-network"
      <> metavar "ID"
      <> help "Request GraphML showing efficacious reactions for agent with that ID when paired against other selected agents")) -< ()

  -- output formats
  mtext <-
    (asA $ flag Nothing (Just PlainTextOutput)
     (long "text"
      <> help "Output as plain text (default)")) -< ()
  mmarkdown <-
    (asA $ flag Nothing (Just MarkdownOutput)
     (long "markdown"
      <> help "Output as Markdown")) -< ()
  myaml <-
    (asA $ flag Nothing (Just YAMLOutput)
     (long "yaml"
      <> help "Output as YAML")) -< ()
  mjson <-
    (asA $ flag Nothing (Just JSONOutput)
     (long "json"
      <> help "Output as JSON")) -< ()
  mgraphml <-
    (asA $ flag Nothing (Just GraphMLOutput)
     (long "graphml"
      <> help "Output as GraphML")) -< ()
  overwrite <-
    (asA $ switch
     (long "overwrite"
      <> help "Overwrite existing files")) -< ()

  -- output destinations
  mtoStdout  <-
    (asA $ flag Nothing (Just AllToStdout)
     (long "stdout"
      <> help "Send output to stdout (default)")) -< ()
  mpathTemplate <-
    (asA $ optional $ strOption
     (long "path-template"
      <> metavar "TEMPLATE"
      <> help "Send output for each agent to its own file using the printf TEMPLATE for the file path, which should include a %d for the agent's id")
    ) -< ()
  -- Convert string options to objects
  let mOnePerFile = ToFilePerAgentId overwrite <$> mpathTemplate
      mIDs = IDs . parseCommaSeparatedNumbers <$> mbyID
      mratedAtLeast = do
        s <- mratedAtLeastStr
        case stringToJSONValue s of
          Left msg ->
            throw (JSONError msg)
          Right v ->
            return (RatedAtLeast v)
      -- These are :: PossibleReportSpec
      mtestSet = Singleton . TestSet <$> mtestSetStr
      mpairDetails = PairCopy . PairDetails . read <$> mpairDetailsStr
      mpairReactionNetwork = PairCopy . PairReactionNetwork . read <$> mpairReactionNetworkStr
  returnA -< Query{
    select = mpick Landmarks [mIDs, mlandmarks, mtop, mratedAtLeast, mliving, mnone],
    reportSpec =
        (mpick (Basic Summary)
         [magent, msummary, mexport, mprogram, mgenome,
           mratingInputs, mefficacious, mfull, mreactionNetwork,
           mpairRatingInputs, mtestSet, mpairDetails, mpairReactionNetwork]),
    outputFormat = mpick PlainTextOutput [mtext, mmarkdown, mjson, myaml, mgraphml],
    outputDestination = mpick AllToStdout [mtoStdout, mOnePerFile],
    ..}

mpick :: t -> [Maybe t] -> t
mpick x ms =
  fromMaybe x (msum ms)

jobParser :: String -> ParserInfo Args
jobParser extra =
  (info jobOpts
   (progDesc
    (concat ["Artificial life simulation engine, version ",
             showVersion P.version,
             "; give --help for more help\n",
             extra])))


--- some odds and ends ---

parseCommaSeparatedNumbers :: String -> [Integer]
parseCommaSeparatedNumbers s =
  let
    intRX :: Regex
    intRX = makeRegex "[0-9]+"
    ms = matchAllText intRX s
    intStrs = [fst (a ! 0) | a <- ms]
  in
   map read intStrs
