{-# LANGUAGE RecordWildCards, NamedFieldPuns, ScopedTypeVariables, FlexibleContexts, FlexibleInstances, ParallelListComp, TypeFamilies, DeriveDataTypeable, OverloadedStrings, TemplateHaskell, DeriveGeneric, MultiParamTypeClasses #-}

{-|

Module: Digiostelium.NeuronProblem

Description: Foundations for single neuron problems

Copyright: © 2019 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

Base definitions for a problem in which there is one neuron-like colony with one cell that receives role input into one pattern.
It receives a stream of input spikes (on or off at any given time step) into a configurable channel, and produces a stream of output spikes.

-}

module Digiostelium.NeuronProblem (
  NeuronProblem(..),
  ToBools(..),
  neuronProblemInputToArenaInput,
  arenaOutputToNeuronProblemOutput,
  boolsToDrawing
  )
where

import GHC.Generics

import Control.DeepSeq

import Data.Aeson.TH
import Data.Binary
import Data.Data
import Data.Default
import Data.Foldable (toList)
import Data.Sequence (Seq)
import Data.Typeable

import BaselMachine.Arena

import Digiostelium.AesonUtil
import Digiostelium.GeneralName


data NeuronProblem = NeuronProblem {
  inputPattern :: Int,
  -- ^ the pattern that gets incremented when the neuron receives
  -- input
  outputPattern :: Int,
  -- ^ when the activation of this pattern goes high enough, the
  -- neuron spikes
  outputThreshold :: Int,
  -- ^ the threshold that the output pattern must reach to spike
  outputCost :: Int
  -- ^ the activation to be subtracted when the neuron spikes,
  -- should be /positive/
  } deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)

instance NFData NeuronProblem
instance Binary NeuronProblem
$(deriveJSON myJSONOpts ''NeuronProblem)

instance Default NeuronProblem where

  def = NeuronProblem {
    inputPattern = 3, -- encodes as 010000g
    outputPattern = 41, -- encodes as 101111g
    outputThreshold = 2,
    outputCost = 2
    }

instance GeneralName NeuronProblem where
  generalName _ = "Neuron"

-- It's not possible to write instance InputOutputProblem NeruonProblem because the inputs are left to various specializations.
-- Instead, we can write some general helper functions:


-- | Helper for implementing instance 'InputOutputProblem'
neuronProblemInputToArenaInput ::
  NeuronProblem
  -> [Bool]
  -> (Int, [[BroadcastState NullLocation NullSignal]])

neuronProblemInputToArenaInput NeuronProblem{..} ns =
    (numSteps, map convert ns)
    where
      numSteps = length ns

      -- At each time step there's one broadcast to the one
      -- colony, and it lasts for one time step
      convert n =
        if n
        then
          [BroadcastState {
              broadcastSignal = NullSignal,
              broadcastMaximumAge = Just 1,
              broadcastFromColonyIx = Nothing,
              broadcastOrigin = NullLocation,
              broadcastAge = 0 }]
        else
          []

-- | Helper for implementing instance 'InputOutputProblem'

arenaOutputToNeuronProblemOutput ::
  NeuronProblem
  -> Seq [BroadcastState location NullSignal]
  -> [Bool]

arenaOutputToNeuronProblemOutput NeuronProblem{..} colonyOutputSeq =
    [ not (null broadcasts)
    | broadcasts <- toList colonyOutputSeq ]


-- | Things that can be represented as a sequence of 'Bool', as
-- in on and off
class ToBools t where
  toBools :: t -> [Bool]


-- | This is a little weird, but it works.
instance ToBools Bool where
  toBools b = [b]

-- | Run lists of Bools together
instance ToBools elt => ToBools [elt] where
  toBools = concatMap toBools


{-



-- === UNCONVERTED === --

singleNeuronColonySpec NeuronProblem{..} =
  ColonySpec{
    inputSpecs = IM.fromList [inputSpec],
    emitterSpecs = [emitterSpec],
    ..}
  where
    numCells = 1
    cellNames = IM.fromList [(0, "Neuron")]
    synapseSpecs = []

    inputSpec =
      (1,     -- message is 1
       ("In", -- named "In"
        0, -- feeds to cell 0
        ReceptorSpec{
          targetPattern = inputPattern,
          targetDelta = inputActivation}))

    emitterSpec =
      ("Out", -- named "Out"
       0, -- reads cell 0
       EmitterSpec{
         sourcePattern = outputPattern,
         sourceThreshold = outputThreshold,
         activationDelta = -outputCost },
       1) -- message is 1

-}

-- | Convert a list of 'Bool's to one character for 'False' and
-- another for 'True'
boolsToDisplay ::
  Char -- ^ character for 'True'
  -> Char -- ^ character for 'False'
  -> [Bool] -> String

boolsToDisplay on off xs =
  [if x then on else off | x <- xs]

-- | Convert a list of 'Bool's to "0" for 'False' and "1" for 'True'
boolsTo01s :: [Bool] -> String
boolsTo01s =
  boolsToDisplay '1' '0'

-- | Convert a list of 'Bool's to ╌ (0x254c) for 'False' and ┸ (0x2538) for 'True'
boolsToDrawing :: [Bool] -> String
boolsToDrawing =
  boolsToDisplay '┸' '╌'


-- === Attic  === --

{-

-- | Convert a list of 'Bool' indicating when the input is high
-- to the correct input format for a single neuron 'Colony'

boolsToColonyInputStream ::
  NeuronProblem
  -> [Bool]
  -> [[Int]]

boolsToColonyInputStream NeuronProblem {..} input =
  [if x
   then [1]
   else []
  | x <- input]


-- | Convert the output from a single neuron 'Colony' to a list
-- of 'Bool's for when the output is high.  The absurdly general
-- type is in case I later decide to change the container type
-- (list) from 'Sequence'

colonyOutputsToBools ::
  Foldable list =>
  NeuronProblem
  -> list [emission] -> [Bool]

colonyOutputsToBools NeuronProblem{..} colonyOutputSeq =
  [not (null emitted)
  | emitted <- F.toList colonyOutputSeq ]
-}


{-
data NeuronProblemTraceSummary = NeuronProblemTraceSummary {
  npts_in :: String,
  -- ^ input word as a string of 0s and 1s
  npts_out :: String
  -- ^ output word as a string of 0s and 1s
  } deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)


$(deriveJSON myJSONOpts{fieldLabelModifier = drop 5}
  ''NeuronProblemTraceSummary)
-}
