{- -- OPTIONS_GHC -fno-ignore-asserts -}

{-# LANGUAGE
RecordWildCards,
NamedFieldPuns,
ScopedTypeVariables,
FlexibleContexts,
MultiParamTypeClasses,
FlexibleInstances,
ParallelListComp,
TypeFamilies,
DeriveDataTypeable,
OverloadedStrings,
TemplateHaskell,
DeriveGeneric,
UndecidableInstances,
FunctionalDependencies
#-}

{-|

Module: Digiostelium.PairCopyProblem

Description : Like CopyProblem but with two separate agents.

Copyright   : © 2019 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

-}

module Digiostelium.PairCopyProblem where

import GHC.Generics

import Control.DeepSeq
import Control.Exception
import Control.Lens hiding ( (??) )

import Data.Aeson
import Data.Aeson.TH
import Data.Binary
import Data.Data
import Data.Default
import qualified Data.IntMap as IntMap
import qualified Data.Map as Map
import Data.Semigroup (Sum(..))


import Numeric.LinearAlgebra

import BaselMachine.Arena
import BaselMachine.Compiler (ProgramInformation, ReactionProgram)
import BaselMachine.Instruction
import WGM.ParUtil

import qualified Digiostelium.Agent as Agent
import Digiostelium.AesonUtil
import Digiostelium.CopyProblemBase
import Digiostelium.GeneralName
import Digiostelium.Genome
import Digiostelium.InputOutputProblem
import Digiostelium.PandocUtil
import Digiostelium.Problem


data PairLocation = Sender | Receiver
  deriving (Enum, Eq, Ord, Read, Show, Data, Generic, Typeable)

instance NFData PairLocation
instance Binary PairLocation
$(deriveJSON defaultOptions ''PairLocation)

instance ToJSONKey PairLocation
instance FromJSONKey PairLocation


data PairCopyProblem = PairCopyProblem {
  wordSize :: Int,
  broadcastWidth :: Int,
  minSteps :: Int,
  maxSteps :: Int,
  correctnessFactor :: Double,
  inputWordIndexToPattern :: Int,
  -- ^ add this to an index into the input word to get the
  -- pattern in the sender that gets incremented
  inputActivation :: Int,
  -- ^ amount added to an input pattern in the sender when the
  -- corresponding bit in the input word is on
  broadcastInputIndexToPattern :: Int,
  -- ^ add this to a bit index into the synaptic word to get the
  -- pattern in the sender that gets read
  broadcastOutputIndexToPattern :: Int,
  -- ^ add this to a bit index into the synaptic word to get the
  -- pattern in the receiver that gets incremented
  broadcastThreshold :: Int,
  -- ^ the broadcast fires when the activation of a broadcast pattern
  -- in the sender reaches this threshold
  broadcastCost :: Int,
  -- ^ the activation subtracted when the broadcast spikes
  broadcastActivation :: Int,
  -- ^ the activation added to a pattern in the receiver when the
  -- broadcast spikes
  outputWordIndexToPattern :: Int,
  -- ^ add this to an index into the output word to get the
  -- pattern in the receiver that gets read
  outputThreshold :: Int,
  -- ^ an output bit is set on when the activation of an output
  -- pattern in the receiver reaches this threshold
  outputCost :: Int,
  -- ^ the activation subtracted when an output bit is set
  beaconPattern :: Int,
  -- ^ this pattern is read for the beacon signal
  beaconThreshold :: Int,
  -- ^ in the receiver cell, if @A['beaconPattern'] ≥
  -- 'beaconThreshold'@ then the beacon signal is emitted
  beaconCost :: Int,
  -- ^ amount to subtract from @A['beaconPattern']@ in the
  -- receiver cell when the beacon signal is emitted
  beaconPoints :: Double,
  -- ^ points awarded for emitting the beacon signal
  minReactions :: Double,
  -- ^ no cost for averaging up to this many reactions per time step
  reactionCostFactor :: Double,
  -- ^ an energy cost; multiply this by ((the number of reactions
  -- used divided by the number of time steps) less
  -- 'minReactions') and subtract that from the total score
  minTotalDeltas :: Double,
  -- ^ no cost for averaging up to this total of reaction deltas
  -- per time step
  deltaCostFactor :: Double,
  -- ^ an energy cost; multiply this by ((the total of all
  -- reaction deltas applied divided by the number of time steps)
  -- less 'minTotalDeltas') and subtract that from the total score
  roleActivation :: Int,
  -- ^ how much activation is added to @A[1]@ in the sender and
  -- @A[2]@ in the receiver on each time step to indicate their
  -- roles.
  ratingSystem :: PairRatingSystem
  } deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)

instance NFData PairCopyProblem
instance Binary PairCopyProblem


-- | In general, to rate an 'Egg', it interacts with all other
-- 'Egg's and living 'Agent's as both sender ('Here' location)
-- and receiver ('There' location).  Its rating is an aggregation
-- of all those ratings.

data PairRatingSystem =
  CompareAll
  -- ^ Average of all scores, equally weighted.  Note that since
  -- many mutations lead to poorly performing, the population
  -- tends to stagnate if all interactions are equally weighted.
  -- Genomes that could perform well can earn low ratings because
  -- there happen to be a lot of deleterious mutations in their
  -- generation.

  | CompareQuantile {
      compareQuantile :: Double,
      selfWeight :: Double,
      -- ^ weight for score when interacting with self
      hereWeight :: Double,
      -- ^ weight for 'Here' score
      thereWeight :: Double
      -- ^ weight for 'There' score
      }
    -- ^ Score is
    --
    -- > (selfWeight * selfScore + hereWeight * hereScore + thereWeight * thereScore)
    -- >  /
    -- >  (selfWeight + hereWeight + thereWeight))
    --
    -- The 'Here' score is the average of ratings earned by an
    -- 'Egg' when interacting in the 'Here' location (sender)
    -- with other 'Egg's or 'Agent's, including itself.  Ditto
    -- for the 'There' score.  In forming those averages, the
    -- scores of the appropriate data set are sorted, and only
    -- the scores above the 'compareQuantile' are considered.
    --
    -- The @selfScore@ component is so that extra points are
    -- earned by interacting well with self.

  deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)

instance NFData PairRatingSystem
instance Binary PairRatingSystem

instance Default PairRatingSystem where
  def = CompareQuantile {
    compareQuantile = 0.5,
    selfWeight = 1.0,
    hereWeight = 1.0,
    thereWeight = 1.0
    }

-- | The default is the @Copy21@ problem with no beacon cost and
-- no energy costs.
instance Default PairCopyProblem where
  def = PairCopyProblem {
    wordSize = 2,
    broadcastWidth = 1,
    minSteps = 20,
    maxSteps = 100,
    correctnessFactor = 100,
    inputWordIndexToPattern = 8,
    inputActivation = 1,
    broadcastInputIndexToPattern = 3,
    broadcastOutputIndexToPattern = 4,
    broadcastThreshold = 10,
    broadcastCost = 0,
    broadcastActivation = 1,
    outputWordIndexToPattern = 12,
    outputThreshold = 10,
    outputCost = 0,
    beaconPattern = 31,
    beaconThreshold = 10,
    beaconCost = 0,
    beaconPoints = 0,
    minReactions = 0,
    reactionCostFactor = 0,
    minTotalDeltas = 0,
    deltaCostFactor = 0,
    roleActivation = 1,
    ratingSystem = def
    }


instance GeneralName PairCopyProblem where
  generalName _ = "PairCopy"

-- | Same as for 'CopyProblem'
type instance Rating PairCopyProblem = Sum Double

type instance Location PairCopyProblem = PairLocation

type instance Signal PairCopyProblem = CopyProblemSignal

instance ArenaProblem PairCopyProblem where
  stopSignal PairCopyProblem{..} = Just Stop

  minLandmarkRating _ = 0


pairProblemInitialArenaSpec ::
  PairCopyProblem
  -> (ProgramInformation, ReactionProgram)
  -> (ProgramInformation, ReactionProgram)
  -> [ColonyInitialState PairLocation (Signal PairCopyProblem)]

pairProblemInitialArenaSpec
  PairCopyProblem{..}
  senderInfoProg
  receiverInfoProg =
  [senderInitialState, receiverInitialState]
  where
    -- I'm sure I'll want to generalize this

    -- Sender colony

    senderColonySpec = ColonySpec {
      colonyCellCount = 1,
      colonyCellIxToName = IntMap.fromList [(0, "S0")],
      colonySynapseSpecs = [],
      colonyBroadcastSpecs = [
          BroadcastSpec {
              broadcastSource = Source {
                  sourceCellIx = 0,
                  sourcePattern = broadcastInputIndexToPattern + j,
                  sourceThreshold = broadcastThreshold,
                  sourceActivationDelta = -broadcastCost },
              broadcastMaximumAge = Just 1,
              broadcastSignal = TransmitBit j }
          | j <- [0 .. broadcastWidth - 1]],
      colonySensorSpecs = senderRoleSpec : senderDataSpecs
      }

    senderRoleSpec = SensorSpec {
      sensorSignal = SenderRole,
        sensorTarget = Target {
          targetCellIx = 0,
            targetPattern = 1,
            targetActivationDelta = roleActivation } }

    senderDataSpecs = [
      SensorSpec {
          sensorSignal = InputBit j,
            sensorTarget = Target {
              targetCellIx = 0,
                targetPattern = inputWordIndexToPattern + j,
                targetActivationDelta = inputActivation } }
      | j <- [0 .. wordSize - 1]]

    senderInitialState = ColonyInitialState {
      initName = "Sender",
      initColonySpec = senderColonySpec,
      initProgram = senderInfoProg,
      initLocation = Sender
      }

    -- Receiver colony

    receiverColonySpec = ColonySpec {
      colonyCellCount = 1,
      colonyCellIxToName = IntMap.fromList [(0, "R0")],
      colonySynapseSpecs = [],
      colonyBroadcastSpecs = [
          -- stop
          BroadcastSpec {
              broadcastSource = Source {
                  sourceCellIx = 0,
                  sourcePattern = 0,
                  sourceThreshold = outputThreshold,
                  sourceActivationDelta = -outputCost },
              broadcastMaximumAge = Just 1,
              broadcastSignal = Stop },
          -- beacon
          BroadcastSpec {
              broadcastSource = Source {
                  sourceCellIx = 0,
                  sourcePattern = beaconPattern,
                  sourceThreshold = beaconThreshold,
                  sourceActivationDelta = -beaconCost },
              broadcastMaximumAge = Just 1,
              broadcastSignal = Beacon }
          ] ++ [
          -- data bits
          BroadcastSpec {
              broadcastSource = Source {
                  sourceCellIx = 0,
                  sourcePattern = outputWordIndexToPattern + j,
                  sourceThreshold = outputThreshold,
                  sourceActivationDelta = -outputCost },
              broadcastMaximumAge = Just 1,
              broadcastSignal = OutputBit j }
          | j <- [0 .. wordSize - 1]],

      colonySensorSpecs = receiverRoleSpec : receiverDataSpecs
      }

    receiverRoleSpec = SensorSpec {
        sensorSignal = ReceiverRole,
        sensorTarget = Target {
            targetCellIx = 0,
            targetPattern = 2,
            targetActivationDelta = roleActivation } }

    receiverDataSpecs = [
        SensorSpec {
            sensorSignal = TransmitBit j,
            sensorTarget = Target {
                targetCellIx = 0,
                targetPattern = broadcastOutputIndexToPattern + j,
                targetActivationDelta = broadcastActivation } }
        | j <- [0 .. broadcastWidth - 1]]

    receiverInitialState = ColonyInitialState {
      initName = "Receiver",
      initColonySpec = receiverColonySpec,
      initLocation = Receiver,
      initProgram = receiverInfoProg
      }


instance InputOutputProblem PairCopyProblem where

  type instance Input PairCopyProblem = CopyProblemInput

  type instance Output PairCopyProblem = CopyProblemOutput


  -- | This is mostly the same as 'Digiostelium.CopyProblem' except that the
  -- sender and receiver are distinct agents at distinct locations

  problemInputToArenaInput PairCopyProblem{..} CopyProblemInput{..} =
    (maxSteps, [senderRole : receiverRole : convertAll inputBits])
    where
      senderRole = BroadcastState {
        broadcastSignal = SenderRole,
        broadcastMaximumAge = Nothing,
        broadcastFromColonyIx = Nothing,
        broadcastOrigin = Sender,
        broadcastAge = 0 }


      receiverRole = BroadcastState {
        broadcastSignal = ReceiverRole,
        broadcastMaximumAge = Nothing,
        broadcastFromColonyIx = Nothing,
        broadcastOrigin = Receiver,
        broadcastAge = 0 }

      convert j = BroadcastState {
        broadcastSignal = InputBit j,
        broadcastMaximumAge = Nothing,
        broadcastFromColonyIx = Nothing,
        broadcastOrigin = Sender,
        broadcastAge = 0 }

      convertAll = map convert

  arenaOutputToProblemOutput _ colonyOutputSeq =
    genericArenaOutputToCopyProblemOutput colonyOutputSeq


  inputsForRating PairCopyProblem{..} =
    allInputsFromWordSize wordSize


-- | Convert input to string of 0s and 1s.
instance Sniplet PairCopyProblem CopyProblemInput where
  toInlines PairCopyProblem{..} pInput =
    toInlines wordSize pInput

-- | Convert output to 0s and 1s, and how many steps.
instance Sniplet PairCopyProblem CopyProblemOutput where
  toInlines PairCopyProblem{..} output =
    toInlines wordSize output


-- | The point of this madness is that compiling the genome
-- involves a lot of bit fiddling and the time consumed will be
-- significant.  So we compile the genomes ahead of time for
-- every egg.  This is the storage object.  Much of what follows
-- is done with lenses and the 'HasGenome' class so that it works
-- on 'Egg' and 'Agent'.

data CompilationCache thingWithGenome =
  CompilationCache {
  _thingWithGenome :: thingWithGenome,
  _instructions :: [Instruction],
  _progAndInfo :: (ProgramInformation, ReactionProgram)
  }

makeClassy ''CompilationCache


compileToCache ::
  (GenomeSpec genomeSpec,
   HasGenome thingWithGenome (Genome genomeSpec))
  =>
  genomeSpec
  -> thingWithGenome
  -> CompilationCache thingWithGenome

compileToCache genomeSpec _thingWithGenome =
  CompilationCache{..}
  where
    _instructions = interpretGenome genomeSpec (getGenome _thingWithGenome)
    _progAndInfo =
      case compile (getInstructionSpec genomeSpec) _instructions of
        Left errors ->
          throw (CompileErrors errors _instructions)
        Right progAndInfo ->
          progAndInfo


compileAllToCache ::
  (GenomeSpec genomeSpec,
   HasGenome thingWithGenome (Genome genomeSpec)
  ) =>
  genomeSpec
  -> [thingWithGenome]
  -> [CompilationCache thingWithGenome]

compileAllToCache genomeSpec things =
  [ compileToCache genomeSpec thing
  | thing <- things]


data InteractionResult =
  InteractionResult {
  numTimeSteps :: Int,
  problemInput :: Input PairCopyProblem,
  problemOutput :: Output PairCopyProblem,
  hereTotalReactionCount :: Int,
  hereTotalReactionDeltas :: Int,
  thereTotalReactionCount :: Int,
  thereTotalReactionDeltas :: Int
  }
  deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)



runPairInteraction ::
  PairCopyProblem
  -> CompilationCache thingWithGenome
  -> CompilationCache thingWithGenome
  -> CopyProblemInput
  -> InteractionResult

runPairInteraction problem@PairCopyProblem{..} here there problemInput =
  InteractionResult{..}
  where
    colonyInitStates =
      (pairProblemInitialArenaSpec
       problem
       (here ^. compilationCache . progAndInfo)
       (there ^. compilationCache . progAndInfo))

    (maxTimeSteps, arenaInput) = problemInputToArenaInput problem problemInput

    ArenaRunSummary{..} =
      (runArena
       colonyInitStates
       (stopSignal problem)
       maxTimeSteps
       arenaInput)

    problemOutput@CopyProblemOutput{..} =
      arenaOutputToProblemOutput problem arenaRunBroadcasts

    hereSummary = arenaColonyIxToResourcesUsed IntMap.! 0
    thereSummary = arenaColonyIxToResourcesUsed IntMap.! 1

    hereTotalReactionCount =
      totalReactionCount
      $ colonyResourcesUsed
      $ hereSummary

    hereTotalReactionDeltas =
      totalReactionDeltas
      $ colonyResourcesUsed
      $ hereSummary

    thereTotalReactionCount =
      totalReactionCount
      $ colonyResourcesUsed
      $ thereSummary

    thereTotalReactionDeltas =
      totalReactionDeltas
      $ colonyResourcesUsed
      $ thereSummary


rateInteractionResult ::
  PairCopyProblem
  -> InteractionResult
  -> (Rating PairCopyProblem, Rating PairCopyProblem)

rateInteractionResult PairCopyProblem{..} InteractionResult{..} =
  (Sum $ score hereTotalReactionCount hereTotalReactionDeltas,
   Sum $ score thereTotalReactionCount thereTotalReactionDeltas)
  where
    CopyProblemInput{..} = problemInput
    CopyProblemOutput{..} = problemOutput

    correctnessScore =
      countCorrectBits wordSize inputBits dataBits

    timeScore = maxSteps - (minSteps `max` numTimeSteps)

    beaconScore =
      if Beacon `elem` nonDataOutputs
      then beaconPoints
      else 0

    score reactionCount reactionDeltas =
      (correctnessFactor * fromIntegral (maxSteps * correctnessScore)
        + fromIntegral timeScore
        + beaconScore
        - reactionCountCost
        - reactionDeltaCost)
      where
        averageReactionCount =
          fromIntegral reactionCount / fromIntegral numTimeSteps

        averageTotalDeltas =
          fromIntegral reactionDeltas / fromIntegral numTimeSteps

        reactionCountCost =
          max 0 (reactionCostFactor * (averageReactionCount - minReactions))

        reactionDeltaCost =
          max 0 (deltaCostFactor * (averageTotalDeltas - minTotalDeltas))



ratePairOn ::
  PairCopyProblem
  -> CompilationCache thingWithGenome
  -> CompilationCache thingWithGenome
  -> CopyProblemInput -> (Rating PairCopyProblem, Rating PairCopyProblem)

ratePairOn problem here there problemInput =
  rateInteractionResult problem
  $ runPairInteraction problem here there problemInput



ratePairOnAll ::
  PairCopyProblem
  -> CompilationCache thingWithGenome
  -> CompilationCache thingWithGenome
  -> (Rating PairCopyProblem, Rating PairCopyProblem)

ratePairOnAll problem@PairCopyProblem{..} here there =
  (sum (map fst ratings), sum (map snd ratings))
  where
    ratings = [ ratePairOn problem here there problemInput
              | problemInput <- allInputsFromWordSize wordSize ]


meanAboveQuantile :: Double -> Vector Double -> Double
meanAboveQuantile quantile xs =
  assert (0.0 <= quantile)
  $ assert (quantile < 1.0)
  sumElements xsKeep / fromIntegral keepCount
  where
    nItems = size xs
    dropCount = floor (quantile * fromIntegral nItems)
    keepCount = nItems - dropCount
    xsSorted = sortVector xs
    xsKeep = subVector dropCount keepCount xsSorted


sumInteractionsQuantile ::
  PairCopyProblem
  -> Double -- ^ quantile
  -> Double -- ^ self weight
  -> Double -- ^ here weight
  -> Double -- ^ there weight
  -> [CompilationCache a]
  -> [CompilationCache a]
  -> [(a, Rating PairCopyProblem)]

sumInteractionsQuantile problem quantile selfWeight0 hereWeight0 thereWeight0 thingsToRate rateAgainst =
  assert (totalWeight > 0)
  ratedThings
  where

    totalWeight = selfWeight0 + hereWeight0 + thereWeight0

    selfWeight = selfWeight0 / totalWeight
    hereWeight = hereWeight0 / totalWeight
    thereWeight = thereWeight0 / totalWeight

    -- Build matrices going across rows.  Each row has the same
    -- here agent.  Each column has the same there agent.  I'm
    -- putting parallelism here because there will generally be
    -- hundreds of pairs of agents.

    ratingMatrices forRows forCols =
      (hereMat, thereMat)
      where
        ratings =
          parListDeep
          [ ratePairOnAll problem here there
          | here <- forRows,
            there <- forCols ]
        hereRatings = map (getSum . fst) ratings
        thereRatings = map (getSum . snd) ratings
        nRows = length forRows
        nCols = length forCols
        hereMat = ((nRows >< nCols) hereRatings)
        thereMat = ((nRows >< nCols) thereRatings)

    -- A: thingsToRate, B: rateAgainst
    (hereAA, thereAA) = ratingMatrices thingsToRate thingsToRate
    (hereAB, _thereAB) = ratingMatrices thingsToRate rateAgainst
    (_hereBA, thereBA) = ratingMatrices rateAgainst thingsToRate
    -- There's no need to do those and those (BB)

    hereMat =
      hereAA ||| hereAB

    thereMat =
      thereAA
      ===
      thereBA

    selfVals = takeDiag hereAA

    -- For here ratings, split into rows
    hereAvgs =
      fromList $ parListDeep
      [ meanAboveQuantile quantile v
      | v <- toRows hereMat]

    -- For there ratings, split into columns
    thereAvgs =
      fromList $ parListDeep
      [ meanAboveQuantile quantile v
      | v <- toColumns thereMat]

    overall = (scale selfWeight selfVals
               + scale hereWeight hereAvgs
               + scale thereWeight thereAvgs)

    ratedThings = [ (thingPlus ^. thingWithGenome, Sum (overall `atIndex` k))
                  | thingPlus <- thingsToRate
                  | k <- [0 ..]]





sumAllPairwiseInteractions ::
  (GenomeSpec genomeSpec,
   HasGenome thingWithGenome (Genome genomeSpec))
  => genomeSpec
  -> PairCopyProblem
  -> [thingWithGenome] -- ^ things to rate
  -> [thingWithGenome] -- ^ rate against those things too
  -> [(thingWithGenome, Sum Double)]

sumAllPairwiseInteractions genomeSpec problem thingsToRate rateAgainst =
  (sumInteractionsQuantile
   problem
   0.0
   0.0
   1.0
   1.0
   (compileAllToCache genomeSpec thingsToRate)
   (compileAllToCache genomeSpec rateAgainst))



sumPairwiseInteractionsQuantile ::
  (GenomeSpec genomeSpec,
   HasGenome thingWithGenome (Genome genomeSpec))
  => genomeSpec
  -> PairCopyProblem
  -> Double -- ^ quantile
  -> Double -- ^ self weight
  -> Double -- ^ here weight
  -> Double -- ^ there weight
  -> [thingWithGenome] -- ^ things to rate
  -> [thingWithGenome] -- ^ rate against those things too
  -> [(thingWithGenome, Sum Double)]

sumPairwiseInteractionsQuantile genomeSpec problem quantile selfWeight hereWeight thereWeight thingsToRate rateAgainst =
  (sumInteractionsQuantile
   problem
   quantile
   selfWeight
   hereWeight
   thereWeight
   (compileAllToCache genomeSpec thingsToRate)
   (compileAllToCache genomeSpec rateAgainst))





-- | This will eventually require some configuration
instance RateEggs PairCopyProblem where

  rateEggs genomeSpec problem@PairCopyProblem{..} livingAgents eggsToRate =
    case ratingSystem of
      CompareAll ->
        return (sumAllPairwiseInteractions genomeSpec problem eggsToRate rateAgainst)
        where
          rateAgainst = map Agent.egg (Map.elems livingAgents)
      CompareQuantile {..} ->
        assert (0.0 < compareQuantile)
        $ assert (compareQuantile <= 1.0)
        $ return (sumPairwiseInteractionsQuantile genomeSpec problem compareQuantile selfWeight hereWeight thereWeight eggsToRate rateAgainst)
        where
          rateAgainst = map Agent.egg (Map.elems livingAgents)




-- === Tracing === --

tracePairOn ::
  PairCopyProblem
  -> CompilationCache thingWithGenome
  -> CompilationCache thingWithGenome
  -> CopyProblemInput
  -> (Input PairCopyProblem, Output PairCopyProblem,
     ArenaTraceSummary (Location PairCopyProblem) (Signal PairCopyProblem))

tracePairOn problem here there problemInput =
  (problemInput, problemOutput, trSummary)
  where
    colonyInitStates =
      (pairProblemInitialArenaSpec
       problem
       (here ^. compilationCache . progAndInfo)
       (there ^. compilationCache . progAndInfo))

    (maxTimeSteps, arenaInput) = problemInputToArenaInput problem problemInput

    trSummary@ArenaTraceSummary{..} =
      (runArenaTrace
       colonyInitStates
       (stopSignal problem)
       maxTimeSteps
       arenaInput)

    ArenaRunSummary{..} = trArenaRunSummary

    problemOutput@CopyProblemOutput{..} =
      arenaOutputToProblemOutput problem arenaRunBroadcasts



tracePairOnAll ::
  PairCopyProblem
  -> CompilationCache thingWithGenome
  -> CompilationCache thingWithGenome
  -> [(Input PairCopyProblem, Output PairCopyProblem,
      ArenaTraceSummary (Location PairCopyProblem) (Signal PairCopyProblem))]

tracePairOnAll problem here there =
  parListDeep
  $ map (tracePairOn problem here there)
  $ inputsForRating problem



-- Has to go at the end because of something being in scope
$(deriveJSON myJSONOpts ''PairCopyProblem)
$(deriveJSON myJSONOpts ''PairRatingSystem)
$(deriveJSON myJSONOpts ''InteractionResult)
