{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE DefaultSignatures #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE OverloadedStrings #-}


module Digiostelium.PandocUtil (
  -- * Show-like classes for 'Pandoc' results
  ToPandoc(..),
  Snip(..),
  Sniplet(..),
  GetTitle(..),
  -- * Tools for rendering stuff to 'Builder'
  buildJSON,
  buildYAML,
  buildPandocPlainNoWrap,
  buildPandocMarkdown,
  buildXMLLight,
  myYAMLPrettyConf,
  -- * Tools for constructing 'Pandoc' elements
  texts,
  paras,
  parat,
  instructionBlock,
  reactionBlock,
  toYAMLCodeBlock
  ) where

import Control.Exception

import Data.Aeson (ToJSON(..))
import qualified Data.Aeson as J
import Data.ByteString.Builder
import qualified Data.ByteString.UTF8 as U
import qualified Data.ByteString.Lazy.UTF8 as LU
import Data.List (intercalate)
import Data.Map (Map)
import qualified Data.Map as Map
import Data.Set (Set)
import qualified Data.Set as Set
import Data.String
import Data.Text.Encoding (encodeUtf8Builder)
import Data.Text (Text)
import Data.Yaml ()
import Data.Yaml.Pretty

import Text.Pandoc
import Text.Pandoc.Builder
import qualified Text.XML.Light as XMLLight

import BaselMachine.Instruction


-- | A 'Snip' object can be converted to Pandoc 'Blocks'.  The
-- @opt@ object is a way to specify options.
class Snip opt t where

  toBlocks :: opt -> t -> Blocks

  -- | The default implementation is to convert the object to
  -- YAML in a code block.
  default toBlocks :: ToJSON t => opt -> t -> Blocks
  toBlocks _ x =
    toYAMLCodeBlock x


-- | A 'Sniplet' object can be converted to Pandoc 'Inlines'.
-- The @opt@ object is a way to specify options.
class Sniplet opt t where

  toInlines :: opt -> t -> Inlines

  -- | The default implementation is to convert the object to
  -- JSON and write that as a line of code.
  default toInlines :: ToJSON t => opt -> t -> Inlines
  toInlines _opt =
    toJSONCodeLine


-- | A 'ToPandoc' object can be converted to a complete 'Pandoc'.
-- The @opt@ object is a way to specify options.

class (GetTitle opt t, Snip opt t) => ToPandoc opt t where

  -- | Show the object as a 'Pandoc'.
  -- Default implementation: Add a title using 'getTitle' to the result of 'showAsBlocks'
  toPandoc :: opt -> t -> Pandoc

  toPandoc opt x =
    setTitle (getTitle opt x)
    $ doc
    $ toBlocks opt x


-- | A 'GetTitle' object has a short title
-- The @opt@ object is a way to specify options.
class GetTitle opt t where
  -- | Short title for the object.
  -- Default is the text \"(Untitled)\"

  getTitle :: opt -> t -> Inlines

  getTitle _ _x = text (fromString "(Untitled)")



-- | Given a 'Map', list out all the objects under a single title.
-- The @opt@ object is a way to specify options.
instance (ToPandoc opt t) => ToPandoc opt (Map Integer t) where

  toPandoc opt rMap =
    if Map.null rMap
    then
      doc mempty
    else
      setTitle (getTitle opt rMap)
      $ doc
      $ toBlocks opt rMap


-- | The title of the 'Map' is the title of the first element

instance (GetTitle opt t) => GetTitle opt (Map Integer t) where
  getTitle opt rMap =
    if Map.null rMap
    then
      text (fromString "(Untitled)")
    else let
      (_, r1) = Map.findMin rMap
      in
        getTitle opt r1


instance (Snip opt t) => Snip opt (Map Integer t) where
  toBlocks opt rMap =
    mconcat [ toBlocks opt r
            | (_, r) <- Map.toList rMap]



-- === Some tools === --

-- | Write a Pandoc object as plain text without line wrapping.
writePlainNoWrap ::
  PandocMonad m =>
  Pandoc
  -> m Text

writePlainNoWrap pandoc =
  writePlain (def {writerWrapText = WrapNone}) pandoc


-- | Spell out a Pandoc writing action, returning the resulting
-- 'Text' or throwing the 'PandocError' as an exception
runPurePandoc :: PandocPure Text -> Text
runPurePandoc axn =
  case runPure axn of
    Right t -> t
    -- This should never happen:
    Left pandocError -> throw pandocError


{- | $toolsBuilder
Tools for rendering text-like things to 'Builder'

The madness:

Pandoc produces Data.Text, which is easily converted to Data.ByteString.ByteString
but not Lazy.

Data.Text can also be converted to Data.ByteString.Builder.Builder.
There's a way to send one of these efficiently to an output
handle (Data.ByteString.Builder.hPutBuilder).

Aeson encodes to Data.ByteString.Lazy.ByteString

Yaml encodes to Data.ByteString.ByteString (strict)

Converting a Lazy ByteString to strict requires a full copy.

I'm going to convert everything to a Data.ByteString.Builder.Builder.

-}


-- | Utility for rendering a JSON-convertible object as JSON in a
-- 'Builder'
buildJSON ::
  (ToJSON t)
  =>
  t -> Builder

buildJSON x =
  lazyByteString
  $ J.encode x -- gives a Data.Lazy.ByteString in UTF8


-- | Utility for rendering a JSON-convertible object as YAML in a
-- 'Builder'
buildYAML ::
  (ToJSON t)
  =>
  t -> Builder

buildYAML x =
  byteString
  $ encodePretty myYAMLPrettyConf x -- gives a Data.ByteString in UTF8


-- | Utility for rendering a 'Pandoc' object to a 'Builder' as
-- unwrapped plain text.
buildPandocPlainNoWrap :: Pandoc -> Builder

buildPandocPlainNoWrap d =
  encodeUtf8Builder
  $ runPurePandoc
  $ writePlainNoWrap d


-- | Utility for rendering a 'Pandoc' object to a 'Builder' as
-- markdown
buildPandocMarkdown :: Pandoc -> Builder

buildPandocMarkdown d =
  encodeUtf8Builder
  $ runPurePandoc
  $ writeMarkdown def d


-- | Encode an 'Text.XML.Light.Element' as a 'Builder'
buildXMLLight :: XMLLight.Element -> Builder

buildXMLLight e =
  stringUtf8
  $ XMLLight.showTopElement e




-- | 'text' composed with 'show'
texts :: Show t => t -> Inlines
texts = text . fromString . show

-- | 'para' composed with 'show'
paras :: Show t => t -> Blocks
paras = para . texts

-- | 'para' composed with 'text'
parat :: String -> Blocks
parat = para . text . fromString

instructionBlock :: [(Int, Instruction)] -> Blocks
instructionBlock indexedInstructions =
  codeBlock
  $ fromString
  $ intercalate "\n"
  $ map printNumberedInstruction
  $ indexedInstructions

reactionBlock :: Set Reaction -> Blocks
reactionBlock reactions =
  codeBlock
  $ fromString
  $ intercalate "\n"
  $ map printReaction (Set.elems reactions)


-- | Convert an object to JSON text and put it in a code 'Inline'
toJSONCodeLine :: ToJSON a => a -> Inlines
toJSONCodeLine x =
  code -- to Inline
  $ fromString -- to Text
  $ LU.toString -- to String
  $ J.encode x -- Data.ByteString.Lazy


-- | Convert an object to YAML text and put it in a code 'Block'
toYAMLCodeBlock :: ToJSON a => a -> Blocks
toYAMLCodeBlock x =
  codeBlock -- to Block
  $ fromString -- to Text
  $ U.toString -- to String
  $ encodePretty myYAMLPrettyConf x -- Data.ByteString



myYAMLPrettyConf :: Config
myYAMLPrettyConf =
  setConfCompare compare defConfig
