{-# LANGUAGE ScopedTypeVariables, BangPatterns,
RecordWildCards, NamedFieldPuns, DisambiguateRecordFields,
TypeFamilies, ParallelListComp, FlexibleInstances, MultiParamTypeClasses,
FlexibleContexts,
DeriveDataTypeable, DeriveGeneric, TemplateHaskell #-}


module Digiostelium.PolyploidGenome (
  PolyploidGenomeSpec(..),
  makeDefaultPolyploidGenomeSpec,
  makeDefaultPolyploidAlignmentSpec
  ) where

import GHC.Generics

import Control.DeepSeq
import Control.Exception
import Control.Monad

import Data.Aeson hiding (Array)
--import Data.Aeson.Types hiding (Array)
import Data.Aeson.TH
import Data.Array
import Data.Binary
import Data.Data
import Data.Default
import Data.List
-- import Data.Monoid
import Data.Map ()
import Data.Map (Map)
import qualified Data.Map as Map
--import Data.Text (Text)
import qualified Data.Text as Text
--import Data.Vector (Vector)
import qualified Data.Vector as Vector
-- import Data.Typeable
--import Debug.Trace
import Text.Printf

import Digiostelium.AesonUtil
import Data.BitSource
import Data.BitString
import Digiostelium.BlockSize
import Digiostelium.Chromosome
import Digiostelium.GeneralName
import Digiostelium.Genome
import Digiostelium.GenomeDiagram
--import BaselMachine.Instruction
import WGM.ParUtil
import System.Random.RandomST
import qualified Digiostelium.RangeMap as R
import WGM.ShortShow
import Digiostelium.SmithWaterman



-- | The exported format is a list of bundles, each of which is a
-- list of chromosomes, each of which is a little-endian string
-- of @0@s and @1@s.
data PolyploidGenomeSpec =
  PolyploidGenomeSpec {
    expansionFactor :: Int,
    -- ^ how many bits in a codon
    instructionSpec :: InstructionSpec,
    crossoverSpec :: CrossoverSpec,
    pGameteChromosomeUnrecombined :: Double,
    -- ^ probability that a gamete gets an un-recombined copy of
    -- a parent cromosome out of a bundle (but still mutated)
    mutationSpec :: SingleChromosomeMutationSpec,
    -- ^ for now, no whole chromosome mutations
    initialNumGenes :: [(Int, Int)],
    -- ^ number of genes in each chromosome in an initial genome;
    -- interpreted as
    -- @initialNumGenes !! bundleNumber = (numStrands, numGenes each)]@
    initialBundlesHomogeneous :: Bool
    -- ^ 'True' means the chromosomes within a bundle are the
    -- same in randomly created genomes, 'False' means they are
    -- created independently
    }
  deriving (Read, Show, Eq, Ord, Typeable, Data, Generic)

instance BlockSize PolyploidGenomeSpec where
  getBlockSize PolyploidGenomeSpec{..} =
    getBlockSize crossoverSpec

  setBlockSize n PolyploidGenomeSpec{..} =
    PolyploidGenomeSpec {
    crossoverSpec = setBlockSize n crossoverSpec,
    mutationSpec = setBlockSize n mutationSpec,
    ..}

instance Default PolyploidGenomeSpec where
  def = makeDefaultPolyploidGenomeSpec 3 1 (2,16) def

-- | Make a 'PolyploidSingletonGenomeSpec' from some basic
-- information.  The important computation this does is the block
-- size.
makeDefaultPolyploidGenomeSpec ::
  Int -- ^ expansion factor
  -> Int -- ^ initial number of bundles
  -> (Int,Int) -- ^ initial number of strands per bundle, and genes per strand
  -> InstructionSpec
  -> PolyploidGenomeSpec
makeDefaultPolyploidGenomeSpec expansionFactor nBundles bundleSpec instructionSpec =
  PolyploidGenomeSpec {..}
  where
    blockSize = expansionFactor * bitsPerInstruction instructionSpec
    crossoverSpec = makeBlockCrossoverSpec (SinglePointAligned def) blockSize
    mutationSpec = setBlockSize blockSize def
    pGameteChromosomeUnrecombined = 0
    initialNumGenes = replicate nBundles bundleSpec
    initialBundlesHomogeneous = True


geneSize :: PolyploidGenomeSpec -> Int
geneSize (PolyploidGenomeSpec {..}) =
  expansionFactor * sizeInBits instructionSpec undefined

-- basically (Int, Int) but with named fields
-- and strictness

data ChromosomeIx = ChromosomeIx {
  bundleNumber :: !Int,
  strandNumber :: !Int
  } deriving (Eq, Ord, Read, Show, Typeable, Data, Generic)

data Strand = Strand {
  chromosome :: BitString,
  source :: R.RangeMap (Int, Int) (FromChromosome PolyploidGenomeSpec)
  } deriving (Read, Show, Eq, Ord, Typeable, Data, Generic)


data PGenome =
  Genome {
    strands :: Map ChromosomeIx Strand
    }
  deriving (Read, Show, Eq, Ord, Typeable, Data, Generic)

data PInterval =
  Interval !ChromosomeIx !Int !Int
  -- ^ chromosome index and [begin,end) bit indices of a
  -- contiguous part of the genome
  deriving (Read, Show, Eq, Ord, Typeable, Data, Generic)

data PMutation =
  Mutation ChromosomeIx SingleChromosomeMutation
  deriving (Read, Show, Eq, Ord, Typeable, Data, Generic)

instance GeneralName PolyploidGenomeSpec where
  generalName _ = "Polyploid"

instance GenomeSpec PolyploidGenomeSpec where
  type Genome PolyploidGenomeSpec = PGenome

  type Interval PolyploidGenomeSpec = PInterval

  type Mutation PolyploidGenomeSpec = PMutation

  getInstructionSpec = instructionSpec

  genomeLength _ Genome{..} =
    sum [toInteger (bsNumBits chromosome)
         | Strand{..} <- Map.elems strands]

  geneDetails genomeSpec@PolyploidGenomeSpec{..} genome@Genome{..} =
    let
      codeOrder = (expansionFactor - 1) `quot` 2
      -- gene size in nucleotide bits
      gsNucBits = geneSize genomeSpec
      -- gene size in codon bits
      (gsCodonBits, rCB) = gsNucBits `quotRem` expansionFactor
    in
     assert (0 == rCB) $ do
       (c, strand) <- Map.toList strands
       let Strand {..} = strand
           nNucBits = chromosomeLength chromosome
           (numGenes, rNG) = nNucBits `quotRem` gsNucBits
           chromosomeCodonBits = bsMajorityDecode codeOrder 0 nNucBits chromosome
       j <- [0 .. numGenes - 1]
       let kNB = j * gsNucBits
           interval = Interval c kNB (kNB + gsNucBits)
           nucleotideBits = chromosomeSubstring kNB (kNB + gsNucBits) chromosome
           kCB = j * gsCodonBits
           codonBits = bsSubstring kCB (kCB + gsCodonBits) chromosomeCodonBits
           (instruction, leftovers) =
             readFromBitList instructionSpec codonBits
           geneSource = intervalSource genomeSpec genome interval
           instructionEncodingDetails =
             (majorityCodeCompleteEncodingDetails
              expansionFactor
              nucleotideBits
              codonBits
              (partialEncodingDetails instructionSpec codonBits))
       (assert (0 == rNG) $
        assert (bsNull leftovers) $
        assert (geneSource /= R.empty) $
        return (GeneDetails {..}))

  exportGenomeData _ Genome{..} =
    toJSON [
      [
        (bsTo01String (chromosome strand))
      | strand <- bundle
      ]
    | bundle <- strandmapToBundles strands
    ]

  importGenomeDataParser _ v =
    let
      -- These are Aeson parsers
      binaryChromosomeP = withText "string of bits" $ \t ->
        case bsFrom01StringChecked (Text.unpack t) of
          Left eMsg ->
            fail eMsg
          Right bs ->
            return bs

      binaryBundleP = withArray "array of strands" $ \a -> do
        mapM binaryChromosomeP (Vector.toList a)

      polyploidGenomeP = withArray "array of chromosome bundles" $ \a ->
        mapM binaryBundleP (Vector.toList a)

    in do
      bundles <- polyploidGenomeP v
      return $ Genome {strands = unsourcedChromosomesToStrandmap bundles}


  intervalSource _ Genome{..} (Interval cIx begin end) =
    let
      Strand{..} = strands Map.! cIx
      result = R.keyIntersection (begin, end) source
    in
     assert (result /= R.empty) $
     assert (R.isWellFormed result) $
     result

  allPossibleMutations PolyploidGenomeSpec{..} Genome{..} = do
    (cIx, strand) <- Map.toList strands
    let Strand {..} = strand
    m <- allPossibleSingleChromosomeMutations mutationSpec (bsNumBits chromosome)
    return (Mutation cIx m)

  applyMutations
    PolyploidGenomeSpec{..}
    wrappedMutations Genome{..} =
    let
      parentIx = -1 -- not the result of a crossover
      mutationsByCIx = Map.fromListWith (++) [
        (cIx, [m])
        | Mutation cIx m <- wrappedMutations]
      -- each new strand is a copy from the one parent
      newIxedStrands = do
        (j, Strand {..}) <- Map.toList strands
        let (newChromosome, annotatedSources) =
              mutateChromosome chromosome (Map.findWithDefault [] j mutationsByCIx)
            newSource = R.RangeMap [
              (r, FromChromosome {inParent = Interval j pBegin pEnd, ..})
              | (r, (pBegin, pEnd), relativeFlips, _chunk) <- annotatedSources]
        (assert (newSource /= R.empty) $
         return (j,
                 Strand { chromosome = newChromosome,
                          source = newSource }))
    in
     Genome { strands = Map.fromList newIxedStrands }

  intervalsChanged PolyploidGenomeSpec{..} (Mutation cIx m) Genome{..} =
    let
      (begin, end) = mutationInterval m
      Strand {..} = strands Map.! cIx
      n = bsNumBits chromosome
    in
     assert (0 <= begin) $
     assert (begin < end) $
     if begin < n
     then [Interval cIx begin (n `min` end)]
     else []

instance R.Range PInterval where

  empty = Interval (ChromosomeIx 0 0) 0 0

  length (Interval _ begin end) =
    end - begin

  take n (Interval cIx begin end) =
    let
      (nBegin, nEnd) = R.take n (begin, end)
    in
     Interval cIx nBegin nEnd

  drop n (Interval cIx begin end) =
    let
      (nBegin, nEnd) = R.drop n (begin, end)
    in
     Interval cIx nBegin nEnd

instance R.BoundedRange PInterval where
  bounds (Interval _ begin end) = (begin, end)

  intersection
    (Interval xCIx xBegin xEnd)
    (Interval yCIx yBegin yEnd) =
      if xCIx == yCIx
      then let
        (zBegin, zEnd) = R.intersection (xBegin, xEnd) (yBegin, yEnd)
        in
         (Interval xCIx zBegin zEnd)
      else
        R.empty


-- | TO DO: This implementation assumes a diploid genome
instance IntervalCoordinates PInterval where
  intervalSegmentCoordinates (Interval (ChromosomeIx {..}) begin end) = (x1, x2, y)
    where
      x1 = fromIntegral begin
      x2 = fromIntegral end
      y = fromIntegral $ -(3 * bundleNumber + strandNumber)
      -- To properly handle a general polyploid genome,
      -- the 3 * bundleNumber here needs to be based on
      -- the number of strands per bundle
      -- The negative is so the bundles and strands appear
      -- in order top to bottom

{-
prettyShowStrand (Strand {..}) =
  let
    lineStrs = [
      printf "[%d,%d) %s pIx=%d b=%d s=%d [%d,%d) %s"
      rBegin
      rEnd
      (bsTo01String (bsSubstring rBegin rEnd chromosome))
      parentIx
      bundleNumber
      strandNumber
      pBegin
      pEnd
      (show relativeFlips)
      | ((rBegin, rEnd), FromChromosome {..}) <- R.rangeMapItems source,
        let Interval cIx pBegin pEnd = inParent,
        let ChromosomeIx {..} = cIx ]
  in
   unlines lineStrs

prettyPrintStrand s =
  putStrLn (prettyShowStrand s)
-}

unsourcedStrand :: ChromosomeIx -> Chromosome -> Strand
unsourcedStrand cIx chromosome =
  let
    nBits = bsNumBits chromosome
    parentIx = -1
    inParent = Interval cIx 0 nBits
    relativeFlips = []
    source = R.RangeMap [((0, nBits), FromChromosome {..})]
  in
   Strand {..}

unsourcedChromosomesToStrandmap :: [[Chromosome]] -> Map ChromosomeIx Strand
unsourcedChromosomesToStrandmap bundles =
  Map.fromList [
    (cIx, unsourcedStrand cIx chromosome)
    | (bundleNumber, bundle) <- zip [0..] bundles,
      (strandNumber, chromosome) <- zip [0..] bundle,
      let cIx = ChromosomeIx {..}]


--- random generation ---

instance Sample PolyploidGenomeSpec PGenome where

  pullSample gs@PolyploidGenomeSpec{..} =
    let
      nGeneBits = geneSize gs

      -- how to pull independent strands in each bundle
      ixedGeneCounts =
        [(ChromosomeIx {..}, nGenes)
        | (bundleNumber, (nStrands, nGenes)) <- zip [0..] initialNumGenes,
          strandNumber <- [0 .. nStrands - 1] ]

      pullStrand cIx nGenes =
        let
          nBits = nGenes * nGeneBits
        in do
          chromosome <- pullSample (UniformBits nBits)
          return (unsourcedStrand cIx chromosome)

      pullIndependentBundles = do
        forM ixedGeneCounts $ \(cIx, nGenes) -> do
          strand <- pullStrand cIx nGenes
          return (cIx, strand)

      -- how to pull one strand and copy it within each bundle
      pullHomogeneousBundles = do
        bundlesGrouped <- forM (zip [0..] initialNumGenes) $
          \(bundleNumber, (nStrands, nGenes)) -> do
            chromosome <- pullSample (UniformBits (nGenes * nGeneBits))
            return [(cIx, unsourcedStrand cIx chromosome)
                   | strandNumber <- [0 .. nStrands - 1],
                     let cIx = ChromosomeIx{..}]
        return (concat bundlesGrouped)

    in do
      ixedStrands <- if initialBundlesHomogeneous
                     then pullHomogeneousBundles
                     else pullIndependentBundles
      let strands = Map.fromList ixedStrands
      return Genome {..}


--- breeding ---

-- | (Internal function, part of 'pullGamete') Produce a
-- baseline 'Genome', giving it parent index 'parentIx'.  The
-- 'IntervalSource's in the resulting 'Genome' all indicate of
-- plain copies of whole 'Strand's in the parent.

germCellBase ::
  Genome PolyploidGenomeSpec -- ^ parent
  -> Int -- ^ index assigned to this parent, 0 or 1 for a diploid architecture
  -> Genome PolyploidGenomeSpec -- ^ germ cell

germCellBase Genome{..} parentIx =
  let
    newIxedStrands =
      [(cIx, selfSource cIx strand parentIx)
       | (cIx, strand) <- Map.toList strands]
  in
   Genome {strands = Map.fromList newIxedStrands}

-- | (Internal function) Given a 'ChromosomeIx' and the 'Strand'
-- there, create a new 'Strand' with the same 'Chromosome' but
-- whose source is simplified to saying the whole thing comes
-- from the corresponding 'ChromosomeIx' in one parent.  The
-- parent's index must also be given.

selfSource ::
  ChromosomeIx
  -> Strand
  -> Int -- ^ parent index to assign
  -> Strand

selfSource cIx (Strand {chromosome}) parentIx =
  let
    cLength = bsNumBits chromosome
    newSource =
      R.RangeMap [((0, cLength),
        FromChromosome {
          parentIx,
          inParent = Interval cIx 0 cLength,
          relativeFlips = []})]
  in
   Strand {source = newSource, ..}

-- | (Internal function) Given a 'Strand' and its corresponding
-- parent index and 'ChromosomeIx', apply a list of
-- 'SingleChromosomeMutation's to it, resulting in an updated
-- 'Strand'.  Its 'source' is updated to reflect any deletion or
-- duplication events, and all flipped bits.

mutateStrand ::
  Int -- ^ parent index
  -> ChromosomeIx
  -> Strand
  -> [SingleChromosomeMutation]
  -> Strand

mutateStrand parentIx cIx (Strand {..}) mutations =
  let
    (mtChromosome, changeMap)
      = mutateChromosome chromosome mutations
    newSource = R.RangeMap [
      (r, FromChromosome {inParent = Interval cIx pBegin pEnd, ..})
      | (r, (pBegin, pEnd), relativeFlips, _) <- changeMap]
  in
   --trace ("mutateStrand: mutations:\n" ++ unlines (map show mutations)) $
   --trace ("mutateStrand: changeMap:\n" ++ unlines (map show changeMap)) $
   --trace ("mutateStrand: newSource:\n" ++ unlines (map show newSource)) $
   assert (newSource /= R.empty) $
   Strand {
     chromosome = mtChromosome,
     source = newSource }

pullMutatedStrand ::
  (Sample (a, Int) [SingleChromosomeMutation])
  => a
  -> Int
  -> ChromosomeIx
  -> Strand
  -> RandomST s Strand

pullMutatedStrand mutationSpec parentIx cIx baseStrand@(Strand {..}) = do
  mutations <- pullSample (mutationSpec, bsNumBits chromosome)
  return (mutateStrand parentIx cIx baseStrand mutations)


-- | Create a germ cell base and apply random 'Mutation's to it.
-- It can now be subject to recombination.  This is in the order
-- things are performed by living organisms.  I did recombination
-- and mutation in the other order in 'HaploidSingletonGenome'
-- because this order sort of wastes time: Mutations are
-- performed on parts of the 'Genome' that never appear in a
-- gamete.  But, this order necessary because if mutations and
-- crossover aren't configured to be gene-aligned, it's possible
-- to end up with weird things like a deletion event that spans a
-- crossover point
--
-- > AAABBB
-- >  [xx)
--
-- which proably can't happen in real DNA.

pullGermCell ::
  PolyploidGenomeSpec
  -> Genome PolyploidGenomeSpec
  -> Int
  -> RandomST s (Genome PolyploidGenomeSpec)

pullGermCell (PolyploidGenomeSpec {..}) g parentIx =
  let
    Genome {..} = germCellBase g parentIx
  in do
    mtStrands <- forM (Map.toList strands) $
      \(cIx, baseStrand) -> do
        mtStrand <- pullMutatedStrand mutationSpec parentIx cIx baseStrand
        return (cIx, mtStrand)
    return (Genome {strands = Map.fromList mtStrands})


-- It has taken DAYS but thanks to RangeMap, I finally have this:
crossoverStrandsAt :: Strand -> Strand -> Int -> Int -> Strand
crossoverStrandsAt sLeft sRight leftEnd rightBegin =
    let
      cLeft = bsTake leftEnd (chromosome sLeft)
      cRight = bsDrop rightBegin (chromosome sRight)
      newChromosome = bsRepack (bsJoin cLeft cRight)
      newLeftSource = R.take leftEnd (source sLeft)
      newRightSource = R.drop rightBegin (source sRight)
      newSource = newLeftSource <> newRightSource

    in
     assert (newSource /= R.empty) $
     assert (R.isWellFormed newSource) $
     Strand {
       chromosome = newChromosome,
       source = newSource }

pullRecombinedStrand :: CrossoverSpec -> [Strand] -> RandomST s Strand
pullRecombinedStrand _ [strand@(Strand {..})] =
  -- trace ("pullRecombinedStrand: strand source:\n" ++ unlines (map show source)) $
  return strand

-- order: Randomly decide which of s1, s2 is treated as the left
-- half or the right half
pullRecombinedStrand crossoverSpec allStrands = do
  order <- pullSample (BernoulliDistribution 0.5)
  let  (s1:s2:strands) = allStrands
       (sL, sR) = if order then (s1, s2) else (s2, s1)
       cL = chromosome sL
       cR = chromosome sR
       --nL = bsNumBits cL
       --nR = bsNumBits cR
  Crossover {..} <- pullSample (crossoverSpec, cL, cR)
  let sOut = (crossoverStrandsAt sL sR leftEnd rightBegin)
  pullRecombinedStrand crossoverSpec (sOut:strands)

-- Maybe recombine
pullStrand :: Double -> CrossoverSpec -> [Strand] -> RandomST s Strand
pullStrand pUnrecombined crossoverSpec allStrands
  | (pUnrecombined == 0) = -- always recombine
    pullRecombinedStrand crossoverSpec allStrands
  | (pUnrecombined == 1) = -- never recombine
    pullSample allStrands
  | otherwise = do -- maybe recombine
    recombine <- pullSample (BernoulliDistribution pUnrecombined)
    if recombine
      then pullRecombinedStrand crossoverSpec allStrands
      else pullSample allStrands


-- | Convert a map of 'ChromosomeIx's to 'Strand's to a list of
-- bundles, each of which is a list of 'Strand's.  The output is
-- supposed to be grouped by bundle, then in order by strand number
-- @
-- [ [bundle 0 strand 0, bundle 0 strand 1,..],
--   [bundle 1 strand 0, bundle 1 strand 1, ..] .. ]
-- @

strandmapToBundles :: Map ChromosomeIx Strand -> [[Strand]]
strandmapToBundles sm =
  let
    -- List of [(strand number, strand)..]
    groupedByBundleMap =
      Map.fromListWith (++) [
        (bundleNumber, [(strandNumber, strand)])
        | (ChromosomeIx {..}, strand) <- Map.toList sm]

    groupedByBundle = Map.elems groupedByBundleMap

    -- List of [strand..]
    groupedByBundleOrdered = [
      map snd sortedIxedStrands
      | ixedStrands <- groupedByBundle,
        let sortedIxedStrands = sort ixedStrands]
  in
   -- trace ("-----") $
   -- trace (show sm) $
   -- trace ("-----") $
   -- trace (show groupedByBundleMap) $
   -- trace ("-----") $
   -- trace (show groupedByBundle) $
   -- trace ("-----") $
   -- trace (show groupedByBundleOrdered) $
   groupedByBundleOrdered

{-
bundlesToStrandmap :: [[Strand]] -> Map ChromosomeIx Strand
bundlesToStrandmap bundles =
  Map.fromList [
    (ChromosomeIx {..}, strand)
    | (bundleNumber, strands) <- zip [0..] bundles,
      (strandNumber, strand) <- zip [0..] strands]
-}

-- | Pull a random gamete, that is, it has half as many 'Strands'
-- in each bundle as the parent.

pullGamete ::
  PolyploidGenomeSpec
  -> Genome PolyploidGenomeSpec -- ^ parent genome
  -> Int -- ^ index of this parent among child's parents
  -> RandomST s [[Strand]] -- ^ resulting half genome as a list of bundles of strands

pullGamete genomeSpec@(PolyploidGenomeSpec {..}) parent parentIx = do
  Genome {..} <- pullGermCell genomeSpec parent parentIx
  let bundles = strandmapToBundles strands
  -- trace ("pullGamete: strands:\n" ++ unlines (map show (Map.toList strands))) $
  forM bundles $
    \ strands ->
    let
      nStrands = length strands
      nToPull = nStrands `div` 2
    in do
      replicateM nToPull (pullStrand
                          pGameteChromosomeUnrecombined
                          crossoverSpec
                          strands)

pullChild ::
  PolyploidGenomeSpec
  -> Genome PolyploidGenomeSpec
  -> Genome PolyploidGenomeSpec
  -> RandomST s (Genome PolyploidGenomeSpec)

pullChild genomeSpec p0 p1 = do
  g0 <- pullGamete genomeSpec p0 0
  g1 <- pullGamete genomeSpec p1 1
  let gametes = [g0, g1]
  -- Yikes!  This next line is supposed to take
  -- gametes = [
  --   [[g0b0s0], [g0b1s0]], (all g0)
  --   [[g1b0s0], [g1b1s0]], (all g1)
  --   ..]
  -- and yield
  -- jointBundles = [ [g0b0s0, g1b0s0], [g0b1s0, g1b1s0], ..]
  let jointBundles = foldl1' (zipWith (++)) gametes
  --let [ [[g0b0s0]], [[g1b0s0]] ] = gametes
  --let jointBundlesT = [ [g0b0s0, g1b0s0] ]
  let ixedStrands =
        [(ChromosomeIx {..}, strand)
        | (bundleNumber, bundle) <- zip [0..] jointBundles,
          (strandNumber, strand) <- zip [0..] bundle]
  let strands = Map.fromList ixedStrands
  let child = Genome {..}
  -- trace ("pullChild: gametes:\n" ++ show gametes) $
  -- trace ("pullChild: jointBundlesT:\n" ++ show jointBundlesT) $
  -- trace ("pullChild: test jointBundles:\n" ++ show (jointBundlesT == jointBundles)) $
  -- trace ("pullChild: ixedStrands keys:\n" ++ show (map fst ixedStrands)) $
  return child

instance
  Sample
  (PolyploidGenomeSpec, PGenome, PGenome)
  PGenome where
    pullSample (genomeSpec, p1, p2) =
      pullChild genomeSpec p1 p2

data PAlignmentSpec =
  AlignmentSpec BlockAlignmentSpec
  deriving (Read, Show, Eq, Ord)

instance Default PAlignmentSpec where
  def = AlignmentSpec def

instance BlockSize PAlignmentSpec where

  getBlockSize (AlignmentSpec s) = getBlockSize s

  setBlockSize n (AlignmentSpec s) = AlignmentSpec (setBlockSize n s)



data PAlignmentResult =
  AlignmentResult {
    resultMatrices :: [Array (Int, Int) (SmithWatermanResult BitString)]
      -- ^ indexed by @!! bundle@, @! (strand number 1, strand number 2)@
    }

makeDefaultPolyploidAlignmentSpec ::
  Int -- ^ block size
  -> PAlignmentSpec
makeDefaultPolyploidAlignmentSpec blockSize =
  AlignmentSpec (makeDefaultBlockAlignmentSpec blockSize)

-- | Alignment is done bundle by bundle, aligning each strand in
-- one genome's bundle with each strand in the other's.  The
-- overall score of an alignment result is to add up the results
-- of so pairing the strands.  All reasonable pairing of strands
-- are aligned because two genomes might have similar alleles
-- that happen to occur on different strands in the same bundle
-- because of recombination.

instance Alignable PolyploidGenomeSpec where

  type AlignmentSpec PolyploidGenomeSpec = PAlignmentSpec

  type AlignmentResult PolyploidGenomeSpec = PAlignmentResult

  bestAlignment _genomeSpec (AlignmentSpec as) gLeft gRight =
    let
      -- leftBundles, rightBundles :: [[Strand]]
      leftBundles = strandmapToBundles (strands gLeft)
      rightBundles = strandmapToBundles (strands gRight)

      resultMatrices =
        [
          let
            leftChrAr = listArray (0, length leftBundle - 1) (map chromosome leftBundle)
            rightChrAr = listArray (0, length rightBundle - 1) (map chromosome rightBundle)
          in
           blockAlignmentMatrix as leftChrAr rightChrAr
         | leftBundle <- leftBundles
         | rightBundle <- rightBundles
         ]

    in
     AlignmentResult {..}

  alignmentScore _genomeSpec _alignmentSpec (AlignmentResult {..}) =
    let
      -- list traversal in monad notation here...
      scores = do
        bundleMatrix <- resultMatrices
        result <- elems bundleMatrix
        return (bestScore result)
    in
     sum scores

  -- Implementation note: This involves a lot of string
  -- concatenation, and it may become necessary to speed this up
  -- so as to avoid repeatedly tranversing strings.

  showAlignment _genomeSpec (AlignmentSpec (BlockAlignmentSpec {..})) (AlignmentResult {..}) =
    let
      showBlock b = ' ' : bsTo01String b
      prettyResult (SmithWatermanResult {..}) =
        prettyShowAlignment (1 + alignmentBlockSize) showBlock workspace

      prettyStuff :: String
      prettyStuff =
        (concat
         [printf "bundle %d:\n%s"
          bundleNum
          (prettyBundle resultsByBundle)
         | resultsByBundle <- resultMatrices
         | bundleNum :: Int <- [0..] ])

      prettyBundle :: Array (Int, Int) (SmithWatermanResult BitString) -> String
      prettyBundle resultMatrix =
        (concat
         [
           (printf "left genome strand %d, right genome strand %d:\n%s\n%s\n"
            j k top bottom)
         | ((j, k), swResult) <- assocs resultMatrix,
           let (top, bottom) = prettyResult swResult])
    in
     prettyStuff

  alignmentMatrix gs as genomeAr =
    let
      bds@(a,b) = bounds genomeAr
      score g1 g2 =
        alignmentScore gs as (bestAlignment gs as g1 g2)

      rDiag = [
        ((j, j), score g g)
        | j <- range bds,
          let g = genomeAr ! j]

      rT1 = [
        ((j, k), score gj gk)
        | j <- range bds,
          k <- range ((j+1), b),
          let gj = genomeAr ! j,
          let gk = genomeAr ! k]

      rT2 = [
        ((k, j), s)
        | ((j, k), s) <- rT1]

      results = concat [rDiag, rT1, rT2]
    in
     array ((a, a), (b, b)) (parListDeep results)

--- routine instances ---

instance Binary Strand where

  put (Strand {..}) = do
    put (bsRepack chromosome)
    put source

  get = do
    rawChromosome <- get
    let chromosome = bsRepack rawChromosome
    source <- get
    return (Strand {..})


instance ShortShow ChromosomeIx where

  shortShow (ChromosomeIx {..}) =
    printf "b%ds%d" bundleNumber strandNumber

instance ShortShow PInterval where

  shortShow (Interval cIx begin end) =
    printf "%s:[%d,%d)" (shortShow cIx) begin end

instance Binary PolyploidGenomeSpec
instance Binary PGenome
instance Binary PInterval
instance Binary PMutation
instance Binary ChromosomeIx

instance NFData PolyploidGenomeSpec
instance NFData PGenome
instance NFData PInterval
instance NFData PMutation
instance NFData ChromosomeIx
instance NFData Strand

$(deriveJSON myJSONOpts ''PolyploidGenomeSpec)
$(deriveJSON myJSONOpts ''PInterval)
$(deriveJSON myJSONOpts ''PMutation)
$(deriveJSON myJSONOpts ''ChromosomeIx)
$(deriveJSON myJSONOpts ''PAlignmentSpec)

instance ToJSONKey ChromosomeIx
instance FromJSONKey ChromosomeIx

instance FromJSON (FromChromosome PolyploidGenomeSpec)
instance ToJSON (FromChromosome PolyploidGenomeSpec)

$(deriveJSON myJSONOpts ''Strand)
$(deriveJSON myJSONOpts ''PGenome)

--- testing ---

_testC1 :: BitString
_testC1 = read "\"0101010101010101\""

_testC2 :: BitString
_testC2 = read "\"0011001100110011\""

_testC3 :: BitString
_testC3 = read "\"0000111100001111\""

_testS1 :: Strand
_testS1 = Strand {
  chromosome = _testC1,
  source = R.RangeMap [((0,16), FromChromosome (-2) (Interval (ChromosomeIx 5 1) 0 16) [])] }

_testS2 :: Strand
_testS2 = Strand {
  chromosome = _testC2,
  source = R.RangeMap [((0,16), FromChromosome (-2) (Interval (ChromosomeIx 5 2) 0 16) [])] }

_testS3 :: Strand
_testS3 = Strand {
  chromosome = _testC3,
  source = R.RangeMap [((0,16), FromChromosome (-2) (Interval (ChromosomeIx 5 3) 0 16) [])] }

_testSC12 :: Strand
_testSC12 = crossoverStrandsAt _testS1 _testS2 8 8

_testS2M :: Strand
_testS2M = mutateStrand (-2) (ChromosomeIx 5 2) _testS2 [DeleteInterval 7 9, Flip 10]

_testSC12M :: Strand
_testSC12M = crossoverStrandsAt _testS1 _testS2M 3 3

_testSC12M3 :: Strand
_testSC12M3 = crossoverStrandsAt _testSC12M _testS3 8 12
