{-# LANGUAGE TypeFamilies, FlexibleContexts, RecordWildCards, NamedFieldPuns #-}

{-|

Module: Digiostelium.Problem

Description : Abstractions for representing a deterministic-rating-based evolutionary problem

Copyright   : © 2019 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

Refactoring notes: This module serves as a bridge between
genomes, populations, and the selection-mutation process.  I'm
moving stuff out of "Evolution" to here, and pushing much of what
used to be in this module down into more specific problem
classes.

-}

module Digiostelium.Problem where


import Data.Map (Map)

import BaselMachine.Arena

import Digiostelium.Agent
import Digiostelium.Genome

import System.Random.RandomST

type family Rating problem :: *
type family Location problem :: *
type family Signal problem :: *

-- | Class 'ArenaProblem' is for types that are associated to
-- proper rating, signal, and location types.
class ArenaProblem problem where

  -- | Maybe a signal that indicates the simulation should stop.
  -- Default is 'Nothing'
  stopSignal ::
    problem -> Maybe (Signal problem)

  stopSignal _ = Nothing


  -- | Rating below which landmarks are not recognized
  minLandmarkRating ::
    problem -> Rating problem



-- | Information needed to initialize an 'Arena'.  Includes
-- compiled genomes and whatever else might be necessary.
type family ArenaInitializationSpec problem :: *

class (ArenaProblem problem)
  =>
  InitializeArena problem where

  -- | Build the initial state of the 'Arena'
  initialArenaState ::
    problem
    -> ArenaInitializationSpec problem
    -> [ColonyInitialState (Location problem) (Signal problem)]



class (ArenaProblem problem)
  =>
  RateEggs problem where

  -- | Assign ratings to 'Egg's
  rateEggs ::
    (GenomeSpec genomeSpec)
    =>
    genomeSpec
    -> problem
    -> Map Integer (Agent (Genome genomeSpec) (Rating problem))
    -- ^ Suriving 'Agent's
    -> [Egg (Genome genomeSpec)]
    -- ^ 'Egg's that need to be rated
    -> RandomST s [(Egg (Genome genomeSpec), Rating problem)]
    -- ^ 'Egg's and their ratings
