{-# LANGUAGE
UndecidableInstances,
FlexibleInstances,
MultiParamTypeClasses,
TypeFamilies,
RecordWildCards,
NamedFieldPuns,
ScopedTypeVariables,
DeriveDataTypeable,
DeriveGeneric,
TemplateHaskell,
FlexibleContexts #-}

{-|

Module: Digiostelium.Query

Description : Abstractions for querying agents and making reports

Copyright   : © 2019 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

-}


module Digiostelium.Query (
  -- * Selecting 'Agent's from a sample run
  SelectAgents(..),
  doSelect,
  -- * Getting properties and reports from individual 'Agent's
  QueryIndividual(..)
  ) where

import GHC.Generics

import Control.Exception
import Control.Monad.Trans

import Data.Aeson
import Data.Aeson.TH
import Data.Binary
import Data.Data
import Data.Kind
import Data.Map (Map)
import qualified Data.Map as Map

import Digiostelium.AesonUtil
import Digiostelium.Agent as Agent
import Digiostelium.Genome as G
import Digiostelium.PopulationDatabase
import Digiostelium.Problem
import Digiostelium.ReportException

-- | Ways of getting a list of 'Agent's out of a sample run
data SelectAgents =
  None
  | Living
  | Landmarks
  | Generation Integer
  | TopRated
  | RatedAtLeast Value
  | IDs [Integer]
  deriving (Show, Eq, Data, Typeable, Generic)

$(deriveJSON myJSONOpts ''SelectAgents)

-- | Fetch 'Agent's from a database as specified by a 'Query'
doSelect ::
  (Binary rating,
   Ord rating,
   FromJSON rating,
   Binary genome)
  =>
     SelectAgents
     -> PopMonad genome rating (Map Integer (Agent genome rating))

doSelect None =
  return Map.empty

doSelect Living = getLivingAgents

doSelect Landmarks = getLandmarks

doSelect (Generation n) = getGeneration n

doSelect (IDs ns) = getAgents ns

doSelect TopRated = do
  living <- getLivingAgents
  let maxRating =
        (maximum
         (map Agent.rating (Map.elems living)))
      topRated =
        (Map.filter
         (\a -> Agent.rating a == maxRating)
         living)
  return topRated

doSelect (RatedAtLeast v)  = do
  living <- getLivingAgents
  case fromJSON v of
    Error msg ->
      liftIO $ throwIO (ParseRatingError msg)
    Success r ->
      (return
       (Map.filter
        (\a -> Agent.rating a >= r)
        living))



-- | Polymorphic function for getting information about an individual 'Agent'
class (GenomeSpec genomeSpec) =>
  QueryIndividual genomeSpec problem reportSpec where

  type QueryResult genomeSpec problem reportSpec :: Type

  queryIndividual ::
    genomeSpec
    -> problem
    -> reportSpec
    -> Agent (Genome genomeSpec) (Rating problem)
    -> QueryResult genomeSpec problem reportSpec
