{-# LANGUAGE DeriveDataTypeable, StandaloneDeriving, RecordWildCards, NamedFieldPuns, ScopedTypeVariables, FlexibleContexts, BangPatterns, UndecidableInstances , DeriveDataTypeable, TemplateHaskell, DeriveGeneric, GADTs #-}

module Digiostelium.SmithWaterman (
  SmithWatermanState(..),
  SmithWatermanResult(..),
  Step(..),
  alignSmithWaterman,
  prettyShowAlignment
  ) where

import GHC.Generics

import Control.DeepSeq
import Control.Monad
import Control.Monad.ST

import Data.Aeson
import Data.Aeson.TH
import Data.Array.IArray
import Data.Array.MArray
import Data.Array.Unsafe (unsafeFreeze)
import qualified Data.Array as A
import qualified Data.Array.ST as STA
import qualified Data.Array.Unboxed as UA
import Data.Binary
import Data.Data
import Data.List
import Data.Typeable
import Text.Printf

import Digiostelium.AesonUtil


data SmithWatermanState ar uar a =
  SmithWatermanState {
    wG :: Int, -- ^ weight of a gap letter
    wA :: a -> a -> Int, -- ^ weight of aligning two letters
    a0 :: [a],
    b0 :: [a],
    a :: ar Int a,
    b :: ar Int a,
    m :: Int,
    n :: Int,
    h :: uar (Int, Int) Int,
    -- ^ dynamic programming matrix of scores
    path :: uar (Int, Int) Int
    -- ^ dynamic programming matrix of backtracking directions
    } deriving (Typeable, Generic)

{-
deriving instance
  (Data (ar Int a),
  Typeable ar,
  Typeable uar,
  Typeable a,
  Data (uar (Int, Int) Int),
  Data a)
  =>
  Data (SmithWatermanState ar uar a)
-}

instance (NFData a,
          NFData (uar (Int, Int) Int),
          NFData (ar Int a))
          => NFData (SmithWatermanState ar uar a) where
  rnf (SmithWatermanState {..}) =
    rnf wG
    `seq` rnf a0
    `seq` rnf b0
    `seq` rnf a
    `seq` rnf b
    `seq` rnf m
    `seq` rnf n
    `seq` rnf h
    `seq` rnf path


data Step a =
  Start
  | Step a a
  | Delete a
  | Insert a
  deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)

instance (NFData a) => NFData (Step a)
instance (Binary a) => Binary (Step a)
$(deriveJSON myJSONOpts ''Step)

-- Contents of the path array
start, north, west, northwest :: Int
start = 0
north = 1
west = 2
northwest = north + west

initSmithWaterman ::
  Int -> (a -> a -> Int) -> [a] -> [a]
  -> ST s (SmithWatermanState A.Array (STA.STUArray s) a)
initSmithWaterman wG wA a0 b0 =
  let
    m = length a0
    n = length b0
    a = listArray (1, m) a0 -- in my references, the algorithm indexes a and b from 1 to m and n.
    b = listArray (1, n) b0 -- but steps are indexed from 0 to m and n.
  in do
    h <- newArray ((0,0), (m,n)) 0
    path <- newArray ((0,0), (m,n)) 0
    return (SmithWatermanState {..})

updateSmithWaterman (SmithWatermanState {..}) i j = do
  let ai = a ! i
      bj = b ! j
      w = wA ai bj
  valN <- readArray h (i - 1, j)
  valNW <- readArray h (i - 1, j - 1)
  valW <- readArray h (i, j - 1)
  let cNW = valNW + w
      cN  = valN + wG
      cW  = valW + wG
  let (val, step) =
        maximumBy compareFst
        [(cNW, northwest),
         (cN, north),
         (cW, west),
         (0, start)]
  writeArray h (i, j) val
  writeArray path (i, j) step

compareFst x y = compare (fst x) (fst y)

loopSmithWaterman s@(SmithWatermanState {..}) !i !j = do
  updateSmithWaterman s i j
  if j < n
    then loopSmithWaterman s i (j + 1)
    else -- j == n, done with this row
    if i < m
    then loopSmithWaterman s (i + 1) 1
    else -- i == m, done entirely
      return ()

data SmithWatermanResult a =
  SmithWatermanResult {
    workspace :: SmithWatermanState A.Array UA.UArray a,
    bestIx :: (Int, Int), -- ^ location in dynamic programming array of highest score
    bestScore :: Int, -- ^ that score
    bestPath :: [(Int, Int, Step a)] -- ^ path backtracking from highest score back to the beginning
    } deriving (Typeable, Generic)

instance Show a => Show (SmithWatermanResult a) where
  show (SmithWatermanResult {..}) =
    show (bestScore, bestIx, bestPath)

instance (NFData a) => NFData (SmithWatermanResult a) where
  rnf (SmithWatermanResult {..}) =
    workspace
    `seq` rnf bestIx
    `seq` rnf bestScore
    `seq` rnf bestPath

-- | Do Smith-Waterman alignment.
alignSmithWaterman ::
  Int -- ^ weight of a gap
  -> (a -> a -> Int) -- ^ weight of two items from the alphabet
  -> [a] -> [a] -- ^ two strings of alphabet items to align
  -> SmithWatermanResult a

alignSmithWaterman wG wA s1 s2 = runST $ do
  s <- initSmithWaterman wG wA s1 s2
  let SmithWatermanState {..} = s
  loopSmithWaterman s 1 1
  h <- unsafeFreeze h
  path <- unsafeFreeze path
  let workspace = SmithWatermanState {..}
      (bestIx, bestScore, bestPath) = alignmentPath workspace
  return (SmithWatermanResult {..})


showMatrix
  :: (IArray ar Int) =>
     ar (Int, Int) Int -> String

showMatrix h =
  let
    ((firstRow, firstCol), (lastRow, lastCol)) = bounds h
    showRow h i =
      concat [printf "%3d" (h ! (i,j))
             | j <- [firstCol .. lastCol] ]
  in
   unlines (map (showRow h) [firstRow .. lastRow])

alignmentPath (SmithWatermanState {..}) =
  let
    elt00 = h ! (0,0)
    allElts = assocs h

    -- Start with position (0,0) and item h ! (0,0)
    -- and track down the highest scoring entry ( (i,j), h ! (i,j) )
    (bestIx@(bestI, bestJ), bestScore) =
      (foldl'
       (\ soFar@(_, bestScoreSoFar) next@(_, nextScore) ->
         if nextScore > bestScoreSoFar
         then next
         else soFar)
       ((0,0), elt00)
       allElts)

    bestPath = bestPathLoop bestI bestJ

    bestPathLoop i j =
      let
        step = path ! (i, j)
      in
       if step == start
       then [(i, j, Start)]
       else
         if step == northwest
         then (i, j, Step (a ! i) (b ! j)) : bestPathLoop (i - 1) (j - 1)
         else
            if step == north
            then (i, j, Delete (a ! i)) : bestPathLoop (i - 1) j
            else
              if step == west
              then (i, j, Insert (b ! j)) : bestPathLoop i (j - 1)
              else
                error (printf "SmithWaterman.alignmentDetails: bad path array item %s at %s"
                       (show step) (show (i,j)))
  in
   (bestIx, bestScore, bestPath)


prettyShowAlignment letterWidth letterToString s@(SmithWatermanState {..}) =
  let
    ((iEnd, jEnd), bestScore, bestPath) = alignmentPath s
    rSteps = reverse bestPath
    (i0, j0, Start) = head rSteps
    leftPad = max i0 j0
    topString = map topSymbol rSteps
    bottomString = map bottomSymbol rSteps

    topSymbol (i, j, Start) =
      replicate (letterWidth * (leftPad - i)) ' ' ++ concatMap letterToString (take i a0)
    topSymbol (_, _, Step x y) = letterToString x
    topSymbol (_, _, Delete x) = letterToString x
    topSymbol (_, _, Insert y) = replicate letterWidth '-'

    bottomSymbol (i, j, Start) =
      replicate (letterWidth * (leftPad - j)) ' ' ++ concatMap letterToString (take j b0)
    bottomSymbol (_, _, Step x y) = letterToString y
    bottomSymbol (_, _, Delete x) = replicate letterWidth '-'
    bottomSymbol (_, _, Insert y) = letterToString y
  in
   (concat
    (concat
     [topString,
      map letterToString (drop iEnd a0)]),
    concat
    (concat
     [bottomString,
      map letterToString (drop jEnd b0)]) )

-- An example --

data Nucleotide = T | C | A | G
                deriving (Read, Show, Eq, Ord, Enum)

ex1, ex2, ex3, ex4 :: [Nucleotide]

ex1 = [A,C,A,C,A,C,T,A]

ex2 = [A,G,C,A,C,A,C,A]

ex3 = [A,G,C,A,C,A,A,A]

ex4 = [A,G,C,A,C,A,T,A]

doEx s1 s2 =
  let
    wEx x y = if x == y then 2 else -1
    SmithWatermanResult {..} = alignSmithWaterman (-1) wEx s1 s2
    SmithWatermanState {..} = workspace
    scoreTableDisp = showMatrix h
    pathTableDisp = showMatrix path
    (topStr, bottomStr) = prettyShowAlignment 1 show workspace
  in do
    putStrLn scoreTableDisp
    putStrLn "==="
    putStrLn pathTableDisp
    putStrLn "==="
    print (bestIx, bestScore)
    putStrLn "==="
    print bestPath
    putStrLn "==="
    putStrLn topStr
    putStrLn bottomStr
