{-# LANGUAGE RecordWildCards, NamedFieldPuns, FlexibleContexts, DeriveDataTypeable, DeriveGeneric, TemplateHaskell #-}

{-|

Module: Digiostelium.StyledNetwork

Description: Tools for building visual representations of gene networks

Copyright: © 2019 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

-}

module Digiostelium.StyledNetwork (
  StyledNode(..),
  NodeStyle(..),
  StyledEdge(..),
  EdgeStyle(..),
  StyledNetwork(..),
  SignalToNode(..),
  networkToGraphMLDocument
  )

where

import GHC.Generics

import Data.Aeson.TH
import Data.Data
import Text.XML.Light
import BaselMachine.Signal

import Digiostelium.AesonUtil
import Digiostelium.GraphML


data StyledNode =
  StyledNode {
    nodeKey :: String,
    nodeLabel :: String,
    nodeStyle :: NodeStyle
    } deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)


data NodeStyle =
  OrdinaryNode
  | SensorNode
  | InputSourceNode
  | InputTargetNode
  | OutputSourceNode
  | OutputTargetNode
  | SynapseSourceNode
  | SynapseTargetNode
  | TransmitNode
  | StopNode
  | ReducedNode
  deriving (Read, Show, Eq, Ord, Enum, Data, Typeable, Generic)

data StyledEdge =
  StyledEdge {
    edgeKey :: String,
    fromKey :: String,
    toKey :: String,
    edgeLabel :: String,
    edgeStyle :: EdgeStyle
    } deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)

data EdgeStyle =
  ActivationEdge
  | InhibitionEdge
  deriving (Read, Show, Eq, Ord, Enum, Data, Typeable, Generic)

data StyledNetwork =
  StyledNetwork {
    networkKey :: String,
    networkLabel :: String,
    nodes :: [StyledNode],
    edges :: [StyledEdge],
    subnetworks :: [StyledNetwork]
    } deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)


$(deriveJSON myJSONOpts ''StyledNode)
$(deriveJSON myJSONOpts ''NodeStyle)
$(deriveJSON myJSONOpts ''StyledEdge)
$(deriveJSON myJSONOpts ''EdgeStyle)
$(deriveJSON myJSONOpts ''StyledNetwork)


class SignalToNode signal where
  -- | Build the 'StyledNode' representation of a @signal@
  signalToNode :: signal -> StyledNode


instance SignalToNode NullSignal where
  signalToNode NullSignal =
    StyledNode {
      nodeKey = "NullSignal",
      nodeLabel = "null sig",
      nodeStyle = ReducedNode }

nodeToGraphML :: StyledNode -> Element
nodeToGraphML StyledNode{..} =
  prettyNode nodeKey nodeLabel shape color
  where
   (shape, color) = nodeStyleToShapeAndColor nodeStyle

nodeStyleToShapeAndColor :: NodeStyle -> ([Char], [Char])
nodeStyleToShapeAndColor s =
  let
    pink = "#ffcccc"
    green = "#ccffcc"
    blue = "#ccccff"
    orange = "#ddddcc"
    purple = "#ddccdd"
    red = "#ff0000"
    yellow = "#ffcc00"
  in case s of
    OrdinaryNode -> ("ellipse", yellow)
    SensorNode -> ("hexagon", pink)
    InputSourceNode -> ("trapezoid2", green)
    InputTargetNode -> ("trapezoid2", blue)
    OutputSourceNode -> ("trapezoid", blue)
    OutputTargetNode -> ("trapezoid", green)
    SynapseSourceNode -> ("ellipse", orange)
    SynapseTargetNode -> ("ellipse", purple)
    TransmitNode -> ("roundrectangle", orange)
    StopNode -> ("octagon", red)
    ReducedNode -> ("diamond", pink)

edgeToGraphML :: StyledEdge -> Element
edgeToGraphML StyledEdge{..} =
  (prettyEdge
   edgeKey
   fromKey
   toKey
   (if null edgeLabel then Nothing else Just edgeLabel)
   (edgeStyle == ActivationEdge))

networkToGraphML :: StyledNetwork -> Element
networkToGraphML StyledNetwork{..} =
  (digraphElement
   networkKey
   (concat [map nodeToGraphML nodes,
            map edgeToGraphML edges,
            map embeddedNetworkToGraphML subnetworks]))

embeddedNetworkToGraphML :: StyledNetwork -> Element
embeddedNetworkToGraphML n@StyledNetwork{..} =
  unode "node"
  (["id" +> "NW:" ++ networkKey,
    "yfiles.foldertype" +> "group"],
   [unode "data"
    (["key" +> "NG"],
     node ("y" <:> "ProxyAutoBoundsNode")
     (node ("y" <:> "Realizers")
      (["active" +> "0"],
       [node ("y" <:> "GroupNode")
        [node ("y" <:> "Fill")
         ["color" +> "#f8ecc9"],
         node ("y" <:> "NodeLabel")
         (["alignment" +> "right",
           "autoSizePolicy" +> "node_width",
           "backgroundColor" +> "#404040",
           "textColor" +> "#ffffff",
           "fontSize" +> "16",
           "modelName" +> "internal",
           "modelPosition" +> "t",
           "x" +> "0.0",
           "y" +> "0.0"],
          networkLabel)],
        node ("y" <:> "GroupNode") () ]))),
    networkToGraphML n])

-- | Convert a 'StyledNetwork' to a complete XML document.
-- This adds a number of necessary scaffold elements.
networkToGraphMLDocument :: StyledNetwork -> Element
networkToGraphMLDocument network =
  scaffold (networkToGraphML network)
