{-# LANGUAGE NoMonomorphismRestriction, MultiParamTypeClasses, RecordWildCards, NamedFieldPuns, DeriveDataTypeable, DeriveGeneric, TemplateHaskell, FlexibleContexts, ParallelListComp, ScopedTypeVariables, OverloadedStrings #-}

module Main where

import Control.Monad.ST

import Data.Aeson
import Data.Aeson.Types
import Data.Foldable (toList)
import Data.IntMap (IntMap)
import qualified Data.IntMap as IM
import Data.IntSet (IntSet)
import qualified Data.IntSet as IS
import Data.List
import Data.Set (Set)
import qualified Data.Set as S
import Data.Sequence (Seq)
import qualified Data.Sequence as Seq
import qualified Data.Yaml as Y

import Text.Printf

import BaselMachine.Colony
import BaselMachine.Compiler
import BaselMachine.Instruction
import Digiostelium.NeuronProblem
import Digiostelium.QueryResult
import RandomColony
import System.Random.RandomST

m1 = "Agents/Agent-1580446.yaml"

-- Based on CD2F
cDist = ColonyDistribution {
  inputTargetPattern = 3,
  inputTargetDelta = 8,
  synapseSourcePattern = 41,
  synapseSourceThreshold = 2,
  synapseActivationDelta = -2,
  synapseTargetPattern = 3,
  numLayers = 1,
  numCellsPerLayer = 100,
  layerCellIndexOffset = 100,
  numDigits = 4,
  synapseDelayDist = [0 .. 12] :: [Int],
  synapseTargetDeltaDist =
    Alternative 0.2 ([-12 .. 12] :: [Int]) (Surely (0 :: Int))
  }

makeReservoir :: RandomST s ColonySpec
makeReservoir = do
  pullSample cDist


(reservoir, _) = withSeedFromInt 8403 $ makeReservoir

extractInstructions :: Value -> Result [(Int,Instruction)]
extractInstructions =
  parse (withObject "" f)
  where
    f x = do
      i <- x .: "allEfficaciousInstructions"
      parseJSON i


withM input = do
  Right (v :: Value) <- Y.decodeFileEither m1
  case extractInstructions v of
    Error msg ->
      error msg
    Success ixedInstructions -> do
      -- mapM_ putStrLn (map printNumberedInstruction ixedInstructions)
      let progAndInfo = compileUnchecked (map snd ixedInstructions)
          (_, _, output) = runColony reservoir (const False) progAndInfo input
      return output

extractTrace colonyOutput j =
  boolsToDrawing $ map (j `elem`) $ toList colonyOutput

makeTracePicture :: Seq [Int] -> String
makeTracePicture colonyOutput =
  intercalate "\n" [
    (printf "%2d %s" j (extractTrace colonyOutput j))
    | j <- IS.toList activeNodeIxs]
  ++
  "\n"
  ++
  edgesToMathematica (toList activeLinks)
  where
    (activeNodeIxs, activeLinks) = reduceNetwork colonyOutput


runM input = do
  y <- withM input
  putStrLn (makeTracePicture y)

makePulses nPulses js =
  intercalate (replicate 32 []) (replicate 3 pulses)
  where
    pulse = replicate 4 js ++ replicate 8 []
    pulses = concat $ replicate nPulses pulse

test4 nPulses js = do
  runM (makePulses nPulses js)


makeAlternation nPulses js1 js2 =
  intercalate (replicate 32 []) (replicate 3 pulses)
  where
    pulse = [js1,js1,js2,js2] ++ replicate 8 []
    pulses = concat $ replicate nPulses pulse


test5 nPulses js1 js2 = do
  runM (makeAlternation nPulses js1 js2)

addClock clockIxs clockPeriod clockDuration input =
  loop (clockDuration - 1) input
  where
    loop _ [] = []
    loop n (js : moreJs) =
      js' : loop n' moreJs
      where
        js' =
          if n < clockDuration
          then clockIxs ++ js
          else js
        n' =
          if n == 0
          then clockPeriod - 1
          else n - 1

reduceNetwork :: Seq [Int] -> (IntSet, Set (Int, Int))
reduceNetwork colonyOutput =
  (activeNodeIxs, activeLinks)
  where
    activeNodeIxs :: IntSet
    activeNodeIxs = IS.unions $ map IS.fromList $ toList colonyOutput
    activeLinks =
      S.fromList [
        (pre, post)
        | (pre, post, _) <- synapseSpecs reservoir,
          (IS.member pre activeNodeIxs
           && IS.member post activeNodeIxs)]

edgesToMathematica edges =
  concat [
    "{",
    intercalate "," [show pre ++ "->" ++ show post | (pre,post) <- edges],
    "}" ]
