{-# LANGUAGE GADTs #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE
RecordWildCards,
NamedFieldPuns,
DeriveDataTypeable,
DeriveGeneric,
ScopedTypeVariables,
ConstraintKinds,
FlexibleContexts,
OverloadedStrings,
Arrows
#-}

{-|
Module      : Copy21PhyloExtract

Description : Extract useful genes for doing phylogenetics experiments

Copyright   : © 2019 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

Extract genes from a sample run.  This is all specialized for my
normal Copy21 setup with beacon.

-}

module Main where

import GHC.Generics

import Control.Applicative
import Control.Exception
import Control.Monad
import Control.Monad.Trans

import Data.Aeson
import Data.Data
import qualified Data.Map as Map
import Data.Maybe
import qualified Data.Set as Set
import Data.Semigroup ( (<>) )
import Options.Applicative
import Options.Applicative.Arrows

import System.Exit
import System.FilePath
import System.IO

import Text.Printf

import BaselMachine.Instruction

import Data.Dynamic
import Data.Monoid (Sum(..) )
import Data.BitString

import Digiostelium.Agent
import Digiostelium.CopyProblem
import Digiostelium.Database
import Digiostelium.Genome
import Digiostelium.InputOutputProblem
import Digiostelium.Job
import Digiostelium.Problem
import Digiostelium.ProblemInventory
import Digiostelium.Query
import Digiostelium.SingletonProblem

import Digiostelium.EngineUtil

data CPEError =
  WrongProblemError String
  deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)

instance Exception CPEError

-- | Command line.
data PhyloExtractJob =
  PhyloExtractJob {
    sampleDir :: FilePath,
    minRating :: Sum Double
    }
  deriving (Show)

pejobParser :: ParserInfo PhyloExtractJob
pejobParser =
  (info pejobOpts
   (progDesc "Copy21-problem-specific phylogenetic information extraction; give --help for more help"))

pejobOpts :: Parser PhyloExtractJob
pejobOpts = runA $ proc () -> do
  sampleDir <-
    (asA $ argument str
     (metavar "SAMPLE_DIR"
      <> help "Data directory storing sample run")
    ) -< ()
  mMinRating <-
    (asA $ optional $ option auto
     (long "min-rating"
      <> metavar "RATING"
      <> help "Consider all agents with at least this rating")
    ) -< ()
  let minRating = Sum (fromMaybe 470320 mMinRating)
  returnA -< PhyloExtractJob{..}


-- | Main routine.  Load a sample run of an Copy21 problem and
-- examine each agent with the given minimum rating.  In the
-- sample run's directory, for each such agent, save the genes
-- for each gene key in a file named @Agent/{agent ID
-- number}/-/{gene key}/.csv@.

main :: IO ()
main = do
  PhyloExtractJob{..} <- execParser pejobParser

  let environmentDirectory = sampleDir
      databaseSpec = DatabaseSpec{..}
      configFile = sampleDir </> "Config.yaml"
      genomeFile = sampleDir </> "Genome.yaml"
      problemFile = sampleDir </> "Problem.yaml"
  assertDirectoryExists sampleDir
  Config{..} <- checkedLoadYAML configFile
  Box problem0 <- loadProblem problemFile problemName
  Box genomeSpec <- loadGenomeSpec genomeFile genomeArchitecture

  let dProblem = toDyn problem0
  problem :: CopyProblem <- do
    case fromDynamic dProblem of
      Just x -> return x
      Nothing -> throwIO (WrongProblemError "Wrong problem; CopyProblem required")


  withGlobalLock sampleDir $ withDatabase databaseSpec $ withPopCacheX genomeSpec problem $ do
    aMap <- doSelect (RatedAtLeast (toJSON minRating))
    when (Map.null aMap) $ liftIO $ do
      hPrintf stderr "In directory ‘%s’ found no agents with rating at least %f\n"
        sampleDir
        (getSum minRating)
      exitFailure
    forM_ (Map.elems aMap) $ \a -> do
      let geneAscs = analyzeAgent genomeSpec problem a
      forM_ geneAscs $ \(key, genes) -> do
        let gFile =
              (environmentDirectory
               </> printf "A%d-%s.txt" (identifier a) key)
            content = unlines (map bsTo01String genes)
        liftIO $ do
          writeFile gFile content
  return ()

-- | (Subroutine) Given a 'GenomeSpec', 'Problem', and 'Agent',
-- return a list of pairs (gene key, genes from the agent that
-- fit that key).

analyzeAgent ::
  (QueryIndividual genomeSpec problem SingletonProblemReportSpec,
   QueryResult genomeSpec problem SingletonProblemReportSpec
   ~ SingletonProblemReport (Rating problem) (Input problem) (Output problem)) =>
  genomeSpec
  -> problem
  -> Agent (Genome genomeSpec) (Rating problem)
  -> [(String, [BitString])]

analyzeAgent genomeSpec problem a@Agent{egg=Egg{..}, ..} = do
  (key, check) <- copy21Elements
  let relevantGenes =
        [nucleotides
        | (j, (instruction, nucleotides)) <- ixedInstructionDetails,
          check (j, instruction)]
  return (key, relevantGenes)
  where
    -- I originally did this with the whole genome.  Need to be
    -- smarter about this... Some agents have genes that refer to
    -- the pattern of interest but are not the genes I'm looking
    -- for (not efficacious)

    EfficaciousReport{..} = queryIndividual genomeSpec problem Efficacious a
    effInstIxs = Set.fromList [j | (j, _) <- allEfficaciousInstructions]
    ixedInstructionDetails = [ (j, (instruction, nucleotideBits))
      | (j, GeneDetails{..}) <- zip [0..] (geneDetails genomeSpec genome),
        j `Set.member` effInstIxs ]


-- | This is a table of gene keys and tests that define them.
-- They are basically the names of different functions genes
-- might have.  They are
--
-- * B: beacon gene, activates pattern 31 (the output beacon)
--
-- * P0, P1: gene switched by patterns 8 or 9 (where the sender receives input)
--
-- * S: gene targets pattern 3 (sender side of synapse)
--
-- * R: gene switched by pattern 4 (receiver side of synapse)
--
-- * Q0, Q1: gene targets patterns 12 or 13 (where the receiver generates output)
--
-- * H: gene targets pattern 0 (receiver's stop signal, as in halt)

copy21Elements :: [(String, (Int, Instruction) -> Bool)]
copy21Elements = [
  ("B", instTMatch 31),
  ("P0", instSMatch 8),
  ("P1", instSMatch 9),
  ("S", instTMatch 3),
  ("R", instSMatch 4),
  ("Q0", instTMatch 12),
  ("Q1", instTMatch 13),
  ("H", instTMatch 0)]



-- | Return whether the given instruction is switched on or off
-- by the given pattern.

instSMatch ::
  Int -- ^ switch pattern to test
  -> (Int, Instruction) -- ^ program index and instruction
  -> Bool

instSMatch s (_j, instruction) =
  switchPattern == s
  where
    Instruction{..} = instruction
    --reactions = splitInstruction (j, instruction)



-- | Return whether the given instruction includes a reaction
-- that targets the given pattern

instTMatch ::
  Int -- ^ target pattern to test
  -> (Int, Instruction) -- ^ program index and instruction
  -> Bool

instTMatch t (j, instruction) =
  t `elem` targetPatterns
  where
    Instruction{..} = instruction
    reactions = splitInstruction (j, instruction)
    targetPatterns = map tPat reactions
