{-# LANGUAGE
RecordWildCards,
NamedFieldPuns,
DeriveDataTypeable,
DeriveGeneric,
ScopedTypeVariables,
ConstraintKinds,
FlexibleContexts,
OverloadedStrings,
TemplateHaskell
#-}

{-|
Module      : Engine

Description : Main simulation engine

Copyright   : © 2019 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

-}

module Main where

import GHC.Generics

import Language.Haskell.TH

import Control.Concurrent
import Control.DeepSeq
import Control.Exception
import Control.Monad
import Control.Monad.State
-- import Control.Monad.Trans

import Data.Aeson hiding (Array)
import qualified Data.Aeson as J
import qualified Data.Aeson.Types as JT
import Data.Binary (Binary)
import qualified Data.ByteString as BS
-- import Data.ByteString.Builder
-- import qualified Data.ByteString.Lazy as BSL
-- import qualified Data.ByteString.Lazy.UTF8 as LU
import Data.Data
import Data.Default
import Data.Either
-- import Data.Foldable
import Data.List (find)
import Data.Map (Map)
import qualified Data.Map as Map
import qualified Data.Sequence as Seq
import Data.Time
--import Data.Typeable
import Data.Word
import qualified Data.Yaml as Yaml
import Options.Applicative
-- import Options.Applicative.Arrows
import System.Directory
-- import System.Environment
-- import System.Exit
import System.FilePath
import System.IO
-- import qualified System.Log.Handler as LH
-- import qualified System.Log.Handler.Simple as LHS
-- import System.Log.Formatter
import System.Log.Logger
-- import System.Posix.Signals
import Text.Printf

import Digiostelium.Agent
import Digiostelium.BasicReport (BasicReport)
import qualified Digiostelium.BasicReport as BasicReport
import Digiostelium.BlockSize
import Digiostelium.Database
import Digiostelium.Evolution
import Digiostelium.GeneralName
import Digiostelium.Genome
import Digiostelium.Job
import Digiostelium.Logging
import Digiostelium.PopulationDatabase
import Digiostelium.PopulationSpec
import Digiostelium.PredicateTools
import Digiostelium.Problem
import Digiostelium.ProblemInventory
import Digiostelium.Query
-- import Digiostelium.Report
import WGM.MonadUtil
import System.Random.RandomST

import Digiostelium.EngineUtil

-- A debugging message: Compile time of this executable
compileTimeStamp :: String
compileTimeStamp = $(stringE =<< runIO (show `fmap` getCurrentTime))

compileLocalTimeStamp :: String
compileLocalTimeStamp = $(stringE =<< runIO
                           (do
                               tz <- getCurrentTimeZone
                               currentUTC <- getCurrentTime
                               let currentLocal = utcToLocalTime tz currentUTC
                               return (formatTime defaultTimeLocale "%c" currentLocal)))

targetAchievedFilePath :: FilePath
targetAchievedFilePath = "TargetAchieved.txt"

main :: IO ()
main = do
  (Args cOpts@CommonOpts{..} cmd) <- execParser (jobParser ("Compiled at " ++ compileTimeStamp ++ " (" ++ compileLocalTimeStamp ++ ")" ))
  -- To start with, the global logger notices everything, but its
  -- only handler sends WARNING and higher messages to stderr
  wHandler <- logToHandle stderr WARNING
  updateGlobalLogger rootLoggerName (
    setLevel DEBUG
    . setHandlers [wHandler]
    )
  -- Do work
  (doCmd cOpts cmd
   `catches`
   [Handler (handleTargetAchievedEx (sampleDir cmd)),
    Handler handleConfigurationEx])


doCmd :: CommonOpts -> Command -> IO ()

doCmd _cOpts List = do
  printf "Problem templates:\n"
  showKeys (Map.keys problemTemplates)
  printf "Genome templates:\n"
  showKeys (Map.keys genomeSpecTemplates)
  where
    prefix = " "
    showKeys keys = do
      forM_ keys $ \key -> do
        putStr prefix
        putStrLn key

doCmd _cOpts Template{..} = do
  createDirectoryIfMissing True sampleDir
  Box prob <- getProblemTemplate problemTemplateName
  Box (genomeSpec :: gs) <- getGenomeSpecTemplate genomeArchitectureTemplateName
  let problemName = generalName prob
      genomeArchitecture = generalName genomeSpec
      archive = Never
      target = def
      config = Config{..}
  yamlEncodeFileSafe (sampleDir </> "Alignment.yaml") (setBlockSize (getBlockSize genomeSpec) (def :: AlignmentSpec gs))
  yamlEncodeFileSafe (sampleDir </> "Config.yaml") config
  yamlEncodeFileSafe (sampleDir </> "Genome.yaml") genomeSpec
  yamlEncodeFileSafe (sampleDir </> "Problem.yaml") prob
  yamlEncodeFileSafe (sampleDir </> "Selection.yaml") (def :: SelectionSpec)

doCmd CommonOpts{..} Init{..} = do
  let environmentDirectory = sampleDir
      databaseSpec = DatabaseSpec{..}
      alignmentFile = sampleDir </> "Alignment.yaml"
      configFile = sampleDir </> "Config.yaml"
      genomeFile = sampleDir </> "Genome.yaml"
      problemFile = sampleDir </> "Problem.yaml"
      selectionFile = sampleDir </> "Selection.yaml"
  createDirectoryIfMissing True environmentDirectory
  quietRemoveFile (sampleDir </> targetAchievedFilePath)
  sendLogToFile (sampleDir </> "Log.txt") logLevel
  Config{..} <- checkedLoadYAML configFile
  Box problem <- loadProblem problemFile problemName
  Box genomeSpec <- loadGenomeSpec genomeFile genomeArchitecture
  alignmentSpec <- checkedLoadYAML alignmentFile
  selectionSpec <- checkedLoadYAML selectionFile
  let populationSize = resultingPopulationSize selectionSpec
      populationSpec = PopulationSpec{..}
  (genomes, randomSeed) <- case (identical, importFrom) of
    (False, "") -> do
      seed0 <- createRandomSeed
      let (genomes, randomSeed) = withSeed seed0 $ do
            pullSampleList populationSize genomeSpec
      return (genomes, randomSeed)
    (False, importPath) -> do
      assertFileNotEmpty importPath
      importBytes <- BS.readFile importPath
      let genomes = checkImportGenomes genomeSpec problem importBytes
      randomSeed <- createRandomSeed
      return (genomes, randomSeed)
    (True, "") -> do
      seed0 <- createRandomSeed
      let (prototype, randomSeed) = withSeed seed0 $ do
            pullSample genomeSpec
      return (replicate populationSize prototype, randomSeed)
    (True, path) -> do
      throwIO $ ConflictingSettings {
        message = "Request for identical starting agents conflicts with request to import from " ++ path
        }
  deepseq genomes $ withGlobalLock sampleDir $ initializeFirstGeneration populationSpec problem genomes randomSeed archive


doCmd CommonOpts{..} Reset{..} = do
  let environmentDirectory = sampleDir
      databaseSpec = DatabaseSpec{..}
      alignmentFile = sampleDir </> "Alignment.yaml"
      configFile = sampleDir </> "Config.yaml"
      genomeFile = sampleDir </> "Genome.yaml"
      problemFile = sampleDir </> "Problem.yaml"
      selectionFile = sampleDir </> "Selection.yaml"
  assertDirectoryExists environmentDirectory
  quietRemoveFile (sampleDir </> targetAchievedFilePath)
  sendLogToFile (environmentDirectory </> "Log.txt") logLevel
  Config{..} <- checkedLoadYAML configFile
  Box problem <- loadProblem problemFile problemName
  Box genomeSpec <- loadGenomeSpec genomeFile genomeArchitecture
  alignmentSpec <- checkedLoadYAML alignmentFile
  selectionSpec <- checkedLoadYAML selectionFile
  let populationSize = resultingPopulationSize selectionSpec
      _populationSpec = PopulationSpec{..}
      initRating = minLandmarkRating problem
      recordGenerations =
        case archive of
          Never -> False
          _ -> True
  ir <- (reinitializePopulationDatabase
         environmentDirectory initRating
         recordGenerations archive)
  let InitializationRec{..} = ir
  withGlobalLock sampleDir $ withDatabase databaseSpec $ withPopCacheX genomeSpec problem $ do
      newEggsRatings <- (withCurrentRandomSeed
                        (rateEggs genomeSpec problem Map.empty
                         [Egg [] g | g <- genomes]))
      -- let newEggsRatings = rateEggs genomeSpec problem Map.empty [Egg [] g | g <- genomes]
      let delta = PopulationBirthDeathDelta{toKill = [], ..}
      applyBirthDeathDelta delta
      nextTime
  return ()


doCmd CommonOpts{..} Run{..} = do
  let environmentDirectory = sampleDir
      databaseSpec = DatabaseSpec{..}
      alignmentFile = sampleDir </> "Alignment.yaml"
      configFile = sampleDir </> "Config.yaml"
      genomeFile = sampleDir </> "Genome.yaml"
      problemFile = sampleDir </> "Problem.yaml"
      selectionFile = sampleDir </> "Selection.yaml"
  assertDirectoryExists sampleDir
  quietRemoveFile (sampleDir </> targetAchievedFilePath)
  sendLogToFile (sampleDir </> "Log.txt") logLevel
  Config{..} <- checkedLoadYAML configFile
  Box problem <- loadProblem problemFile problemName
  Box genomeSpec <- loadGenomeSpec genomeFile genomeArchitecture
  alignmentSpec <- checkedLoadYAML alignmentFile
  selectionSpec <- checkedLoadYAML selectionFile
  let populationSize = resultingPopulationSize selectionSpec
      populationSpec = PopulationSpec{..}

  t <- myThreadId
  checkSig <- makeSigStopper (Just (t, UserInterrupt))
  checkMaxSecondsClock <- maybeHelperIO maxSeconds (makeClockChecker MaxSecondsElapsed)

  createDirectoryIfMissing True environmentDirectory

  -- Hard-coded during development; move to config file soon
  let snapshotSpec = SnapshotSpec {
        snapshotSeconds = 60*15, -- every 15 minutes
        snapshotMaxCount = 3
        }

  -- Main loop:
  -- Return cleanly if time runs out by catching and discarding exception MaxSecondsElapsed.
  -- Return cleanly if a signal was sent to this process and the loop had time to notice via checkSig.
  -- When snapshot clock runs out, the exception system exits the inner loop, and closes the database;
  -- exception handler for SnapshotSecondsElapsed makes a snapshot;
  -- and re-enter the inner loop.
  _ <- handle (\MaxSecondsElapsed -> return ()) $ withGlobalLock sampleDir $ forever $ do
    checkSnapshotClock <- makeSnapshotChecker snapshotSpec
    handle (\SnapshotSecondsElapsed -> makeSnapshot databaseSpec snapshotSpec) $
      withDatabase databaseSpec $ withPopCache $ forever $ do -- inner loop
        quickReport
        writeStatus
        checkSig
        checkMaxSecondsClock
        checkSnapshotClock
        checkTarget genomeSpec problem target
        fullProblemNextGeneration populationSpec problem
  return ()

-- Note that Continue{..} jobs don't remove the 'targetAchievedFilePath'
doCmd CommonOpts{..} Continue{..} = do
  let environmentDirectory = sampleDir
      databaseSpec = DatabaseSpec{..}
      alignmentFile = sampleDir </> "Alignment.yaml"
      configFile = sampleDir </> "Config.yaml"
      genomeFile = sampleDir </> "Genome.yaml"
      problemFile = sampleDir </> "Problem.yaml"
      selectionFile = sampleDir </> "Selection.yaml"
  assertDirectoryExists sampleDir
  sendLogToFile (sampleDir </> "Log.txt") logLevel
  Config{..} <- checkedLoadYAML configFile
  Box problem <- loadProblem problemFile problemName
  Box genomeSpec <- loadGenomeSpec genomeFile genomeArchitecture
  alignmentSpec <- checkedLoadYAML alignmentFile
  selectionSpec <- checkedLoadYAML selectionFile
  let populationSize = resultingPopulationSize selectionSpec
      populationSpec = PopulationSpec{..}

  assertDirectoryExists environmentDirectory
  t <- myThreadId
  checkSig <- makeSigStopper (Just (t, UserInterrupt))

  withGlobalLock sampleDir $ withDatabase databaseSpec $ withPopCache $ do
    n <- getCurrentTimeStep
    let goalGenerations = n + moreGenerations
        checkDone = do
          t <- getCurrentTimeStep
          (msgDebug
           (printf
            "Checking current generation number %d against goal %d"
            t goalGenerations))
          when (goalGenerations <= t) $ liftIO $ throwIO MaxGenerationsReached

    untilTrue $ do
      quickReport
      writeStatus
      checkDone
      checkSig
      fullProblemNextGeneration populationSpec problem
      return False
  return ()



doCmd CommonOpts{..} Query{..} = do
  let environmentDirectory = sampleDir
      databaseSpec = DatabaseSpec{..}
      configFile = sampleDir </> "Config.yaml"
      genomeFile = sampleDir </> "Genome.yaml"
      problemFile = sampleDir </> "Problem.yaml"
  assertDirectoryExists sampleDir
  sendLogToFile (sampleDir </> "Log.txt") logLevel
  Config{..} <- checkedLoadYAML configFile
  Box problem <- loadProblem problemFile problemName
  Box genomeSpec <- loadGenomeSpec genomeFile genomeArchitecture

  aMap <- withGlobalLock sampleDir $ withDatabase databaseSpec $ withPopCacheX genomeSpec problem $ do
    doSelect select

  case boxReportSpec genomeSpec problem reportSpec aMap of
    Just reportBox -> do
      liftIO $ writeReportsBox reportBox outputFormat outputDestination
    Nothing -> do
      throw ImpossibleReport {
        message = printf "Report is impossible: %s" (show reportSpec)
        }


initializeFirstGeneration ::
  (UsefulProblem prob,
   UsefulGenomeSpec gs) =>
  PopulationSpec gs
  -> prob
  -> [Genome gs]
  -> [Word32]
  -> CompoundMPredicate ArchivePredicate
  -> IO ()
initializeFirstGeneration PopulationSpec{..}
  problem genomes randomSeed archivePredicate =
  let
    -- problemName = generalName problem
    recordGenerations =
      case archivePredicate of
        Never -> False
        _ -> True
    initRec = InitializationRec{..}
    initRating = minLandmarkRating problem
  in do
    initializePopulationDatabase databaseSpec initRating initRec
    withDatabase databaseSpec $ withPopCacheX genomeSpec problem $ do
      newEggsRatings <- (withCurrentRandomSeed
                        (rateEggs genomeSpec problem Map.empty
                         [Egg [] g | g <- genomes]))
      -- let newEggsRatings = rateEggs genomeSpec problem Map.empty [Egg [] g | g <- genomes]
      let delta = PopulationBirthDeathDelta{toKill = [], ..}
      applyBirthDeathDelta delta
      nextTime


data ImportException =
  ImportException String
  deriving (Show, Eq, Typeable, Data)

instance Exception ImportException

-- | The problem argument is a dummy to get the types to work out
checkImportGenomes ::
  forall gs prob.
  (UsefulGenomeSpec gs,
   UsefulProblem prob) =>
  gs -> prob -> BS.ByteString -> [Genome gs]
checkImportGenomes gs _problem bytes =
  if null errors
  then successes
  else
    throw (ImportException (head errors))
  where
    exports :: Map Integer (BasicReport (Genome gs) (Rating prob))
    exports = case Yaml.decodeEither' bytes of
      Left exc ->
        throw exc
      Right exports ->
        exports
    parseAttempts =
      [JT.parseEither (importGenomeDataParser gs) genome
        | BasicReport.ExportReport{..} <- Map.elems exports]
    errors = lefts parseAttempts
    successes = rights parseAttempts


-- | Exceptions thrown when a run reaches is target.
data TargetAchievedException =
  MaxGenerationsReached
  | LandmarkBornReached
  | LandmarkDiedReached
  deriving (Read, Show, Eq, Data, Typeable, Generic)

instance Exception TargetAchievedException

handleTargetAchievedEx :: FilePath -> TargetAchievedException -> IO ()
handleTargetAchievedEx sampleDir ex = do
  let logMsg = case ex of
        MaxGenerationsReached ->
          "Reached maximum number of generations"
        LandmarkBornReached ->
          "Requested landmark agent has been born"
        LandmarkDiedReached ->
          "Requested landmark agent has died"
  msgInfo logMsg
  writeFile (sampleDir </> targetAchievedFilePath) logMsg



-- | Throw a 'TargetAchievedException' if one of the targets has
-- been achieved.
checkTarget ::
  forall gs prob.
  (UsefulGenomeSpec gs,
   UsefulProblem prob) =>
  gs
  -> prob
  -> Target
  -> PopMonad (Genome gs) (Rating prob) ()

checkTarget _genomeSpec _problem Target{..} = do

  case maxGenerations of
    Nothing -> return ()
    Just tLast -> do
      t <- getCurrentTimeStep
      (msgDebug
       (printf
        "Checking current generation number %d against max generations %d"
        t tLast))
      when (tLast <= t) $ liftIO $ throwIO MaxGenerationsReached

  case landmarkBorn of
    Nothing -> return ()
    Just targetRatingV ->
      case fromJSON targetRatingV of
        J.Success targetRating -> do
          (msgDebug
           (printf
            "Checking for landmark born with rating >= %s"
            (show targetRating)))
          withLandmarkRatedMoreThan targetRating $ \agent -> do
            (msgDebug
             (printf "Found target agent (ID %s)" (show (identifier agent))))
            liftIO $ throwIO $ LandmarkBornReached
        J.Error msg ->
            liftIO $ throwIO $ BadRating {
              badRating = targetRatingV,
              parseErrorMessage = msg
              }

  case landmarkDied of
    Nothing -> return ()
    Just targetRatingV ->
      case fromJSON targetRatingV of
        J.Success targetRating -> do
          (msgDebug
           (printf
            "Checking for landmark died with rating >= %s"
            (show targetRating)))
          withLandmarkRatedMoreThan targetRating $ \agent -> do
            case lifespan agent of
              Left _ -> return ()
              Right (_birthTime, _deathTime) -> do
                (msgDebug
                 (printf "Found target agent (ID %s)" (show (identifier agent))))
                liftIO $ throwIO $ LandmarkDiedReached
        J.Error msg ->
            liftIO $ throwIO $ BadRating {
              badRating = targetRatingV,
              parseErrorMessage = msg
              }


withLandmarkRatedMoreThan ::
  (Binary genome,
   Binary rating,
   Ord rating) =>
  rating
  -> (Agent genome rating -> StateT (PopCache genome rating) (DBMonad TableName) ())
  -> StateT (PopCache genome rating) (DBMonad TableName) ()

withLandmarkRatedMoreThan targetRating axn = do
  landmarks <- getLandmarks
  let mTargetLandmark = find (\x -> rating x >= targetRating) (Map.elems landmarks)
  case mTargetLandmark of
    Nothing -> return ()
    Just agent ->
      axn agent


quickReport ::
  (Show rating) =>
  PopMonad genome rating ()
quickReport = do
  seed <- getCurrentRandomSeed
  t <- getCurrentTimeStep
  n <- getCurrentIdentifier
  r <- getBestRating
  msgInfo ("At time " ++ show t ++ " agent #" ++ show n
           ++ " best rating so far: " ++ show r)
  msgDebug ("random seed: " ++ show (take 4 seed))

writeStatus ::
  (ToJSON rating,
  Binary rating,
  Binary genome) =>
  PopMonad genome rating ()

writeStatus = do
  t <- getCurrentTimeStep
  n <- getCurrentIdentifier
  -- r <- getBestRating
  lm <- getLandmarks
  genClockTimes <- getGenerationClockTimes
  nowUTC <- liftIO getCurrentTime
  thisTimeZone <- liftIO getCurrentTimeZone
  let nowLocal = utcToLocalTime thisTimeZone nowUTC
  let (newestLandmarkID, Agent{..}) = Map.findMax lm
      avgGenClockTime :: Double
      avgGenClockTime =
        if Seq.null genClockTimes
        then -1.0
        else
          sum genClockTimes / fromIntegral (Seq.length genClockTimes)
      rObj = object [
        "generation" .= t,
        "nextAgentID" .= n,
        -- "bestRatingSoFar" .= r,
        "newestLandmark" .= (newestLandmarkID, rating),
        "timeStamp" .= show nowLocal,
        "avgGenClockTime" .= avgGenClockTime
        ]
  (d :: Database TableName) <- lift get
  let Database{dbSpec = DatabaseSpec{environmentDirectory} } = d
  liftIO $ Yaml.encodeFile (environmentDirectory </> "Status.yaml") rObj





-- | Advance to the next generation

fullProblemNextGeneration ::
  (UsefulGenomeSpec gs,
   UsefulProblem prob)
  =>
  PopulationSpec gs
  -> prob
  -> PopMonad (Genome gs) (Rating prob) ()

fullProblemNextGeneration PopulationSpec{..} problem =
  nextGeneration selectionSpec genomeSpec problem



-- | (Private) Utility for removing a file quietly
quietRemoveFile :: FilePath -> IO ()
quietRemoveFile path = do
  v <- doesFileExist path
  when v $ removeFile path
