{-# LANGUAGE RecordWildCards, NamedFieldPuns, DeriveDataTypeable,
DeriveGeneric, ScopedTypeVariables, ConstraintKinds,
FlexibleContexts, OverloadedStrings, Arrows #-}

{-|
Module      : HammingMatrix

Description : Extract useful genes for doing phylogenetics experiments

Copyright   : © 2019 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

-}

module Main where

import Control.Applicative
import Control.DeepSeq
import Control.Exception
import Control.Monad

import Data.Array.IArray
import Data.List
import Data.Map.Strict (Map)
import qualified Data.Map.Strict as Map
import Data.Maybe
import Data.Set (Set)
import Data.Semigroup ( (<>) )
import qualified Data.Set as Set
import Data.Typeable

import Options.Applicative
import Options.Applicative.Arrows

import System.Directory
import System.FilePath
import System.IO

import Text.Printf
import Text.Regex.PCRE

import Data.BitString

import WGM.ParUtil


data HammingMatrixJob =
  HammingMatrixJob {
    outDir :: FilePath,
    -- ^ diretory to write files into
    extrDirs :: [FilePath]
    -- ^ extraction directories, each of which is where genome
    -- data from one sample run has been extracted to
    }
  deriving (Show)


data JobException =
  FileEmptyException FilePath
  | GeneListEmptyException
  | NoGenomeFiles FilePath
  | NoGenomeFilesForKey String FilePath
  deriving (Show, Typeable)

instance Exception JobException

-- | Command line parser

hmjobParser :: ParserInfo HammingMatrixJob

hmjobParser =
  (info hmjobOpts
   (progDesc "\
\Given many directory paths to Copy21PhyloExtract runs on the command line, output the Hamming distance matrix."))


-- | Command line parser

hmjobOpts :: Parser HammingMatrixJob

hmjobOpts = runA $ proc () -> do
  outDir <-
    (asA $ argument str
     (metavar "DIR"
      <> help "Output directory where HM*.csv will be written")
    ) -< ()
  extrDirs <-
    (asA $ many $ argument str
     (metavar "DIRS"
      <> help "Copy21 run directories")
    ) -< ()
  returnA -< HammingMatrixJob{..}



-- | Given a file path, read each line of 0s and 1s as a
-- 'BitString'

read01StringFile :: FilePath -> IO [BitString]

read01StringFile filePath = do
  fileData <- readFile filePath
  let fileLines = lines fileData
      genes = map bsFrom01String fileLines
  deepseq genes $ return genes


-- | (Subroutine) Given a gene key and a file path, if the file path
-- is of the correct form for where the gene given by the key
-- from an agent should have been extracted to, return 'Just' the
-- file contents as a list of 'BitString'.  Otherwise, 'Nothing'.
-- This is used by 'loadGenomeFiles' to filter out extracted gene
-- files by key from a list of all files in a directory.

loadAgentFile ::
  String -- ^ gene key
  -> FilePath -- ^ file path
  -> IO (Maybe [BitString])

loadAgentFile key path =
  let
      rx = "A([0-9]+)-" ++ key
      stem = takeBaseName path
      _before :: String
      _matching :: String
      _after :: String
      subgroups :: [String]
      (_before, _matching, _after, subgroups) =
        stem =~ rx
  in
   -- I don't actually need n here, I just need to know that the
   -- file stem fits the pattern of "A" followed by an agent ID,
   -- a hyphen -, then the gene key.  There's probably a simpler
   -- way to do this, but I don't want to change it right now...
   case subgroups of
     [nStr] ->
       let
         _n :: Integer
         _n = read nStr
       in do
         genes <- read01StringFile path
         when (null genes) $ do
           throwIO (FileEmptyException path)
         return (Just genes)
     _  ->
       return Nothing


-- | Given a gene key and a directory, load all files in the
-- directory that contain gene data from some agent for that gene
-- key.

loadGenomeFiles ::
  String -- ^ gene key
  -> FilePath -- ^ directory to search
  -> IO (Set [BitString])

loadGenomeFiles key dir = do
  allFiles <- getDirectoryContents dir
  when (null allFiles) $ do
    throw (NoGenomeFiles dir)
  agentDataMs <- mapM (loadAgentFile key) [dir </> p | p <- allFiles]
  let agentDatas = catMaybes agentDataMs
  when (null agentDatas) $ do
    throw (NoGenomeFilesForKey key dir)
  return (Set.fromList agentDatas)



-- | Given a list of item, return a 'Map' from items to how many
-- times they appeared in the list.

countItems ::
  Ord k =>
  [k]
  -> Map k Int

countItems xs =
  Map.unionsWith (+) [Map.singleton x 1 | x <- xs]



-- | Invert a map, that is, swap the roles of keys and values.

invertMap ::
  (Ord k, Ord v) =>
  Map k v
  -> Map v k

invertMap m =
  Map.fromList [(v, k) | (k, v) <- Map.toList m]


-- | Return the 'BitString' in the list that is closest in
-- Hamming distance to the reference 'BitString'

closestTo ::
  BitString -- ^ reference
  -> [BitString]
  -> BitString

closestTo _ [] = throw GeneListEmptyException

closestTo ref xs = closest
  where
    distances = [(bsHammingDistance ref x, x) | x <- xs]
    (_leastDistance, closest) = minimum distances



-- | Reduce a list of 'BitString's to the one closest to the
-- given reference string.  The result is a list, so this can be
-- optionally used as a way to refine the calculation of
-- distances between populations

reduceToClosest :: BitString -> [BitString] -> [BitString]
reduceToClosest ref genes = [closestTo ref genes]



-- | Return the gene whose total Hamming distance to all others
-- is least.

mostCentral ::
  [BitString]
  -> BitString

mostCentral genes = g
  where
    geneCounts = Map.toList (countItems genes)

    -- Make a list of (total distance to all other genes, gene)
    geneToTotalDistance =
      parListDeep [
        (sum [(bsHammingDistance gLeft gRight) * nRight
             | (gRight, nRight) <- geneCounts],
         gLeft)
        | (gLeft, _nLeft) <- geneCounts
        ]

    (_d, g) = minimum geneToTotalDistance



-- | Given lists of genes (presumably for the same key) from two
-- individuals, return the Hamming distance for the closest match
-- for one gene from each individual.

scoreGeneSets :: [BitString] -> [BitString] -> Maybe Int

scoreGeneSets [] _ = Nothing

scoreGeneSets _ [] = Nothing

scoreGeneSets aStrings bStrings =
  let
    hammingDistances = sort [
      minimum hds
      | a <- aStrings,
        let hds = [bsHammingDistance a b | b <- bStrings]]
  in case hammingDistances of
    (n:_) -> Just n
    [] -> Nothing


-- | The 'Set's of gene lists passed to this function mean: For
-- each individual in a population, include all of its genes for
-- a particular key.  Thus, each item in each set is the list of
-- relevant genes from one individual in a population.  Return
-- the Hamming score from 'scoreGeneSets' averaged over all
-- pairings of an individual from each set.  If an item is the
-- empty list, it's ignored and not included at all in figuring
-- the average.

scoreAgentSets ::
  Set [BitString]
  -> Set [BitString]
  -> Maybe Double

scoreAgentSets leftASet rightASet =
  if null scores
  then Nothing
  else Just (floatMean scores)

  where
    scoreMs :: [Maybe Int]
    scoreMs = do
      a <- Set.elems leftASet
      b <- Set.elems rightASet
      return (scoreGeneSets a b)

    scores = catMaybes scoreMs



-- | Given a lsit of 'Int's, compute their average as a 'Double'.

floatMean :: [Int] -> Double

floatMean xs =
  fromIntegral sumXs / fromIntegral n
  where
    n = length xs
    sumXs = sum xs


-- | Gene keys are the names of different functions genes might have.
-- They are
--
-- * B: beacon gene, activates the output beacon
--
-- * P0, P1: gene switched by patterns 8 or 9 (where the sender receives input)
--
-- * S: gene targets pattern 3 (sender side of synapse)
--
-- * R: gene switched by pattern 4 (receiver side of synapse)
--
-- * Q0, Q1: gene targets patterns 12 or 13 (where the receiver generates output)
--
-- * H: gene targets pattern 0 (receiver's stop signal, as in halt)
--
-- But for now, we only need B.

geneKeys :: [[Char]]
geneKeys = ["B"]

-- geneKeys = ["B", "P0", "P1", "S", "R", "Q0", "Q1", "H"]


-- | Main routine.  From the command line, get a
-- 'HammingMatrixJob'.  For each item in 'geneKeys', construct a
-- matrix of the result of 'scoreAgentSets' obtained by scoring
-- each pair of populations from 'extrDirs'.  The extracted files
-- are assumed to have file names as used by the
-- @Copy21PhyloExtract@ program, as in
-- @Agent/{agent ID number}/-/{gene key}/.csv@.
-- The matrices are written out in CSV format to files
-- in 'outDir' named @HM-/{gene key}/.csv@

main :: IO ()
main = do
  job@HammingMatrixJob{..} <- execParser hmjobParser
  onException (mainWork job) $ do
    hPutStrLn stderr "Exception caught"
    hPrint stderr job

mainWork :: HammingMatrixJob -> IO ()
mainWork HammingMatrixJob{..} = do
  forM_ geneKeys $ \key -> do
    runsDataAll <- forM extrDirs $ loadGenomeFiles key
    let allTogether :: [BitString]
        allTogether = concat (Set.toList (Set.unions runsDataAll))
        geneCounts = countItems allTogether
        (_, geneMostFrequent) = Map.findMax (invertMap geneCounts)
        geneMostCentral = mostCentral allTogether
        grandReferenceGene :: BitString
        grandReferenceGene = geneMostFrequent
        runsData :: [Set [BitString]]
        runsData = map (Set.map (reduceToClosest grandReferenceGene)) runsDataAll
    let nRuns = length runsData
        hammingScores =
          parListDeep [if j == k
           then [((j,k), x)]
           else [((j,k), x), ((k,j), x)]
            | j <- [0 .. nRuns - 1],
              k <- [j .. nRuns - 1],
              let leftRun = runsData !! j,
              let rightRun = runsData !! k,
              let x = scoreAgentSets leftRun rightRun]
        hammingMatrix :: Array (Int, Int) (Maybe Double)
        hammingMatrix =
          array ((0,0), (nRuns - 1, nRuns - 1)) (concat hammingScores)
        hammingRows = [
          intercalate "," [
             mShow (hammingMatrix ! (j, k))
             | k <- [0 .. nRuns - 1]]
          | j <- [0 .. nRuns - 1]]
        matOutPath = outDir </> (printf "HM-%s.csv" key)
        genesOutPath = outDir </> (printf "Genes-%s.xml" key)
    withFile genesOutPath WriteMode $ \h -> do
      hPutStrLn h "<?xml version='1.0'?>"
      hPutStrLn h "<populations>"
      hPutStrLn h "<reference>"
      hPutStr h "<gr>"
      hPutStr h (bsTo01String grandReferenceGene)
      hPutStrLn h "</gr>"
      hPutStr h "<gmf>"
      hPutStr h (bsTo01String geneMostFrequent)
      hPutStrLn h "</gmf>"
      hPutStr h "<gmc>"
      hPutStr h (bsTo01String geneMostCentral)
      hPutStrLn h "</gmc>"
      hPutStrLn h "</reference>"
      hPutStrLn h "<inventory>"
      let sortFn :: (BitString, Int) -> (BitString, Int) -> Ordering
          sortFn (_g1, c1) (_g2, c2) = compare c2 c1
      forM_ (sortBy sortFn (Map.toList geneCounts)) $ \(g, count) -> do
        hPrintf h "<g count='%d'>" count
        hPutStr h (bsTo01String g)
        hPutStrLn h "</g>"
      hPutStrLn h "</inventory>"
      forM_ runsDataAll $ \population -> do
        hPutStrLn h "<population>"
        forM_ (Set.toList population) $ \allGenes -> do
          hPutStrLn h "<agent>"
          hPutStr h "<gx>"
          hPutStr h (bsTo01String (head (reduceToClosest grandReferenceGene allGenes)))
          hPutStrLn h "</gx>"
          forM_ allGenes $ \gene -> do
            hPutStr h "<g>"
            hPutStr h (bsTo01String gene)
            hPutStrLn h "</g>"
          hPutStrLn h "</agent>"
        hPutStrLn h "</population>"
      hPutStrLn h "</populations>"
    writeFile matOutPath (unlines hammingRows)
  return ()

-- | (Utility) For 'Just' an object, show the object.  For
-- 'Nothing', return "undefined".

mShow :: Show t => Maybe t -> String

mShow Nothing = "\"undefined\""

mShow (Just x) = show x
