{-# LANGUAGE ForeignFunctionInterface, EmptyDataDecls, ScopedTypeVariables #-}

{- |

Module: System.Random.CSFMT

Description : Basic wrapper ardound the Mersenne Twister random number generator.

Copyright   : © 2019 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

This wraps a C implementation of the Mersenne twister algorithm,
specifically, a modification of the dSFMT package by Saito and
Matsumoto.  This package is the basic wrapper around

-}

module System.Random.CSFMT (
  Generator,
  newGeneratorFromInt,
  newGeneratorFromWords,
  randomUnitDouble,
  randomInt)

where

import Foreign
import Foreign.C
import Control.Monad.ST
import Control.Monad.ST.Unsafe
import System.IO
import System.IO.Unsafe


data CSFMT_prim

type CSFMT = Ptr CSFMT_prim

type Generator = ForeignPtr CSFMT_prim

foreign import ccall unsafe "CachingSFMT.h csfmt_new"
  csfmt_new :: CInt -> IO CSFMT

foreign import ccall unsafe "CachingSFMT.h csfmt_newFromWords"
  csfmt_newFromWords :: Ptr Word32 -> CInt -> IO CSFMT

foreign import ccall unsafe "CachingSFMT.h csfmt_genrand_close_open"
  csfmt_genrand_close_open :: CSFMT -> IO Double

foreign import ccall unsafe "CachingSFMT.h csfmt_finalize"
  csfmt_finalize :: CSFMT -> IO ()

foreign import ccall unsafe "CachingSFMT.h &csfmt_finalize"
  csfmt_finalize_p :: FunPtr (CSFMT -> IO ())

{-# NOINLINE newGeneratorFromInt #-}
newGeneratorFromInt :: Int -> Generator
newGeneratorFromInt seed = unsafePerformIO $ do
  pGen <- csfmt_new (fromIntegral seed)
  newForeignPtr csfmt_finalize_p pGen

{-# NOINLINE newGeneratorFromWords #-}
newGeneratorFromWords :: [Word32] -> Generator
newGeneratorFromWords seeds = unsafePerformIO $ do
  pGen <- withArrayLen seeds $ \nWords ptr ->
    csfmt_newFromWords ptr (fromIntegral nWords)
  newForeignPtr csfmt_finalize_p pGen

{-# NOINLINE randomUnitDouble #-}
{- | Create a random double in the range [0,1) -}
randomUnitDouble :: Generator -> ST s Double
randomUnitDouble fpGen =
  unsafeIOToST $ withForeignPtr fpGen $ \pGen ->
     csfmt_genrand_close_open pGen

randomInt :: forall s n. (Integral n, Bounded n) =>
             Generator -> ST s n
randomInt g =
  let
    a = fromIntegral (minBound :: n)
    b = fromIntegral (maxBound :: n)
    spread = b + 1 - a
  in do
    x <- randomUnitDouble g
    let n = floor (a + spread * x)
    n `seq` return n

_testNoFinalizer :: IO ()
_testNoFinalizer = do
  hPutStrLn stderr "Start"
  gen <- csfmt_new 12345
  hPutStrLn stderr "Generator initialized"
  x <- csfmt_genrand_close_open gen
  hPutStrLn stderr "Generator run"
  putStrLn (show x)
  putStrLn "Ok"
