{-# LANGUAGE DefaultSignatures, TypeFamilies, MultiParamTypeClasses,
  FlexibleInstances, FlexibleContexts, UndecidableInstances #-}

{- |

Module: WGM.Snapshot

Description: How to record the state of a dynamic object.  Since
objects in Haskell are constant outside of special monads like IO
and ST, most of this machinery works within one of these monads.

Copyright: © 2019 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC)

-}

module WGM.Snapshot where

import Control.Monad
import Control.Monad.ST

import qualified Data.Graph.Inductive.PatriciaTree as GPT
import qualified Data.Graph.Inductive.Graph as G

import Data.Foldable

import Data.IntMap (IntMap)
import qualified Data.IntMap as IntMap

import Data.Map (Map)
import qualified Data.Map as Map

import Data.STRef

import Data.Sequence (Seq)
import qualified Data.Sequence as Seq

class Snapshot m t where
  -- | The type of the memento generated by taking a snapshot.
  -- The default is the same as 't'.
  type Memento t :: *
  type instance Memento t = t


  -- | Extract a memento from a dynamic object.
  -- The default is to return the object itself.
  snapshot :: Monad m => t -> m (Memento t)
  default snapshot :: (Monad m, Memento t ~ t) => t -> m (Memento t)
  snapshot x = return x


class Snapshot m t => Reconstruct m t where

  -- | Reconstruct an object from a memento.
  -- The default is to return the memento itself.
  reconstruct :: Monad m => Memento t -> m t
  default reconstruct :: (Monad m, Memento t ~ t) => Memento t -> m t
  reconstruct x = return x


instance Snapshot m Int
instance Reconstruct m Int
instance Snapshot m Char
instance Reconstruct m Char
instance Snapshot m Double
instance Reconstruct m Double




instance
  (Snapshot m t)
  => Snapshot m [t]   where
  type Memento [t] = [Memento t]
  snapshot = mapM snapshot

instance
  (Reconstruct m t)
  => Reconstruct m [t]   where
  reconstruct = mapM reconstruct




instance
  (Snapshot m t)
  => Snapshot m (Maybe t)   where
  type Memento (Maybe t) = Maybe (Memento t)
  snapshot Nothing = return Nothing
  snapshot (Just x) = do
    y <- snapshot x
    return (Just y)


instance
  (Reconstruct m t)
  => Reconstruct m (Maybe t) where
  reconstruct Nothing = return Nothing
  reconstruct (Just y) = do
    x <- reconstruct y
    return (Just x)




instance
  (Snapshot m t1,
   Snapshot m t2)
  => Snapshot m (t1, t2)   where
  type Memento (t1, t2) = (Memento t1, Memento t2)
  snapshot (x1, x2) = do
    m1 <- snapshot x1
    m2 <- snapshot x2
    return (m1, m2)

instance
  (Reconstruct m t1,
   Reconstruct m t2)
  => Reconstruct m (t1, t2)   where
  reconstruct (y1, y2) = do
    x1 <- reconstruct y1
    x2 <- reconstruct y2
    return (x1, x2)



instance Snapshot (ST s) t =>
  Snapshot (ST s) (STRef s t) where
  type Memento (STRef s t) = Memento t

  snapshot ref = do
    y <- readSTRef ref
    snapshot y

instance Reconstruct (ST s) t =>
  Reconstruct (ST s) (STRef s t) where

  reconstruct y = do
    x <- reconstruct y
    newSTRef x



instance
  (Snapshot m value) =>
  Snapshot m (IntMap value) where

  type Memento (IntMap value) = IntMap (Memento value)

  snapshot im = do
    pairs <- forM (IntMap.toList im) $ \ (key, value) -> do
      valueM <- snapshot value
      return (key, valueM)
    return (IntMap.fromList pairs)

instance
  (Reconstruct m value) =>
  Reconstruct m (IntMap value) where

  reconstruct m = do
    pairs <- forM (IntMap.toList m) $ \ (key, valueM) -> do
      value <- reconstruct valueM
      return (key, value)
    return (IntMap.fromList pairs)



instance
  (Snapshot m nodeLabel,
   Snapshot m edgeLabel)
  =>
  Snapshot m (GPT.Gr nodeLabel edgeLabel) where

  -- | The 'Gr' types are missing a bunch of instnaces
  -- and there's not really a way to write them, so
  -- the 'Memento' has to be a list of notes and a list of edges.
  type Memento (GPT.Gr nodeLabel edgeLabel) =
    ([G.LNode (Memento nodeLabel)], [G.LEdge (Memento edgeLabel)])

  snapshot g = do
    -- mementos for all nodes
    nodeMs <- forM (G.labNodes g) $ \(n, nLabel) -> do
      nodeM <- snapshot nLabel
      return (n, nodeM)

    -- mementos for all edges
    edgeMs <- forM (G.labEdges g) $ \(pre, post, eLabel) -> do
      edgeM <- snapshot eLabel
      return (pre, post, edgeM)

    return (nodeMs, edgeMs)


instance
  (Reconstruct m nodeLabel,
   Reconstruct m edgeLabel)
  =>
 Reconstruct m (GPT.Gr nodeLabel edgeLabel) where

  reconstruct (nodeMs, edgeMs) = do
    -- mementos for all nodes
    nodes <- forM nodeMs $ \(n, nLabel) -> do
      node <- reconstruct nLabel
      return (n, node)

    -- mementos for all edges
    edges <- forM edgeMs $ \(pre, post, eLabel) -> do
      edge <- reconstruct eLabel
      return (pre, post, edge)

    return (G.mkGraph nodes edges)


instance (Monad m, Snapshot m t) => Snapshot m (Seq t) where

  type Memento (Seq t) = Seq (Memento t)

  snapshot xs = do
    ys <- snapshot (toList xs)
    return (Seq.fromList ys)

instance (Monad m, Reconstruct m t) => Reconstruct m (Seq t) where
  reconstruct ys = do
    xs <- reconstruct (toList ys)
    return (Seq.fromList xs)


instance
  (Ord key,
   Snapshot m key,
   Ord (Memento key),
   Snapshot m value)
  => Snapshot m (Map key value) where

  type Memento (Map key value) = Map (Memento key) (Memento value)

  snapshot m = do
    pairMs <- mapM snapshot (Map.toList m)
    return (Map.fromList pairMs)
