{-# LANGUAGE BangPatterns #-}

{- |

Module: WGM.Util

Description: All manner of odd utility functions that I've needed
from time to time.

Copyright: © 2017 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : Portable

-}


module WGM.Util where

import Control.DeepSeq
import Data.List
import Data.Ord
import qualified Data.IntMap as IM
import Text.Printf


-- | Given [a, b, c] [x, y, z], reverse the first list and
-- prepend the elements to the second list, giving [c, b, a, x,
-- y, z]

prependReverse :: [a] -> [a] -> [a]
prependReverse [] vs = vs
prependReverse (u:us) vs = prependReverse us (u:vs)

{-# INLINABLE prependReverse #-}


-- | (private) Accumulator helper for 'splitFirst'
splitFirstHelper :: (a -> Bool) -> [a] -> [a] -> ([a], Maybe a, [a])

splitFirstHelper pr seenRev (x:xs) =
    if pr x then (seenRev, Just x, xs)
    else splitFirstHelper pr (x:seenRev) xs

splitFirstHelper _ seenRev [] =
    (seenRev, Nothing, [])

{-# INLINABLE splitFirstHelper #-}


-- | Given [1,2,3,-1,4,5,6] and (\< 0), return ([3,2,1], Just -1,
-- [4,5,6]).  That is, return a reversed list of items that don't
-- satisfy the predicate, then the first item that does, then the
-- rest.  If no items satisfy it, return ([...],Nothing,[])

splitFirst :: (a -> Bool) -> [a] -> ([a], Maybe a, [a])

splitFirst pr xs = splitFirstHelper pr [] xs

{-# INLINABLE splitFirst #-}


-- | Form a list of @(x, f x)@
keepMap :: (a -> b) -> [a] -> [(a, b)]
keepMap f xs =
  [(x, f x) | x <- xs]

{-# INLINABLE keepMap #-}


-- | Form a list of @(f x, x)@ -}
mapKeep :: (a -> b) -> [a] -> [(b, a)]
mapKeep f xs =
  [(f x, x) | x <- xs]

{-# INLINABLE mapKeep #-}

-- | A strict loop
loop :: Int -> (a -> a) -> a -> a
loop 0 _ !x0 = x0
loop n f !x = loop (n-1) f (f x)

{-# INLINABLE loop #-}

-- | A strict monadic loop
loopM :: Monad m => Int -> (a -> m a) -> a -> m a
loopM 0 _ !x0 = return x0
loopM n axn !x = do
  xNext <- axn x
  loopM (n-1) axn xNext

{-# INLINABLE loopM #-}


-- | A strict accumulating monadic loop
iterateCM :: Monad m => Int -> (a -> m a) -> a -> m [a]
iterateCM 0 _ !x0 = return [x0]
iterateCM nSteps axn !x = do
  xNext <- axn x
  rest <- iterateCM (nSteps - 1) axn xNext
  return (xNext:rest)

{-# INLINABLE iterateCM #-}


-- | A very strict loop
loopDS :: NFData a => Int -> (a -> a) -> a -> a
loopDS 0 _ x0 = deepseq x0 x0
loopDS n f x =
  let result = loopDS (n-1) f (f x)
  in deepseq result result

{-# INLINABLE loopDS #-}


-- | A very strict monadic loop
loopMDS :: (NFData a, Monad m) => Int -> (a -> m a) -> a -> m a
loopMDS 0 _ !x0 = return x0
loopMDS n axn !x = do
  xNext <- axn x
  deepseq xNext (loopMDS (n-1) axn xNext)

{-# INLINABLE loopMDS #-}


-- | A very strict monadic loop
replicateMDS ::
  (Monad m, NFData a) =>
  Int -> m a -> m [a]

replicateMDS 0 _ = return []
replicateMDS n axn = do
  next <- axn
  deepseq next (do
                   rest <- replicateMDS (n-1) axn
                   return (next:rest))

{-# INLINABLE replicateMDS #-}


-- | Reverse an 'Ordering' so that 'LT' maps to 'GT', 'GT' maps
-- to 'LT', 'EQ' maps to itself.

reverseOrdering :: Ordering -> Ordering
reverseOrdering EQ = EQ
reverseOrdering LT = GT
reverseOrdering GT = LT

{-# INLINABLE reverseOrdering #-}


-- | Reverse a comparison function

reverseComparison :: Ord a => (a -> a -> Ordering) -> a -> a -> Ordering
reverseComparison f x y =
  reverseOrdering (f x y)

{-# INLINABLE reverseComparison #-}


-- | Sort objects using a veiw function
sortThrough :: Ord b => (a -> b) -> [a] -> [a]
sortThrough f xs = sortBy (comparing f) xs

{-# INLINABLE sortThrough #-}


-- | Add copies of z to the front of xs to produce a list of length n
padFrontToLength :: Int -> a -> [a] -> [a]
padFrontToLength n z xs =
  let m = length xs
      k = n - m
  in (replicate k z) ++ xs

{-# INLINABLE padFrontToLength #-}


-- | A helper for 'allBitPatterns' that takes an accumulator
allBitPatternsS :: Int -> [[Bool]] -> [[Bool]]
allBitPatternsS 0 xss = xss
allBitPatternsS n xss =
  allBitPatternsS (n-1) (concatMap (\xs-> [True:xs,False:xs]) xss)

{-# INLINABLE allBitPatternsS #-}


-- | All 2^n bit patterns you can make form n bits
allBitPatterns  :: Int -> [[Bool]]
allBitPatterns n = allBitPatternsS n [[]]

{-# INLINABLE allBitPatterns #-}


-- | All 2^n sublists you can make from a list of n items
allSubLists :: [a] -> [[a]]
allSubLists [] = [[]]
allSubLists (x:xs) =
  let ys = allSubLists xs
  in [x:y | y <- ys] ++ ys

{-# INLINABLE allSubLists #-}


-- | Do an action for each item in a list, with a between action in between

forMB_ :: Monad m =>
          [t] -> m () -> (t -> m ()) -> m ()
forMB_ [x] _ f =
  f x

forMB_ (x:xs) b f = do
  f x
  b
  forMB_ xs b f

{-# INLINABLE forMB_ #-}


-- | A more error-handlable version of 'read'
readM :: (Monad m, MonadFail m, Read a) => String -> m a
-- To compile with older Haskell libraries before `MonadFail` was introduced,
-- use this type instead:
-- readM :: (Monad m, Read a) => String -> m a
readM s = case [x | (x,t) <- reads s, ("","")  <- lex t] of
  [x] -> return x
  []  -> fail ("readM: unable to parse: " ++ take 20 s)
  _   -> fail ("readM: ambiguous parse: " ++ take 20 s)

{-# INLINABLE readM #-}

transposeWith :: ([e] -> t) -> [[e]] -> [t]

transposeWith _ ([]:_) =
  []

transposeWith f rows =
  let
    col = map head rows
    restOfRows = map tail rows
  in
   f col : transposeWith f restOfRows

-- | Return 'True' if all items are equal.  An empty list yields
-- 'True' vacuously (because there are no items in the list that
-- are unequal).

allSame :: Eq a => [a] -> Bool
allSame (x1:x2:xs) = x1 == x2 && allSame (x2:xs)
allSame _ = True

{-# INLINABLE allSame #-}


-- | Given a list of @xs@, covert it to a list @[(x1, n1), (x2,
-- n2) ...]@ such that the original list consists of @n1@ copies
-- of @x1@, @n2@ copies of @x2@, ...
--
-- NOTE: On 2013-05-13 I reversed the tuples to be more
-- compatible with Mathematica and to make it easier to produce a
-- @Data.Map@ from items to counts.

runLengthCompress :: Eq a => [a] -> [(a, Int)]
runLengthCompress [] = []
runLengthCompress (x:xs) = helper 1 x xs
  where
    helper :: Eq a => Int -> a -> [a] -> [(a, Int)]
    helper !n y ys =
      case ys of
        [] -> [(y,n)]
        (z:zs) -> if y == z
                  then helper (n+1) y zs
                  else (y,n) : helper 1 z zs

{-# INLINABLE runLengthCompress #-}

-- | Given a list of @[(x1, n1), (x2, n2) ...]@ produce a list
-- consisting of @n1@ copies of @x1@ followed by @n2@ copies of
-- @x2@ .... This function is the inverse of 'runLengthCompress'

runLengthExpand :: [(a, Int)] -> [a]
runLengthExpand items =
  case items of
    [] ->
      []
    (x, n) : more ->
      loop x n more
  where
    loop x n more =
      if n <= 0
      then
        case more of
          [] -> []
          (x2, n2) : evenMore ->
            loop x2 n2 evenMore
      else
        x : loop x (n-1) more

{-# INLINABLE runLengthExpand #-}

-- | Slightly more informative checked lookup of a key in an 'IntMap'.
checkedLookup :: (Show v) => String -> IM.IntMap v -> Int -> v
checkedLookup msg m k =
  case IM.lookup k m of
    Just v -> v
    Nothing ->
      error (printf "%s: key %d is not present in map:\n%s"
             msg k (show (IM.toList m)))

{-# INLINABLE checkedLookup #-}

{-
-- | Split a list up into sublists of the given length.
chunksOf :: Int -> [a] -> [[a]]

chunksOf _ [] = []
chunksOf n xs =
  let (chunk, rest) = splitAt n xs
  in (chunk : chunksOf n rest)

chunks = chunksOf
-}
