{-# LANGUAGE
MultiParamTypeClasses,
TypeFamilies,
FlexibleContexts,
UndecidableInstances,
FlexibleInstances,
StandaloneDeriving,
RecordWildCards,
NamedFieldPuns,
ScopedTypeVariables,
OverloadedStrings,
TemplateHaskell,
DeriveDataTypeable,
DeriveGeneric
-- AllowAmbiguousTypes
#-}

{-|

Module      : Digiostelium.Genome

Description : Genome-related type classes, data types, and functions

Copyright   : © 2019 W. Garrett Mitchener

License     : GPL-3

Maintainer  : garrett.mitchener@gmail.com

Stability   : experimental

Portability : non-portable (GHC, POSIX)

-}


module Digiostelium.Genome where

import GHC.Generics

import Control.DeepSeq
import Control.Exception
import Control.Monad
import Control.Monad.Identity (runIdentity)
import Control.Monad.State (runStateT)

import Data.Array
import Data.Aeson hiding (Array)
import Data.Aeson.Types hiding (Array)
import Data.Aeson.TH
import Data.Binary
import Data.Data
import Data.Default
import Data.List
import Data.Either
import qualified Data.Vector as V

import Text.Printf

import Digiostelium.AesonUtil
import Digiostelium.BlockSize
import Data.BinaryReflectedGrayCode
import Data.BitSource
import Data.BitString
import BaselMachine.Compiler
import BaselMachine.Instruction
import qualified Digiostelium.RangeMap as R
import WGM.ShortShow
import WGM.ParUtil

-- | Class for things that have a genome
class HasGenome t genome where
  -- | Retrieve a genome
  getGenome :: t -> genome




-- | Encoding of pattern and threshold integers.
--
-- The 'PlusMinus' requires some explanation:
--
-- @
-- if m > 0 then m else m - 1
-- @
--
-- @[PlusMinus]@ with a one bit number
-- translates 1 to 1 and 0 to -1.
--
-- @[Offset (-1), PlusMinus]@ with a two bit number
-- translates @[0..3]@ to @[-2,-1,1,2]@
--
-- @[Offset (1-2^(n-1)), PlusMinus]@ with an @n@ bit number
-- translates @[0..2^n-1]@ to @[-2^(n-1), ..., -1, 1, ..., 2^(n-1)]@
--
-- That way, the result has matching positive and negative
-- results and no zero.

data GenomeIntSpec =
  Constant Int -- ^ always the same
  | TwosComplement Int -- ^ twos complement (no additional encoding)
  | ReflectedGrayEncoding Int -- ^ Gray code
  | PlusMinus
  | Offset Int GenomeIntSpec -- ^ add an offset
  | SumMany Int GenomeIntSpec -- ^ add several together
  deriving (Read, Show, Eq, Ord, Typeable, Data, Generic)

instance NFData GenomeIntSpec
instance Binary GenomeIntSpec
$(deriveJSON myJSONOpts ''GenomeIntSpec)

instance BitSpec GenomeIntSpec where

  type FromBitsResult GenomeIntSpec = Int

  readFromBits (Constant m) =
    return m

  readFromBits (TwosComplement n) = do
    readFromBits (IntSpec n)

  readFromBits (ReflectedGrayEncoding n) = do
    m <- readFromBits (IntSpec n)
    return (reflectedGrayToBinary m)

  readFromBits PlusMinus = do
    m <- readFromBits (IntSpec 1)
    return (if m > 0 then m else m - 1)

  readFromBits (Offset d enc) = do
    m <- readFromBits enc
    return (m + d)

  readFromBits (SumMany n enc) = do
    ms <- replicateM n (readFromBits enc)
    return (sum ms)



instance BitSpecSize GenomeIntSpec where

  sizeInBits (Constant _) _ = 0

  sizeInBits (TwosComplement n) _ = n

  sizeInBits (ReflectedGrayEncoding n) _ = n

  sizeInBits PlusMinus _ = 1

  sizeInBits (Offset _ enc) x =
    sizeInBits enc x

  sizeInBits (SumMany n enc) x =
    n * sizeInBits enc x



data InstructionSpec = InstructionSpec {
  numReactions :: Int,
  patternSpec :: GenomeIntSpec,
  -- ^ how bit patterns in the genome are decoded
  thresholdSpec :: GenomeIntSpec,
  -- ^ how thresholds (positive or signed integers) in the genome
  -- are decoded
  deltaSpec :: GenomeIntSpec,
  -- ^ how deltas (signed integers) in the genome are decoded
  deltaSign :: DeltaSign
  -- ^ further options for the signs of reaction deltas
  } deriving (Read, Show, Eq, Ord, Typeable, Data, Generic)

data DeltaSign =
  JustDecode
  -- ^ sign of a reaction delta is left unchanged from the
  -- decoded value
  |
  ToggleByEvenOdd
  -- ^ the delta of an even numbered reaction is as decoded, the
  -- delta of an odd numbered reaction is multiplied by -1.
  deriving (Read, Show, Eq, Ord, Typeable, Data, Generic)

instance NFData DeltaSign
instance Binary DeltaSign
$(deriveJSON myJSONOpts ''DeltaSign)

adjustSign :: DeltaSign -> Int -> Int -> Int

adjustSign JustDecode _ delta = delta

adjustSign ToggleByEvenOdd n delta =
  if n `mod` 2 == 0
  then delta
  else -delta


-- | The default is 6 pattern bits, 4 threshold bits, 2
-- reactions, 0 delta bits, and delta signs are determined by the
-- even or odd index of the reaction.  The first reaction has
-- index 0.  Integers are decoded by reflected gray encoding.
-- This replicates the original Utrecht machine instruction spec.
instance Default InstructionSpec where
  def = InstructionSpec {
    numReactions = 2,
    patternSpec = ReflectedGrayEncoding 6,
    thresholdSpec = ReflectedGrayEncoding 4,
    deltaSpec = Constant 1,
    deltaSign = ToggleByEvenOdd
    }

instance NFData InstructionSpec
instance Binary InstructionSpec
$(deriveJSON myJSONOpts ''InstructionSpec)


instance BitSpec InstructionSpec where

  type FromBitsResult InstructionSpec = Instruction

  readFromBits InstructionSpec{..} = do
    switchPattern <- readFromBits patternSpec
    switchThreshold <- readFromBits thresholdSpec
    targetsAndDeltas <- forM [0 .. numReactions - 1] $ \n -> do
      targetPattern <- readFromBits patternSpec
      delta <- readFromBits deltaSpec
      return (targetPattern, adjustSign deltaSign n delta)
    return Instruction{..}



-- | By the way, the existnce of this instance restricts genome
-- architectures to a fixed instruction size.
instance BitSpecSize InstructionSpec where

  sizeInBits (InstructionSpec {..}) _ =
    (sizeInBits patternSpec 0
      + sizeInBits thresholdSpec 0
      + (numReactions *
          (sizeInBits patternSpec 0
            + sizeInBits deltaSpec 0)))



-- | How many codon bits does an instruction take
bitsPerInstruction :: InstructionSpec -> Int
bitsPerInstruction spec = sizeInBits spec undefined



-- | Structure of a segment of a gene (switch pattern, threshold,
-- reaction).  This requires some back and forth between this
-- module and the 'GenomeSpec' because the 'GenomeSpec' is
-- responsible for translating from nucleotide bits to codon
-- bits.

data SegmentEncoding = SegmentEncoding {
  segmentNucleotideBitRange :: (Int, Int), -- ^ [begin, end) range of bit indices into the gene's nucleotide bits
  segmentNucleotideBits :: BitString, -- ^ nucleotide bits
  segmentCodonBitRange :: (Int, Int), -- ^ [begin, end) range of bit indices in gene's codon bits
  segmentCodonBits :: BitString, -- ^ codon bits
  segmentIntValue :: Int -- ^ integer value of these bits
  } deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)

instance NFData SegmentEncoding
instance Binary SegmentEncoding
$(deriveJSON myJSONOpts ''SegmentEncoding)

data EncodingDetails = EncodingDetails {
  switchPatternSegment :: SegmentEncoding, -- ^ how the switch pattern is encoded
  thresholdSegment :: SegmentEncoding, -- ^ how the threshold is encoded
  reactionSegments :: [(SegmentEncoding, SegmentEncoding)] -- ^ how the reactions are encoded
  } deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)

instance NFData EncodingDetails
instance Binary EncodingDetails
$(deriveJSON myJSONOpts ''EncodingDetails)


-- | (Subroutine) Make a partially completed 'EncodingDetails'
-- structure for a 'BitString' of nucleotide bits.  The
-- 'segmentNucleotideBitRange' field is initialized to @(0,0)@.
-- The 'segmentNucleotideBits' field is initialized to 'bsEmpty'.
-- The intention is for the 'geneDetails' function related to a
-- specific instance of 'GenomeSpec' to use this function and
-- fill in the missing information.  See also
-- 'majorityCodeCompleteEncodingDetails'.

partialEncodingDetails ::
  InstructionSpec
  -> BitString -- ^ codon bits for the instruction
  -> EncodingDetails

partialEncodingDetails instSpec codonBits =
  details
  where
    -- looseBits should be empty
    (details, _looseBits) = runIdentity (runStateT (partialEncodingDetailsM instSpec) codonBits)

partialEncodingDetailsM ::
  ReadBits m =>
  InstructionSpec
  -> m EncodingDetails

partialEncodingDetailsM InstructionSpec{..} = do
  swp <- readFromBits patternSpec
  let swpEnd = sizeInBits patternSpec swp
      switchPatternSegment = SegmentEncoding {
        segmentCodonBitRange = (0, swpEnd),
        segmentIntValue = swp,
        ..}

  thr <- readFromBits thresholdSpec
  let thrBegin = swpEnd
      thrEnd = thrBegin + sizeInBits thresholdSpec thr
      thresholdSegment = SegmentEncoding {
        segmentCodonBitRange = (thrBegin, thrEnd),
        segmentIntValue = thr,
        ..}

  let patSize = sizeInBits patternSpec 0
      deltaSize = sizeInBits deltaSpec 0
      rxnSize = patSize + deltaSize

  reactionSegments <- forM [0 .. numReactions - 1] $ \n -> do
    tgt <- readFromBits patternSpec
    delta0 <- readFromBits deltaSpec
    let tgtBegin = thrEnd + (n * rxnSize)
        tgtEnd = tgtBegin + patSize
        delta = adjustSign deltaSign n delta0
        deltaBegin = tgtEnd
        deltaEnd = deltaBegin + deltaSize
    return (SegmentEncoding {
               segmentCodonBitRange = (tgtBegin, tgtEnd),
               segmentIntValue = tgt,
               ..},
            SegmentEncoding {
               segmentCodonBitRange = (deltaBegin, deltaEnd),
               segmentIntValue = delta,
               ..})
  return EncodingDetails {..}
    where
      -- captured by SegmentEncoding
      segmentNucleotideBitRange = (0,0)
      segmentNucleotideBits = bsEmpty
      segmentCodonBits = bsEmpty


-- | (Subroutine) Complete the results of 'partialEncodingDetails' assuming a
-- majority code from nucleotide bits to codon bits.

majorityCodeCompleteEncodingDetails ::
  Int
  -> BitString
  -> BitString
  -> EncodingDetails
  -> EncodingDetails

majorityCodeCompleteEncodingDetails expansionFactor nucleotideBits codonBits EncodingDetails{..} =
  EncodingDetails {
    switchPatternSegment = completeSegment switchPatternSegment,
    thresholdSegment = completeSegment thresholdSegment,
    reactionSegments = map (\(t,d) -> (completeSegment t, completeSegment d)) reactionSegments }
  where
    completeSegment SegmentEncoding{segmentCodonBitRange = segmentCodonBitRange@(codonBegin, codonEnd), ..} =
      SegmentEncoding {
        segmentNucleotideBitRange = (expansionFactor * codonBegin, expansionFactor * codonEnd),
        segmentNucleotideBits = bsSubstring (expansionFactor * codonBegin) (expansionFactor * codonEnd) nucleotideBits,
        segmentCodonBits = bsSubstring codonBegin codonEnd codonBits,
        ..}



-- | For building, manipulating, and interpreting genomes.
--
-- The trickiest part in all of this seems to be keeping up with the
-- histories of intervals in the genome.  I have to do that so that
-- I can track the histories of interesting genes.
--
-- This public interface should have only bit indices, not block
-- (=whole gene) indices, because I want to write stuff that still
-- works if I switch to a non-block genome architecture.
--
-- See "ProblemInventory" for the type instances expected for a
-- 'GenomeSpec'.

class GenomeSpec gs where
  -- | Genome type
  type Genome gs :: *

  -- | Associated gene interval type that represents a contiguous
  -- part of the genome
  type Interval gs :: *

  -- | Associated mutation type
  type Mutation gs :: *

  -- | Get the 'InstructionSpec' used to decode bits into
  -- an 'Instruction'
  getInstructionSpec :: gs -> InstructionSpec

  -- | Length of the 'Genome'.  Since this can potentially be
  -- more than 4 billion letters, it reutrns an 'Integer'
  genomeLength :: gs -> Genome gs -> Integer

  -- | Return full details of how each gene is constructed and
  -- interpreted.
  geneDetails :: gs -> Genome gs -> [GeneDetails gs]

  -- | Export just the bit data as JSON, giving sufficient
  -- information to reconstruct the genome but not necessarily
  -- any of its history, etc.  To be used for copying genomes
  -- from one sample run to another, maybe even solving a
  -- different problem, or for analysis by some entirely other
  -- package.
  exportGenomeData :: gs -> Genome gs -> Value

  -- | Import the genome data, creating a genome with no history,
  -- etc., just bits.  Return 'Left' with an error message if it
  -- can't be done.
  importGenomeDataParser :: gs -> Value -> Parser (Genome gs)

  -- | Map an 'Interval' to its sources in parent genomes.
  intervalSource :: gs -> Genome gs -> Interval gs -> IntervalSource gs

  -- | Produce a list of all possible 'Mutation's that the given
  -- 'Genome' can experience
  allPossibleMutations :: gs -> Genome gs -> [Mutation gs]

  -- | Apply a list of 'Mutation' to a 'Genome', for a knockout
  -- experiment, etc.
  applyMutations :: gs -> [Mutation gs] -> Genome gs -> Genome gs

  -- | Get the 'Interval' changed by a 'Mutation'
  intervalsChanged :: gs -> Mutation gs -> Genome gs -> [Interval gs]

-- | An 'IntervalSource' records where an 'Interval' of a
-- 'Chromosome' came from, either from a parent 'Chromosome', or
-- synthesized from some other source like an indel mutation or
-- the random generator.  It should be possible to fully
-- reconstitute this part of the chromosome from the
-- 'IntervalSource' and the parents' 'Genome's.  Deletions and
-- duplications should not be required-- They should be accounted
-- for when building the 'IntervalSource'.  The number of bits
-- coming from an 'IntervalSource' should be equal to the number
-- of bits in the original interval.

type IntervalSource gs = R.RangeMap (Int, Int) (FromChromosome gs)

data FromChromosome gs =
  FromChromosome {
    parentIx :: !Int,
    -- ^ which parent in the list of parents it comes from, or -1
    -- if none
    inParent :: !(Interval gs),
    -- ^ where in the parent
    relativeFlips :: ![Int]
    -- ^ bits to flip in the resulting string, indexed relative
    -- to its beginning
    }
  deriving (Typeable, Generic)

instance
  (GenomeSpec gs,
   ShortShow (Interval gs)) =>
  ShortShow (FromChromosome gs) where
    shortShow (FromChromosome {..}) =
      printf "p%d:%s,flip%s" parentIx (shortShow inParent) (shortShow relativeFlips)

instance
  (GenomeSpec gs,
   R.Range (Interval gs))
  => R.Range (FromChromosome gs) where

    empty =
      FromChromosome (-1) R.empty []

    length (FromChromosome {..}) =
      R.length inParent

    take n (FromChromosome {..}) =
      FromChromosome {
        inParent = R.take n inParent,
        relativeFlips = [f | f <- relativeFlips, f < n],
        ..}

    drop n (FromChromosome {..}) =
      FromChromosome {
        inParent = R.drop n inParent,
        relativeFlips = [f - n | f <- relativeFlips, n <= f],
        ..}

instance
  (GenomeSpec gs,
   R.BoundedRange (Interval gs),
   R.Range (Interval gs))
  => R.BoundedRange (FromChromosome gs) where

    bounds (FromChromosome {..}) =
      R.bounds inParent

deriving instance (Read (Interval gs)) => Read (FromChromosome gs)
deriving instance (Show (Interval gs)) => Show (FromChromosome gs)
deriving instance (Eq (Interval gs)) => Eq (FromChromosome gs)
deriving instance (Ord (Interval gs)) => Ord (FromChromosome gs)
deriving instance (Typeable gs, Data gs, Data (Interval gs)) => Data (FromChromosome gs)

-- | TODO: This instance can probably be generated automatically
-- by the compiler
instance (Binary (Interval gs)) => Binary (FromChromosome gs) where
  put (FromChromosome {..}) = do
    put parentIx
    put inParent
    put relativeFlips


  get = do
    parentIx <- get
    inParent <- get
    relativeFlips <- get
    return (FromChromosome {..})

instance (NFData (Interval gs)) => NFData (FromChromosome gs) where
  rnf (FromChromosome {..}) = do
    rnf parentIx
    `seq` rnf inParent
    `seq` rnf relativeFlips

data GeneDetails gs = GeneDetails {
  instruction :: Instruction,
  interval :: Interval gs,
  codonBits :: BitString,
  nucleotideBits :: BitString,
  geneSource :: IntervalSource gs, -- ^ source strings in parent genomes, to concatenate
  instructionEncodingDetails :: EncodingDetails
  }

deriving instance (Read (Interval gs)) => Read (GeneDetails gs)
deriving instance (Show (Interval gs)) => Show (GeneDetails gs)
deriving instance (Eq (Interval gs)) => Eq (GeneDetails gs)
deriving instance (Ord (Interval gs)) => Ord (GeneDetails gs)



data CompileError =
  NoError
  | InvalidSwitchPattern Reaction
  | InvalidTargetPattern Reaction
  deriving (Read, Show, Eq, Ord, Data, Typeable, Generic)

data CompileErrors =
  CompileErrors [CompileError] [Instruction]
  deriving (Read, Show, Data, Typeable, Generic)

instance Exception CompileErrors


-- | If all goes well, the list of errors should be [].

checkSplitInstructions ::
  Int -- ^ number of patterns, 0 through this number - 1
  -> [(Int, Instruction)] -- ^ indexed instructions
  -> ([CompileError], [Reaction]) -- ^ results

checkSplitInstructions nPatterns ixInst =
  let
    reactions = concatMap splitInstruction ixInst

    checkReaction r@Reaction{..} = do
      when (sPat < 0 || sPat >= nPatterns) $
        Left (InvalidSwitchPattern r)
      when (tPat < 0 || tPat >= nPatterns) $
        Left (InvalidTargetPattern r)
      Right r
  in
   partitionEithers (map checkReaction reactions)

-- | Compile the given list of 'Instructions', checking for lots
-- of mistakes.

compile ::
  InstructionSpec
  -> [Instruction] -- ^ the program
  -> Either [CompileError] (ProgramInformation, ReactionProgram)

compile instSpec instructions =
  let
    InstructionSpec{..} = instSpec
    nPatterns = 2 ^ (sizeInBits patternSpec 0)
    (compileErrors, _reactions) =
      checkSplitInstructions nPatterns (zip [0..] instructions)

  in
   case compileErrors of
     [] -> Right (compileUnchecked instructions)
     _ -> Left compileErrors




-- | Import a list of 'Genome's from an Aeson 'Value', returning
-- in either 'Left' with an error message, or 'Right' with a list
-- of 'Genome's.

importGenomesParser ::
  (GenomeSpec gs) =>
  gs -> Value -> Parser [Genome gs]

importGenomesParser gs =
  withArray "array of genomes" $ \a ->
  forM (V.toList a) $
  withObject "agent data object" $ \agentObj -> do
    genomeObj <- agentObj .: "genome"
    importGenomeDataParser gs genomeObj


-- | Return just the instructions encoded in a genome.
interpretGenome ::
  (GenomeSpec gs) =>
  gs -> Genome gs -> [Instruction]
interpretGenome gs genome =
  map instruction (geneDetails gs genome)

-- | Return each possible single-mutation variant of the 'Genome'
eachPossibleMutant ::
  GenomeSpec gs =>
  gs -> Genome gs -> [Genome gs]

eachPossibleMutant genomeSpec genome =
  [applyMutations genomeSpec [m] genome
   | m <- allPossibleMutations genomeSpec genome]


-- | (Utility) Go over an 'EncodingDetails' object and produce a
-- list of the result of applying a given function to all
-- segments.  The segments are reported in the order @[switch
-- pattern, threshold, target 0, delta 0, target 1, delta 1,
-- ...]@
flatTraverseEncodingDetails :: (SegmentEncoding -> result) -> EncodingDetails -> [result]
flatTraverseEncodingDetails f EncodingDetails{..} =
  concat [[f switchPatternSegment],
          [f thresholdSegment],
          concat [[f tgtSeg, f deltaSeg]
                 | (tgtSeg, deltaSeg) <- reactionSegments]]


instance
  (GenomeSpec gs,
   ShortShow (Interval gs)) =>
  ShortShow (GeneDetails gs) where
    shortShow GeneDetails{..} =
      (printf "%s:\n  %s\n  %s\n  %s\n  %s\n  %s\n  %s\n  %s\n"
       (shortShow interval)
       (shortShow instruction)
       (intercalate " " (flatTraverseEncodingDetails (bsTo01String . segmentCodonBits) instructionEncodingDetails))
       (intercalate " " (flatTraverseEncodingDetails (bsTo01String . segmentNucleotideBits) instructionEncodingDetails))
       (bsTo01String codonBits)
       (bsTo01String nucleotideBits)
       (bsToHexString nucleotideBits)
       (shortShow geneSource))


class (GenomeSpec gs, BlockSize (AlignmentSpec gs)) => Alignable gs where

  -- | For any parameters that might be required
  type AlignmentSpec gs :: *

  -- | Holds data computed by aligning two genomes
  type AlignmentResult gs :: *

  -- | Compute the best alignment between two genomes.  The first
  -- one is considered the reference sequence if the alignment
  -- system is asymmetric.
  bestAlignment :: gs -> AlignmentSpec gs -> Genome gs -> Genome gs -> AlignmentResult gs

  -- | Return the score resulting from genome alignment.  Higher
  -- means the genomes are more alike.
  alignmentScore :: gs -> AlignmentSpec gs -> AlignmentResult gs -> Int

  -- | Show the result of aligning two genomes.
  showAlignment :: gs -> AlignmentSpec gs -> AlignmentResult gs -> String

  -- | Compute a matrix of alignment scores.  This is a method of
  -- this class in case there's some optimization such as caching
  -- or symmetry that can be used to speed up mass alignment.
  -- Elements of the returned array should be numbered from @[0,
  -- n)@ where @n@ is the number of given genomes.  The default
  -- implementation just does the obvious thing, calling
  -- 'bestAlignment' and 'alignmentScore' for each pair of
  -- genomes.  In preliminary profiling experiments with
  -- 'HaploidSingletonGenome' and 'PolyploidGenome', most of the
  -- time in this is spent doing unavoidable Hamming distance
  -- calculations, so it's probably not worth spending time
  -- optimizing this unless you can speed up Hamming distance.
  alignmentMatrix :: gs -> AlignmentSpec gs -> Array Int (Genome gs) -> Array (Int, Int) Int

  alignmentMatrix gs as genomeArray =
    let
      genomeIxRange@(a,b) = bounds genomeArray
      results = [
        ((i, j), alignmentScore gs as ar)
        | i <- range genomeIxRange,
          j <- range genomeIxRange,
          let gi = genomeArray ! i,
          let gj = genomeArray ! j,
          let ar = bestAlignment gs as gi gj]
    in
     array ((a,a), (b,b)) (parListDeep results)
